package Metrics;

import java.util.ArrayList;

import workerCode.Coord;
import workerCode.Robot;
import workerCode.Queen;
import workerCode.Worker;

import simulation.Scout;
import simulation.SimulatedRobot;
import simulation.VirtualLocomotionDriver;
import simulation.VirtualScanDriver;

public class MetricRobotInformation {

	//General Information
	private Robot ID;
	private int level;
	
	//RunTime Information
	private long startTime=0;
	private long endTime=0;
	private long runtimeTotal=0;
	private long runtimeCalc=0;
	private long runtimeScan=0;
	private long runtimeDrive=0;
	
	//Data Transfer Information
	private int packetsSent=0;
	private int packetsReceived=0;
	
	//Driving Information
	private double distanceDriven=0;
	private int numPaths=0;
	
	//Scanning Information
	private int numScans=0;
	private int areaScanned=0;
	private int blockedSeen=0; //Unused
	
	/**
	 * Public Constructor for the metric information
	 * @param r
	 */ 
	public MetricRobotInformation(SimulatedRobot sr) {
		this.level = sr.getSwarmLevel();
	}
	
	public void update(SimulatedRobot sr) {
			this.numPaths = sr.getNumPaths();
			this.areaScanned = (int) sr.getAreaScanned();
			
			this.numScans = sr.getNumScans();
			this.distanceDriven = sr.getDistanceDriven();
	}
	
	/**
	 * Increment the number of packets sent by 1
	 */
	public void incrementPacketsSent(){
		this.packetsSent++;
	}
	
	/**
	 * Increment the number of packets receieved by 1
	 */
	public void incrementPacketsReceived(){
		this.packetsReceived++;
	}
	
	/**
	 * Updates the information pertaining to the robot driving
	 * @param startLocation
	 * @param path
	 */
	public void updatePathInformation(Coord startLocation, ArrayList<Coord> path){
		
		double additionalDriveDistance=0; //In cm
		
		//Update the number of paths the robot has been told to drive
		this.numPaths++;
		
		//Update the drive distance of the robot
		Coord temp1 = startLocation;
		Coord temp2 = new Coord();
		
		for (Coord wayPoint : path){
			temp2 = wayPoint;
			additionalDriveDistance += this.calculateDistance(temp1, temp2);
			temp1 = wayPoint;
		}
		
		this.distanceDriven += additionalDriveDistance;
		this.runtimeDrive += ((additionalDriveDistance / VirtualLocomotionDriver.driveSpeed) * 1000);
	}
	
	/**
	 * Caluclates the drive distance between two points
	 * @param temp1
	 * @param temp2
	 * @return
	 */
	private double calculateDistance(Coord temp1, Coord temp2) {
		return Math.sqrt(Math.pow(temp2.getX()-temp1.getX(), 2) + Math.pow(temp2.getY() - temp1.getY(), 2));
	}

	public void updateScanInformation(int radius){
		
		//Increment the number of times scanned
		this.numScans++;
		
		//Update the area scanned
		this.areaScanned += (Math.PI * Math.pow((double)radius, 2));
		
		//Update the time spent scannning
		//TODO PRIORITY 4 find a way of determining the cell conversions
		double degreeIncrement = Math.atan((double)10 / (double)radius);
		int numberOfScans = (int)((Math.PI * 2) / degreeIncrement);
		numberOfScans = numberOfScans / VirtualScanDriver.numServos;
		this.runtimeScan += numberOfScans * (VirtualScanDriver.rangeFinderSettleTime + VirtualScanDriver.servoSettleTime);
		
	}

	public void setStartTime() {
		this.startTime = System.currentTimeMillis();
	}

	public void setEndTime() {
		this.endTime = System.currentTimeMillis();
		
		this.runtimeCalc = this.endTime - this.startTime;
		this.runtimeTotal = this.runtimeCalc + this.runtimeDrive + this.runtimeScan;
		
	}
	
	@Override
	public String toString(){
		return new String(new String(
									 //this.ID + ", " +
									 this.getLevel() + ", " +
									 //this.runtimeTotal + ", " +
									 //this.runtimeScan + ", " +
									 //this.runtimeDrive + ", " +
									 //this.packetsSent + ", " +
									 //this.packetsReceived + ", " +
									 this.numPaths + ", " +
									 this.distanceDriven + ", " +
									 this.numScans + ", " +
									 this.areaScanned));
	}

	public void addToCount() {
		this.startTime++;
	}

	public long getStartTime() {
		return this.startTime;
	}

	public void setLevel(int level) {
		this.level = level;
	}

	public int getLevel() {
		return level;
	}

}
