package simulation;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;

import communication.Comm;
import communication.VirtualRemoteCommDriver;

import workerCode.Coord;
import workerCode.Robot;


public class SimulatedRobot implements Cloneable{
	private Robot robot;
	private SimulatedRobot parentRobot;
	private Collection<SimulatedRobot> children;
	
	//these fields technically exist in the robot however since it has 
	//pretty much no public access methods we need to redundantly
	//store them here.  This really should be fixed in the future
	private Coord currentLocation;
	private ArrayList<Coord> currentPath;
	private int currentScanRadius;
	private Comm robotComm;
	private VirtualRemoteCommDriver robotVirtualRemoteCommDriver;
	private HashMap<Integer, SimulatedRobot> idToRobotMap= new HashMap<Integer, SimulatedRobot>();
	private int swarmLevel; //0 = queen, 1= worker, 2= scout..etc
	
	private int state; //is the robot driving? scanning? idle?
	private boolean stateChanged; //track if this robot has changed state
	
	public static final int IDLE_STATE= 0;
	public static final int DRIVING_STATE= 1;
	public static final int SCANNING_STATE= 2;
	
	
	//Metric Data
	private int numPaths = 0;
	private int numScans = 0;
	private int distanceDriven = 0;
	private double areaScanned = 0;
	private int energyUsed = 0;
	
	public SimulatedRobot(Robot robot, SimulatedRobot parentRobot, Collection<SimulatedRobot> children, Coord startingLocation) {
		this.setRobot(robot);
		this.addParent(parentRobot);
		this.children = new ArrayList<SimulatedRobot>();
		for(SimulatedRobot child : children) {
			this.addChild(child);
		}
		
		this.setCurrentLocation(startingLocation);
		this.currentPath= new ArrayList<Coord>(); //empty path to start
	}
	
	public SimulatedRobot(Robot robot, Coord startingLocation) {
		this.children= new ArrayList<SimulatedRobot>();
		this.setRobot(robot);
		this.setCurrentLocation(startingLocation);
		this.currentPath= new ArrayList<Coord>(); //empty path to start
	}


	/**
	 * @param robot the robot to set
	 */
	public void setRobot(Robot robot) {
		this.robot = robot;
	}


	/**
	 * @return the robot
	 */
	public Robot getRobot() {
		return robot;
	}


	/**
	 * @param parentRobot the parentRobot to set
	 */
	public void setParentRobot(SimulatedRobot parentRobot) {
		this.parentRobot = parentRobot;
	}


	/**
	 * @return the parentRobot
	 */
	public SimulatedRobot getParentRobot() {
		return parentRobot;
	}


	/**
	 * @param children the children to set
	 */
	public void setChildren(Collection<SimulatedRobot> children) {
		this.children = children;
	}


	/**
	 * @return the children
	 */
	public Collection<SimulatedRobot> getChildren() {
		return children;
	}


	/**
	 * @param currentLocation the currentLocation to set
	 */
	public void setCurrentLocation(Coord currentLocation) {
		this.currentLocation = currentLocation.deepCopy();
	}


	/**
	 * @return the currentLocation
	 */
	public Coord getCurrentLocation() {
		return currentLocation.deepCopy();
	}


	/**
	 * @param currentPath the currentPath to set
	 */
	public void setCurrentPath(List<Coord> currentPath) {
		for(Coord c: currentPath) {
			this.currentPath.add(c.deepCopy());
		}
	}


	/**
	 * @return the currentPath
	 */
	public ArrayList<Coord> getCurrentPath() {
		ArrayList<Coord> retVal= new ArrayList<Coord>();
		for(Coord c : this.currentPath) {
			retVal.add(c.deepCopy());
		}
		return retVal;
	}
	
	/**
	 * Clear the current path.
	 */
	public void clearCurrentPath() {
		this.currentPath.clear();
	}

	/**
	 * @param currentScanRadius the currentScanRadius to set
	 */
	public void setCurrentScanRadius(int currentScanRadius) {
		this.currentScanRadius = currentScanRadius;
	}


	/**
	 * @return the currentScanRadius
	 */
	public int getCurrentScanRadius() {
		return currentScanRadius;
	}


	/**
	 * @param robotComm the robotComm to set
	 */
	public void setRobotComm(Comm robotComm) {
		this.robotComm = robotComm;
	}


	/**
	 * @return the robotComm
	 */
	public Comm getRobotComm() {
		return robotComm;
	}


	/**
	 * @param robotVirtualRemoteCommDriver the robotVirtualRemoteCommDriver to set
	 */
	public void setRobotVirtualRemoteCommDriver(
			VirtualRemoteCommDriver robotVirtualRemoteCommDriver) {
		this.robotVirtualRemoteCommDriver = robotVirtualRemoteCommDriver;
	}

	/**
	 * @return the robotVirtualRemoteCommDriver
	 */
	public VirtualRemoteCommDriver getRobotVirtualRemoteCommDriver() {
		return robotVirtualRemoteCommDriver;
	}

	/**
	 * @param swarmLevel the swarmLevel to set
	 */
	public void setSwarmLevel(int swarmLevel) {
		this.swarmLevel = swarmLevel;
	}


	/**
	 * @return the swarmLevel
	 */
	public int getSwarmLevel() {
		return swarmLevel;
	}


	/**
	 * @param state the state to set
	 */
	public void setState(int state) {
		if(this.state!= state) {
			this.stateChanged= true;
		}
		this.state = state;
	}


	/**
	 * @return the state
	 */
	public int getState() {
		return state;
	}
	
	/**
	 * Set the state and update the appropriate field.
	 * If the state is set to idle stateInfo can be null to simply clear out
	 * the path/scan radius fields or set to a Coord to update the current
	 * location of the robot.
	 * If the state is set to driving then the stateinfo must be set to the
	 * path the robot is driving along.
	 * If the state is set to scanning then the stateInfo must be an Integer
	 * (or an int) containing the new scan radius.
	 * In all cases, fields which are no longer relevant (such as the scan
	 * radius while in driving state) will be cleared or set to 0.
	 */
	@SuppressWarnings("unchecked")
	public void setState(int state, Object stateInfo) {
		this.setState(state);
		if(state == SimulatedRobot.IDLE_STATE) {
			//clear the path
			this.clearCurrentPath();
			//clear scan radius
			this.setCurrentScanRadius(0);
			if(stateInfo != null) {
				this.setCurrentLocation((Coord)stateInfo);
			}
		}
		else if(state == SimulatedRobot.DRIVING_STATE) {
			//stateinfo is the path
			this.setCurrentPath((List<Coord>) stateInfo);
			this.setCurrentScanRadius(0);
			this.updatePathMetricInfo((List<Coord>) stateInfo);
		}
		else if(state == SimulatedRobot.SCANNING_STATE) {
			this.setCurrentScanRadius((Integer)stateInfo);
			this.clearCurrentPath();
			this.updateScanMetricInfo((Integer) stateInfo);
		}
	}
	
	private void updateScanMetricInfo(Integer stateInfo) {
//		System.out.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@" + numScans);
		this.numScans ++;
		this.areaScanned += Math.PI * stateInfo * stateInfo;
		
	}

	private void updatePathMetricInfo(List<Coord> stateInfo) {
//		System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" + numPaths);
		this.numPaths++;
		if (stateInfo.isEmpty()){
			System.err.println("PATH EMTPY");
		}
		else if (stateInfo.size() == 1){
			System.err.println("PATH Size 1");
		}
		
		for (int i = 1; i < stateInfo.size(); i++){
			Coord pos1 = stateInfo.get(i - 1);
			Coord pos2 = stateInfo.get(i);
			this.distanceDriven += Math.sqrt(Math.pow(pos1.getX()-pos2.getX(), 2) + Math.pow(pos1.getY() - pos2.getY(), 2));
		}
	}

	/**
	 * Check if the state of this robot has changed.  Checking the state will
	 * also clear the state change.
	 * @return
	 */
	public boolean hasChangedState() {
		if(this.stateChanged) {
			this.stateChanged= false;
			return true;
		}
		else {
			return false;
		}
	}
	
	/**
	 * Add a child to this simulated robot.
	 */
	public void addChild(SimulatedRobot child) {
		this.children.add(child);
		this.idToRobotMap.put(this.children.size(), child);
	}
	
	/**
	 * Add a parent to this simulated robot
	 */
	public void addParent(SimulatedRobot parent) {
		this.parentRobot= parent;
		this.idToRobotMap.put(0, parent);
	}
	
	/**
	 * Look up a robot from the ID.  The id's are only unique to a single
	 * robot so this function will translate the robot-specific ID # into
	 * a globally unique SimulatedRobot object.
	 * @param senderID The id to look up.
	 * @return The SimulatedRobot which wraps the robot associated with robotID
	 */
	public SimulatedRobot lookupRobotFromID(int robotID) {
		return this.idToRobotMap.get(robotID);
	}
	
	@Override
	public Object clone() {
		SimulatedRobot retVal= new SimulatedRobot(this.robot, this.getParentRobot(), this.getChildren(), this.getCurrentLocation());
		retVal.setCurrentPath(this.getCurrentPath());
		retVal.setState(this.getState());
		retVal.setCurrentPath(this.getCurrentPath());
		retVal.setCurrentScanRadius(this.getCurrentScanRadius());
		retVal.setRobotComm(this.getRobotComm());
		retVal.setSwarmLevel(this.getSwarmLevel());
		retVal.setNumPaths(this.getNumPaths());
		retVal.setNumScans(this.getNumScans());
		retVal.setDistanceDriven(this.getDistanceDriven());
		retVal.setAreaScanned(this.getAreaScanned());
		retVal.setEnergyUsed(this.getEnergyUsed());
		
		return retVal;
	}
	
	@Override
	public int hashCode(){
		return this.robot.hashCode();
	}
	
	@Override
	public boolean equals(Object o){
		if (this == o)
			return true; // Return true if the objects are the same object
		if (o == null)
			return false; // Return false if they are not the same object and o
							// is null
		if (this.getClass() != o.getClass())
			return false; // Return false if they are a differetn class
		else { // They are not a different class
			SimulatedRobot sr = (SimulatedRobot)o;
			if (sr.getRobot() == this.getRobot()) return true;
			return false;
		}
	}

	public void setNumPaths(int numPaths) {
		this.numPaths = numPaths;
	}

	public int getNumPaths() {
		return numPaths;
	}

	public void setNumScans(int numScans) {
		this.numScans = numScans;
	}

	public int getNumScans() {
		return numScans;
	}

	public void setDistanceDriven(int distanceDriven) {
		this.distanceDriven = distanceDriven;
	}

	public int getDistanceDriven() {
		return distanceDriven;
	}

	public void setAreaScanned(double areaScanned) {
		this.areaScanned = areaScanned;
	}

	public double getAreaScanned() {
		return areaScanned;
	}

	public void setEnergyUsed(int energyUsed) {
		this.energyUsed = energyUsed;
	}

	public int getEnergyUsed() {
		return energyUsed;
	}

}
