package simulation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Observable;
import java.util.Observer;

import communication.*;

import workerCode.Coord;
import workerCode.ProbMap;

public class Simulation extends Observable implements Observer, Runnable{
	private ProbMap realWorldMap; //real world map with actual object locations
	private ProbMap observedMap;  //map as seen by the swarm
	
	private Collection<SimulatedRobot> simRobots; //the robots themselves
		
	public Simulation(ProbMap realMap, Collection<SimulatedRobot> simRobots) {
		this.realWorldMap= realMap;
		this.simRobots= simRobots;
		this.observedMap= new ProbMap(realMap.getxSize(), realMap.getySize(),
				realMap.getCellConversion(), ProbMap.NO_OFFSET);
		//observe the comm of every simulated robot
		for(SimulatedRobot simr : simRobots) {
			simr.getRobotComm().addObserver(this);
		}
	}

	@Override
	public synchronized void update(Observable arg0, Object arg1) {
		//got a packet
		if(arg0 instanceof Comm && arg1 instanceof SwarmCommunication) {
			this.rawSwarmCommunicationCallback((Comm)arg0, (SwarmCommunication)arg1);
		}
	}

	@Override
	public void run() {
		//start the swarm
		for(SimulatedRobot simr : this.simRobots) {
//			System.out.println(simr.getRobot());
			Thread t= new Thread(simr.getRobot());
			t.start();
		}
		
	}
	
	/**
	 * Callback which is called whenever any packet is received by any robot
	 * in the swarm.
	 * @param recipientComm The comm which received this packet.
	 * @param receivedCommunication The wrapper containing the packet and the
	 * comm-specific ID identifying the sender.
	 */
	private void rawSwarmCommunicationCallback(Comm recipientComm,
			                                   SwarmCommunication receivedCommunication) {
		//lookup the recipient
		SimulatedRobot recipient= this.lookupRecipientRobot(recipientComm);
		//lookup the sender
		SimulatedRobot sender= recipient.lookupRobotFromID(receivedCommunication.getID());
		this.packetInterceptCallback(sender, recipient, receivedCommunication.getPacket());
		this.setChanged();
		this.notifyObservers(receivedCommunication);
	}
	

	/**
	 * Callback which is called whenever the simulation intercepts any packet
	 * which is received by any robot in the swarm.
	 * @param sender The SimulatedRobot wrapping the robot which sent the packet
	 * @param recipient The SimulatedRobot wrapping the robot which received the packet.
	 * @param packet The packet which was received.
	 */
	private void packetInterceptCallback(SimulatedRobot sender,
			                             SimulatedRobot recipient,
			                             SwarmPacket packet) {
		//check what kind of packet was received
		if(packet instanceof ScanPacket) {
			this.scanPacketInterceptHandler(sender, recipient, (ScanPacket) packet);
		}
		else if(packet instanceof PathPacket) {
			this.pathPacketInterceptHandler(sender, recipient, (PathPacket) packet);
		}
		else if(packet instanceof ScanCompletedUnknownShadowingPacket) {
			this.scanCompleteUnknownShadowingPacketInterceptHandler(sender,
					recipient, (ScanCompletedUnknownShadowingPacket) packet);
		}
		else if(packet instanceof ScanCompletedKnownShadowingPacket) {
			this.scanCompletedKnownShadowingPacketInterceptHandler(sender,
					recipient, (ScanCompletedKnownShadowingPacket) packet);
		}
		else if(packet instanceof MapPacket) {
			this.mapPacketInterceptHandler(sender, recipient, (MapPacket)packet);
		}
		else if(packet instanceof ReachedDestinationPacket) {
			this.reachedDestinationPacketInterceptHandler(sender, recipient,
					(ReachedDestinationPacket)packet);
		}
		else if(packet instanceof ChildInformationPacket) {
			this.childInformationPacketInterceptHandler(sender, recipient,
					(ChildInformationPacket)packet);
		}
		else if(packet instanceof AddPointToMapPacket) {
			this.addPointToMapPacketInterceptHandler(sender, recipient,
					(AddPointToMapPacket)packet);
		}
		else if(packet instanceof KillPacket) {
			this.killPacketInterceptHandler(sender, recipient, (KillPacket)packet);
		}
	}

	private void scanPacketInterceptHandler(SimulatedRobot sender,
			SimulatedRobot recipient, ScanPacket packet) {
		recipient.setState(SimulatedRobot.SCANNING_STATE, packet.getRadius());
	}
	
	private void pathPacketInterceptHandler(SimulatedRobot sender,
			SimulatedRobot recipient, PathPacket packet) {
		recipient.setState(SimulatedRobot.DRIVING_STATE, packet.getPath());
	}
	
	private void scanCompleteUnknownShadowingPacketInterceptHandler(
			SimulatedRobot sender, SimulatedRobot recipient,
			ScanCompletedUnknownShadowingPacket packet) {
		this.updateObservedMapShadowing(sender, sender.getCurrentScanRadius());
		sender.setState(SimulatedRobot.IDLE_STATE, null);
	}
	
	private void scanCompletedKnownShadowingPacketInterceptHandler(
			SimulatedRobot sender, SimulatedRobot recipient,
			ScanCompletedKnownShadowingPacket packet) {
		
		this.updateObservedMapShadowing(sender.getCurrentLocation(),
				sender.getCurrentScanRadius(), packet.getShadowedPoints());
		
		sender.setState(SimulatedRobot.IDLE_STATE, null);
	}
	
	private void mapPacketInterceptHandler(SimulatedRobot sender,
			SimulatedRobot recipient, MapPacket packet) {
		//TODO we can probably ignore these?
	}
	
	private void reachedDestinationPacketInterceptHandler(
			SimulatedRobot sender, SimulatedRobot recipient,
			ReachedDestinationPacket packet) {
		sender.setState(SimulatedRobot.IDLE_STATE, packet.getDestination());
		sender.clearCurrentPath();
	}
	
	private void childInformationPacketInterceptHandler(SimulatedRobot sender,
			SimulatedRobot recipient, ChildInformationPacket packet) {
		sender.setState(SimulatedRobot.IDLE_STATE, packet.getLocation());
	}
	
	private void addPointToMapPacketInterceptHandler(SimulatedRobot sender,
			SimulatedRobot recipient, AddPointToMapPacket packet) {
		//update our map
		this.updateObservedMap(packet.getLocation(), packet.getStdDev());
	}
	
	private void killPacketInterceptHandler(SimulatedRobot sender,
			SimulatedRobot recipient, KillPacket packet) {
//		System.out.println("Simulation got a kill packet");
		//TODO ignore these for now
	}

	private SimulatedRobot lookupRecipientRobot(Comm receivingComm) {
		for(SimulatedRobot r : this.simRobots) {
			if(r.getRobotComm() == receivingComm) {
				return r;
			}
		}
		return null;
	}
	

	public synchronized void updateObservedMapShadowing(SimulatedRobot robot, int scanRadius) {
		this.observedMap.updateCompletedScanAreaUnknownShadowingSimulation(
				robot.getCurrentLocation(), scanRadius, robot.getParentRobot().getCurrentLocation(),
				robot.getParentRobot().getCurrentScanRadius());
	}
	
	public synchronized void updateObservedMapShadowing(Coord robotLocation, int scanRadius, Collection<Coord> shadowPoints) {
		this.observedMap.updateCompletedScanAreaUnseenSpaces(
				robotLocation, scanRadius, shadowPoints);
	}
	/**
	 * @param observedMap the observedMap to set
	 */
	public synchronized void updateObservedMap(Coord objectLocation, int stdDeviation) {
		this.observedMap.updateCellAreaBlockValue(objectLocation, stdDeviation);
	}

	/**
	 * @return the observedMap
	 */
	public synchronized ProbMap getObservedMap() {
		return new ProbMap(observedMap);
	}
	
	/**
	 * @return the realWorldMap
	 */
	public ProbMap getRealworldMap() {
		return this.realWorldMap;
	}

	/**
	 * @return the simRobots
	 */
	public Collection<SimulatedRobot> getSimRobots() {
		ArrayList<SimulatedRobot> simRobotsCopy= new ArrayList<SimulatedRobot>();
		for(SimulatedRobot sr : this.simRobots) {
			simRobotsCopy.add((SimulatedRobot) sr.clone());
		}
		return simRobotsCopy;
	}
	
	public static final void main(String[] args) {
		ProbMap map= new ProbMap(200, 200, 10, ProbMap.NO_OFFSET);
		
		for (int x = (map.getxSize() / 2) - 20; x < (map.getxSize() / 2); x++){
			for (int y = (map.getySize() / 2) - 20; y < (map.getySize() / 2); y++){
				map.cellSetBlockValue(new Coord(x,y), 1.0);
			}
		}
		
		
		ArrayList<Coord> workerStartingLocations= new ArrayList<Coord>();
		workerStartingLocations.add(new Coord(80, 0));
		workerStartingLocations.add(new Coord(240, 0));
		
		ArrayList<Coord> scoutStartingLocations= new ArrayList<Coord>();
		scoutStartingLocations.add(new Coord(100, 0));
		scoutStartingLocations.add(new Coord(140, 0));
		scoutStartingLocations.add(new Coord(260, 0));
		scoutStartingLocations.add(new Coord(300, 0));
		SimulationStarter swarmMaker= new SimulationStarter();
		Collection<SimulatedRobot> swarmBots;
		swarmBots= swarmMaker.setupSwarm(2, workerStartingLocations, 2, scoutStartingLocations, map);
		Simulation sim= new Simulation(map, swarmBots);
		sim.run();
	}

}
