package communication;

import gnu.io.CommPortIdentifier;
import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Vector;


public class XbeeAPIFrameRouter implements Runnable, SerialPortEventListener {

	private XbeeRemoteCommDriver xbeeDriver; //the xbee driver to route to
	
	private CommPortIdentifier serialPortID; //serial port manager
	private SerialPort serialPort;           //the serial port itself
	private ArrayList<Byte> xbeePacketBuffer;
	private Vector<Byte> rawSerialReadBuffer= new Vector<Byte>();
	private boolean isRunning; //verifies that driver should still be running
	
	private Object txStatusSemaphore= new Object();
	private int txDeliveryStatus;
	private int outboundFrameID= 1;
	private int rxFrameID= 0;
	private boolean waitingForTxStatus= false;
	private ArrayList<RxPacketAPI> rxPacketBacklog= new ArrayList<RxPacketAPI>();
	
	
	public XbeeAPIFrameRouter(CommPortIdentifier serialPortID, XbeeRemoteCommDriver xbeeDriver) {
		this.xbeeDriver= xbeeDriver;
		this.serialPortID= serialPortID;

		try {
			//try to open the given serial port using the name Xbee HSR
			//and timing out after 1000ms
			this.serialPort= (SerialPort) this.serialPortID.open("XBee HSR", 1000);
			//set up the serial port for 11500, 8 data bits, 1 stop bit, no parity
			this.serialPort.setSerialPortParams(115200,
                    							SerialPort.DATABITS_8,
                    							SerialPort.STOPBITS_1,
                    							SerialPort.PARITY_NONE);
			//no flow control
			this.serialPort.setFlowControlMode(SerialPort.FLOWCONTROL_NONE);
			//set up to notify that data is available
			this.serialPort.notifyOnDataAvailable(true);
			//register this as a listener
			this.serialPort.addEventListener(this);
			
		} catch (Exception e) {
			//if an exception occurs then something is screwed up
			e.printStackTrace();
		}
		this.xbeePacketBuffer= new ArrayList<Byte>();
		this.isRunning= true;
	}

	@Override
	public void run() {
		//while the driver is running
		while(this.isRunning) {
			boolean packetComplete= false;
			int totalLength= Integer.MAX_VALUE;
			byte checkSum= 0;
			//while the packet is not complete
			while(!packetComplete) {
				//wait for data to arrive and add it to the buffer
				this.xbeePacketBuffer.add(serialPortRead());
				//this is kind of ghetto but this is here to make sure that
				//when the serial port is closed by kill we return to kill
				//this thread
				if(this.isRunning == false) {
					return;
				}
				//make sure the start delim is the first byte
				if(this.xbeePacketBuffer.size() == 1) {
					//flush out the buffer if the first byte is wrong
					if(this.xbeePacketBuffer.get(0) != XbeeAPIFrame.XBEE_START_DELIM) {
						this.xbeePacketBuffer.clear();
					}
				}
				//if enough bytes have been read in then set the packet length
				if(this.xbeePacketBuffer.size() == 3) {
					byte tempBuf[]= new byte[2];
					tempBuf[0]= this.xbeePacketBuffer.get(1);
					tempBuf[1]= this.xbeePacketBuffer.get(2);
					totalLength= TypeConverter.bytesToShort(tempBuf)+4;
				}
				//if the entire packet has been read in
				if(this.xbeePacketBuffer.size() >= totalLength) {
					packetComplete= true;
				}
			}
			//packet is fully formed
			//check the checksum byte
			checkSum= calculateChecksum(this.xbeePacketBuffer);
			//if the checksum is good then move on, otherwise....crap
			if(checkSum != this.xbeePacketBuffer.get(this.xbeePacketBuffer.size() - 1).byteValue()) {
				//crap
			}
			
			else {
				//handle the packet
				this.handleXbeePacket(this.xbeePacketBuffer);
			}
			
			//reset for next run
			this.xbeePacketBuffer.clear();
		}
	}

	public boolean send(XbeeRemoteCommDriver.XbeeAddressEntry entry, byte[] serialData) {
		boolean txSuccess= false; //return value
		//if entry is a null then just return right away
		//technically we "failed" to send but we don't want to retry
		//or anything so count sending to a null as a success.
		if(entry == null) {
			return true;
		}
		//segment data
		int numSegments= serialData.length/251 + 1;
		
		for(int i= 0; i < numSegments; i++) {
			byte[] segmentedSerialData= this.getNextSegment(serialData, numSegments, i);
			
			
			//create tx api frame
			TxReqAPI txFrame= new TxReqAPI(this.outboundFrameID, entry.addr64bit,
					                       entry.addr16bit, 0, 0, segmentedSerialData);
			
			//serialize the frame
			byte[] serialFrame= txFrame.serialize();
			//write out serial port and set the waiting for tx status flag to true
			this.setWaitingForTxStatus(true);
			try {
				this.serialPort.getOutputStream().write(serialFrame);
			} catch (IOException e) {
				e.printStackTrace();
			}
			
			
			//wait for tx status
			synchronized(this.txStatusSemaphore) {
				while(this.rxFrameID != this.outboundFrameID) {
//					System.out.println("waiting for id "+this.outboundFrameID);
					try {
						this.txStatusSemaphore.wait();
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				//no longer waiting for tx status
				this.setWaitingForTxStatus(false);
				if(this.txDeliveryStatus == 0x00) {
					txSuccess= true;
				}
				else {
					txSuccess= false;
				}
			}
			//increment frame id for next time
			this.outboundFrameID++;
			if(this.outboundFrameID > 0xFF) this.outboundFrameID= 1;
		}

		return txSuccess;
	}

	public void kill() {
		//do cleanup stuff
		this.isRunning= false;
		synchronized(this.serialPort) {
			this.serialPort.notifyAll();
		}
		this.serialPort.close();
	}
	/**
	 * Event handler for serial port events.
	 */
	@Override
	public void serialEvent(SerialPortEvent event) {
		//if data is available
		if(event.getEventType() == SerialPortEvent.DATA_AVAILABLE) {
			synchronized(this.serialPort) {
				int c= -1;
				do {
					try {
						c= this.serialPort.getInputStream().read();
					} catch (IOException e) {
						e.printStackTrace();
					}
					if (c != -1) {
						this.rawSerialReadBuffer.add((byte)c);
						this.serialPort.notifyAll();
					}
				} while(c != -1);
				
			}
		}
		
	}
	
	
	/**
	 * Read a single byte off of the serial port.  This function blocks if
	 * there is currently no data available to read.
	 * @return The next byte on the input stream of the serial port.
	 */
	private Byte serialPortRead() {
		synchronized(this.serialPort) {
			while(this.rawSerialReadBuffer.size() == 0) {
				try {
					this.serialPort.wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			return this.rawSerialReadBuffer.remove(0);
		}
	}
	
	/**
	 * Calculate the checksum of a serial API frame.  This is the same
	 * as the checksum method defined in XbeeAPIFrame but this one works
	 * for an arraylist input
	 * @param serialFrame The frame to calculate the checksum of.
	 * @return The expected checksum value for the given frame.
	 */
	private byte calculateChecksum(ArrayList<Byte> serialFrame) {
		int retVal= 0;
		int frameLength;
		byte[] tempBuf= new byte[2];
		tempBuf[0]= serialFrame.get(1);
		tempBuf[1]= serialFrame.get(2);
		frameLength= TypeConverter.bytesToShort(tempBuf);
		for(int i= 3; i < frameLength + 3; i++) {
			retVal+= serialFrame.get(i).byteValue();
		}
		//remove all but last byte
		retVal&= 0xFF;
		
		return (byte)(0xFF - retVal);
	}
	
	private void handleXbeePacket(ArrayList<Byte> serialFrame) {
		int frameType= serialFrame.get(3).byteValue();
		frameType&= 0xFF;
		
		if(frameType == XbeeAPIFrame.TX_STATUS) {
			TxStatusAPI txStatusPacket= new TxStatusAPI();
			txStatusPacket.deSerialize(this.copyRxBuffer());
			this.handleTxStatusAPI(txStatusPacket);
			//if there are any waiting rx packets in backlog do those
			while(this.rxPacketBacklog.size() != 0) {
				this.handleRxAPI(this.rxPacketBacklog.remove(0));
			}
		}
		
		else if(frameType == XbeeAPIFrame.RX_PACKET) {
			RxPacketAPI rxPacket= new RxPacketAPI();
			rxPacket.deSerialize(this.copyRxBuffer());
			if(this.isWaitingForTxStatus()) {
				//buffer the packet
				this.rxPacketBacklog.add(rxPacket);
			}
			else {
				this.handleRxAPI(rxPacket);
			}
		}
		
		else if(frameType == XbeeAPIFrame.MODEM_STATUS) {
			//don't really care about this
		}
	}
	
	private byte[] copyRxBuffer() {
		byte[] retVal= new byte[this.xbeePacketBuffer.size()];
		for(int i= 0; i < this.xbeePacketBuffer.size(); i++) {
			retVal[i]= this.xbeePacketBuffer.get(i);
		}
		
		return retVal;
	}
	
	private void handleRxAPI(RxPacketAPI rxPacket) {
		this.xbeeDriver.setRxPacket(rxPacket);
	}
	
	private void handleTxStatusAPI(TxStatusAPI txStatusPacket) {
//		System.out.println("got frame id "+ txStatusPacket.getFrameID());
		synchronized(this.txStatusSemaphore) {
			this.rxFrameID= txStatusPacket.getFrameID();
			this.txDeliveryStatus= txStatusPacket.getDeliveryStatus();
			this.txStatusSemaphore.notifyAll();
		}
	}
	
	private synchronized void setWaitingForTxStatus(boolean b) {
		this.waitingForTxStatus= b;
	}
	private synchronized boolean isWaitingForTxStatus() {
		return this.waitingForTxStatus;
	}
	

	
	/**
	 * Separate out the serial data into sub packets that are 255 bytes long
	 * and contain the sequence information to reconstruct them later.
	 * @param serialData The data to segment
	 * @param totalSegments The total number of segments
	 * @param segmentNumber Which segment we are up to.
	 * @return An array of byte arrays containing 
	 */
	private byte[] getNextSegment(byte[] serialData, int totalSegments, int segmentNumber) {
		byte[] retVal;
		int retValLength= serialData.length - 251*segmentNumber + 4;
		if(retValLength > 255) retValLength = 255;
		retVal= new byte[retValLength];
		byte[] tempByte;
		tempByte= TypeConverter.shortToBytes((short)totalSegments);
		retVal[0]= tempByte[0];
		retVal[1]= tempByte[1];
		tempByte= TypeConverter.shortToBytes((short)segmentNumber);
		retVal[2]= tempByte[0];
		retVal[3]= tempByte[1];
		for(int i= 4; i < retVal.length; i++) {
			retVal[i]= serialData[i - 4 + 251*segmentNumber];
		}
		
		return retVal;
		
	}

}
