package communication;

import java.util.ArrayList;


import gnu.io.*;

public class XbeeRemoteCommDriver extends RemoteCommDriver {

	private boolean isRunning= true;
	private XbeeAPIFrameRouter apiRouter;
	private ArrayList<RxPacketAPI> rxPacketBuffer= new ArrayList<RxPacketAPI>();
	private ArrayList<XbeeAddressEntry> addressTable= new ArrayList<XbeeAddressEntry>();
	private ArrayList<ArrayList<Byte>> segmentPacketBuffer;
	private static short ADDR_16_UNKNOWN= (short) 0xFFFE;
	private static long ADDR_64_UNKNOWN= 0xFFFFFFFFFFFFFFFFL;
	
	
	
	/**
	 * Creates a driver capable of talking to an xbee module via a serial
	 * port.  The serial port to be used must be passed in as an argument
	 * @param serialPortID The ID of the serial port to use.
	 */
	public XbeeRemoteCommDriver(CommPortIdentifier serialPortID) {
		this.apiRouter= new XbeeAPIFrameRouter(serialPortID, this);
		this.segmentPacketBuffer= new ArrayList<ArrayList<Byte>>();
		Thread apiRouterThread= new Thread(this.apiRouter);
		apiRouterThread.start();
	}

	@Override
	public void run() {
		while(this.isRunning) {
			RxPacketAPI rxPacket= this.getRxPacket();
			XbeeAddressEntry address= new XbeeAddressEntry(rxPacket.getAddr64bit(), rxPacket.getAddr16bit());
			int ad= lookupAddressDescriptor(address);
			//if address is from a valid sender
			if (ad >= 0) {
				//update the 16 bit address in the table
				this.addressTable.get(ad).addr16bit= rxPacket.getAddr16bit();
				byte[] rxData=  rxPacket.getRxData();
				for(int i= 4; i < rxData.length; i++) {
					this.segmentPacketBuffer.get(ad).add(rxData[i]);
				}
				//check if sequence is complete
				byte[] tempBuf= new byte[2];
				tempBuf[0]= rxData[0];
				tempBuf[1]= rxData[1];
				int numSegments= TypeConverter.bytesToShort(tempBuf);
				tempBuf[0]= rxData[2];
				tempBuf[1]= rxData[3];
				int segmentNumber= TypeConverter.bytesToShort(tempBuf);
				if(segmentNumber == (numSegments - 1)) {
					//full packet assembled
					byte[] fullRxPacket= new byte[this.segmentPacketBuffer.get(ad).size()];
					for(int i= 0; i < fullRxPacket.length; i++) {
						fullRxPacket[i]= this.segmentPacketBuffer.get(ad).get(i).byteValue();
					}
					this.segmentPacketBuffer.get(ad).clear();
					RemoteCommDriver.ReceiveDataArray rx;
					rx= new RemoteCommDriver.ReceiveDataArray((byte)ad, fullRxPacket);
					this.setChanged();
					this.notifyObservers(rx);
				}

			}
		}
	}


	@Override
	public boolean send(byte id, byte[] serialData) {
		XbeeAddressEntry entry= this.addressTable.get(id);
		return this.apiRouter.send(entry, serialData);
	}

	@Override
	public void kill() {
		this.apiRouter.kill();
		this.isRunning= false;
		synchronized(this) {
			this.notifyAll();
		}
	}
	
	public synchronized void setRxPacket(RxPacketAPI rxPacket) {
		this.rxPacketBuffer.add(rxPacket);
//		System.out.println("set rx packet");
		this.notifyAll();
	}
	
	private synchronized RxPacketAPI getRxPacket() {
		while(this.rxPacketBuffer.isEmpty()) {
			try {
				this.wait();
			}
			catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		return this.rxPacketBuffer.remove(0);
	}
	
	
	private int lookupAddressDescriptor(XbeeAddressEntry entry) {
		for(int i= 0; i < this.addressTable.size(); i++) {
			if(this.addressTable.get(i) != null) {
				if(this.addressTable.get(i).equals(entry)) {
					return i;
				}
			}
		}
		return -1;
	}
	
	/**
	 * Add the given 64 bit address to the table.
	 * @param addr64bit The address to add.
	 * @return The address descriptor associated with the given address.
	 */
	public int addAddressTableEntry(long addr64bit) {
		XbeeAddressEntry newEntry= new XbeeAddressEntry(addr64bit, ADDR_16_UNKNOWN);
		this.addressTable.add(newEntry);
		this.segmentPacketBuffer.add(new ArrayList<Byte>());
		return this.addressTable.indexOf(newEntry);
	}
	
	public int addAddressTableEntry(long addr64bit, int index) {
		XbeeAddressEntry newEntry= new XbeeAddressEntry(addr64bit, ADDR_16_UNKNOWN);
		while(index >= this.addressTable.size()) {
			this.addressTable.add(null);
		}
		this.addressTable.set(index, newEntry);
		
		while(index >= this.segmentPacketBuffer.size()) {
			this.segmentPacketBuffer.add(null);
		}
		this.segmentPacketBuffer.set(index, new ArrayList<Byte>());
		
		return index;
	}
	
	protected class XbeeAddressEntry {
		public long addr64bit;
		public short addr16bit;
		
		public XbeeAddressEntry(long addr64bit, int addr16bit) {
			this.addr64bit= addr64bit;
			this.addr16bit= (short)addr16bit;
		}
		@Override
		public boolean equals(Object obj) {
			if(!(obj instanceof XbeeAddressEntry)) {
				return false;
			}
			else {
				XbeeAddressEntry entry= (XbeeAddressEntry) obj;
				if((this.addr64bit == entry.addr64bit) || (this.addr16bit == entry.addr16bit)) {
					return true;
				}
				else { 
					return false;
				}
			}
		}
	}
	

}
