package workerCode;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

/**
 * 
 */

/**
 * @author hsrmqp
 *
 */
public class Child {

	//Needs to be populated from robot information
	private Coord currentLocation;
	private int myMinScanRadius = 10;
	private int myMaxScanRadius = 80;
	private boolean requiresMap;
	
	//Status Flags
	private boolean busy = false;
	
	//Used for The Coverage Algorithm
	private ArrayList<Coord> drivableList = new ArrayList<Coord>();
	private ArrayList<Coord> placeableList = new ArrayList<Coord>();
	private boolean shouldRunDijkstra = false;
	private boolean shouldTryToPlaceFlag = false;
	
	private Coord possibleNextLocation = new Coord();
	private Coord nextLocation = new Coord();
	private int scanRadiusToUse = 10;
	
	private boolean move = false;
	private boolean scan = false;
	
	//Used by PathPlanner
	private int wave;
	private ArrayList<Coord> allPointsInPathList = new ArrayList<Coord>();
	private boolean pathListMade;
	
	//Used by Dijkstra
	private ProbMap myDijkstraMap = null;
	private ArrayList<Coord> dijkstraFinalPath = new ArrayList<Coord>();
	private int ID = -1;
	
	/**
	 * Blank Constructor
	 */
	public Child(){
		//blank by design
	}
	
	/**
	 * Constructor for a child, only three things are needed when it starts which are the position, and scan radius information
	 * 
	 * @param startPosition
	 * @param minScanRadius
	 * @param maxScanRadius
	 */
	public Child(int ID, Coord startPosition, int minScanRadius, int maxScanRadius, boolean requiresMap) {
		this.ID = ID;
		this.setCurrentLocation(startPosition);
		this.setMinScanRadius(minScanRadius);
		this.setMaxScanRadius(maxScanRadius);
		this.setRequiresMap(requiresMap);
		this.setBusy(true);
	}
	
	public Child(int ID, Coord startPosition, int minScanRadius, int maxScanRadius) {
		this.ID = ID;
		this.setCurrentLocation(startPosition);
		this.setMinScanRadius(minScanRadius);
		this.setMaxScanRadius(maxScanRadius);
		this.setRequiresMap(false);
		this.setBusy(true);
	}
	

	/**
	 * Sets the current location of the child
	 * 
	 * Used by the coverage algorithm and then updated by locomotion
	 * 
	 * @param currentLocation
	 */
	public void setCurrentLocation(Coord currentLocation) {
		this.currentLocation = currentLocation.deepCopy();
	}
	
	/**
	 * Gets the current location of the child
	 * @return
	 */
	public Coord getCurrentLocation() {
		return currentLocation.deepCopy();
	}
	
	/**
	 * Sets the next possible location of the child
	 * 
	 * Used by the coverage algorithm to keep track of where the robot may be placed next
	 * @param possibleNextLocation
	 */
	public void setPossibleNextLocation(Coord possibleNextLocation) {
		this.possibleNextLocation = possibleNextLocation;
	}
	
	/**
	 * Gets the next possible location of the child
	 * 
	 * Used by the coverage algorithm
	 * @return
	 */
	public Coord getPossibleNextLocation() {
		return possibleNextLocation;
	}
	
	/**
	 * Sets the next end location of the robot (the final destination of its next path)
	 * @param nextLocation
	 */
	public void setNextLocation(Coord nextLocation) {
		this.nextLocation = nextLocation;
	}
	
	/**
	 * Gets the next end location of the robot
	 * @return
	 */
	public Coord getNextLocation() {
		return nextLocation;
	}
	
	/**
	 * Get the list of coords that the robot can successfully drive to
	 * @return
	 */
	public ArrayList<Coord> getDrivableList() {
		return drivableList;
	}
	
	/**
	 * Get the list of coords that the robot can be placed at this iteration
	 * @return
	 */
	public ArrayList<Coord> getPlaceableList() {
		return placeableList;
	}
	
	/**
	 * Set the should run flag, indicates that the coverage algorithm should run for this child
	 * @param coverageAlgorithmRunFlag
	 */
	public void setShouldRunFlag(boolean coverageAlgorithmRunFlag) {
		this.shouldRunDijkstra = coverageAlgorithmRunFlag;
	}
	
	/**
	 * Gets the should run flag, indicates whether or not the coverage algorithm should run for this child
	 * @return
	 */
	public boolean getCoverageAlgorithmShouldRunFlag() {
		return shouldRunDijkstra;
	}
	
	/**
	 * Set the map that will be used by dijkstra's algorithm
	 * @param myMap
	 * @param occupiedPoints Collection of points which are occupied in
	 * external units.
	 * and cannot be driven through.
	 */
	public void setDijkstraMap(ProbMap myMap, Collection<Coord> occupiedPointsExternal) {
		//if occupiedPointsExternal is null, then create an empty list to
		//satisfy the for each loop
		if(occupiedPointsExternal == null) {
			occupiedPointsExternal= new ArrayList<Coord>();
		}
		this.myDijkstraMap = myMap;
		for(Coord c : occupiedPointsExternal) {
			Coord cInternal= c.scaleCoordDown(this.myDijkstraMap.getCellConversion());
			if(this.myDijkstraMap.cellInMap(c)) {
				this.myDijkstraMap.getCell(cInternal).setOccupied(true);
			}
		}
	}
	
	/**
	 * Get the map that will be used by dijkstra's algorithm
	 * @return
	 */
	public ProbMap getDijkstraMap() {
		return myDijkstraMap;
	}
	
	/**
	 * Set the should try to place flag
	 * 
	 * Setting this to true means the coverage algorithm will try to place this child next iteration
	 * @param shouldTryToPlaceFlag
	 */
	public void setShouldTryToPlaceFlag(boolean shouldTryToPlaceFlag) {
		this.shouldTryToPlaceFlag = shouldTryToPlaceFlag;
	}
	
	/**
	 * This method returns whether or not the coverage algorithm should try to place the child
	 * @return true if it should try to place the child
	 */
	public boolean shouldTryToPlace() {
		return shouldTryToPlaceFlag;
	}
	
	/**
	 * Sets the minimum scan radius of the child
	 * @param myMinScanRadius
	 */
	public void setMinScanRadius(int myMinScanRadius) {
		this.myMinScanRadius = myMinScanRadius;
	}
	
	/**
	 * Gets the min scan radius of the child
	 * @return
	 */
	public int getMinScanRadius() {
		return myMinScanRadius;
	}
	
	/**
	 * Sets the max scan radius of the child
	 * @param myMaxScanRadius
	 */
	public void setMaxScanRadius(int myMaxScanRadius) {
		this.myMaxScanRadius = myMaxScanRadius;
	}
	
	/**
	 * Gets the max scan radius of the child
	 * @return
	 */
	public int getMaxScanRadius() {
		return myMaxScanRadius;
	}
	
	/**
	 * Set the scan radius this child should use next time it runs
	 * @param scanRadiusToUse
	 */
	public void setScanRadius(int scanRadiusToUse) {
		this.scanRadiusToUse = scanRadiusToUse;
	}
	
	/**
	 * Get the scan radius this child should use
	 * @return
	 */
	public int getScanRadius() {
		return scanRadiusToUse;
	}
	
	/**
	 * Set if this child requires a map before scanning.
	 */
	public void setRequiresMap(boolean b) {
		this.requiresMap= b;
	}
	
	/**
	 * Get if this child requires a map before scanning.
	 */
	public boolean getRequiresMap() {
		return this.requiresMap;
	}
	
	/**
	 * Get the path for the robot to take
	 * 
	 * @return the dijkstraFinalPath
	 */
	public ArrayList<Coord> getFinalPath() {
		return dijkstraFinalPath;
	}
	
	public void setFinalPath(ArrayList<Coord> newPath){
		this.dijkstraFinalPath = newPath;
	}

	/**
	 * Set whether or not the child is busy
	 * @param busy
	 */
	public void setBusy(boolean busy) {
		this.busy = busy;
	}

	/**
	 * Find out whether or not the child is busy
	 * @return
	 */
	public boolean isBusy() {
		return busy;
	}

	/**
	 * Get the ID of the child
	 * @return
	 */
	public int getID() {
		return this.ID;
	}
	
	/**
	 * Set the ID of the child
	 * @return
	 */
	public void setID(int newID) {
		this.ID = newID;
	}

	/**
	 * This method is called after set Final Path in order to add a path to where the robot was already going to travel
	 * @param pathToAdd
	 */
	public void addToPath(ArrayList<Coord> pathToAdd) {
		
		for (Coord point : pathToAdd){
			this.dijkstraFinalPath.add(point);
		}
		
	}

	/**
	 * Removes a given number of points from the final path of a child
	 * @param pointsToDelete
	 */
	public void removePointsFromEnd(int pointsToDelete) {
		int index = this.dijkstraFinalPath.size() - 1;
		
		for (int i = 0; i < pointsToDelete; i++){
			this.dijkstraFinalPath.remove(index);
			index--;
		}
		
	}
	
	/**
	 * Check if the final path of this child is valid. (length > 0)
	 * @return false if the path is not valid. true otherwise
	 */
	public boolean isPathValid() {
		if(this.dijkstraFinalPath.size() > 0) {
			return true;
		}
		else {
			return false;
		}
	}

	public void setWave(int wave) {
		this.wave = wave;
	}

	public int getWave() {
		return wave;
	}

	public void setAllPointsInPath(ArrayList<Coord> allPointsInPath) {
		this.allPointsInPathList = allPointsInPath;
	}

	public ArrayList<Coord> getAllPointsInPath() {
		return allPointsInPathList;
	}
	
	/**
	 * Get all points that the robot drives through.
	 * This differers from getAllPointsInPath because this will return all
	 * points that any part of the body of the robot passes through. rather
	 * than simply the list of points that the ideal robot with zero
	 * physical size would pass through.
	 * @return A set of all points which any part of the robot will
	 * pass through.
	 */
	public Set<Coord> getAllPointsRobotDrivesThru() {
		Set<Coord> retVal= new CopyOnWriteArraySet<Coord>();
		//for each cell in the path of the robot
		for(Coord c : this.allPointsInPathList) {
			retVal.addAll(this.myDijkstraMap.getRobotBodyCells(c));
		}
		return retVal;
	}

	public void setPathListMade(boolean b) {
		this.pathListMade = b;
	}
	
	public boolean isPathListMade(){
		return this.pathListMade;
	}

	
	/**
	 * Set whether or not the robot will have to move
	 * @param move
	 */
	public void setMove(boolean move) {
		this.move = move;
	}

	/**
	 * Get whether or not the robot will have to move
	 * 
	 * True if the robot has a path and needs to be moved
	 * @return
	 */
	public boolean isMove() {
		return move;
	}

	
	/**
	 * Set whether or not the robot will scan
	 * @param scan
	 */
	public void setScan(boolean scan) {
		this.scan = scan;
	}

	/**
	 * Get whether or not the robot will scan
	 * 
	 * True if the robot will perform a scan
	 * @return
	 */
	public boolean isScan() {
		return scan;
	}


}
