package workerCode;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

import dijkstra.DijkstraAlgorithm;

public class CoverageAlgorithm {
	
	private ProbMap myMap = null; //The map the coverage algorithm will use
	private ArrayList<Child> childList = new ArrayList<Child>(); //The list of children that may be placed
	private int internalMinScanRadius = 1; //The minimum scan radius the CA can go down to
	private int internalMaxScanRadius = 8; //The max scan radius the CA should start at 
	private int currentScanRadius = 8; //The current scan radius being used
	
	private DijkstraAlgorithm myDijk; //The Dijkstra's algorithm the CA will use
	private PathPlanner myPathPlanner = new PathPlanner();
	
	private ArrayList<Coord> completeValidList = new ArrayList<Coord>(); //Complete list of valid points in the map
	private ArrayList<Coord> tempPlaceList = new ArrayList<Coord>(); //Temp list of locations of robots
	
	//Outputs of the coverage algorithm
	public static final int WORK_TO_DO = 1000;
	public static final int NO_MORE_VALID_SPACES = 1001;
	public static final int UNREACHABLE_SPACES = 1002;
	
	//Used for value determination
	public static int DISTANCECONSTANT = 3;
	public static int UNCOVEREDCONSTANT = 1;
	
	
	//Location of the parent
	private Coord parentLocation = new Coord();
	
	/**
	 * This method will run the coverage algorithm on a given list of children
	 * 
	 * The coverage algorithm will try to optimally place the given children on the map.
	 * 
	 * It will start with their max scan radius and determine if it can place all the children successfully at that radius.
	 * If it cannot, it will begin decreasing the scan radius until it reaches the minimum, at which point it will begin
	 * removing robots.  It will try to remove the robot that is closet to the others in order to avoid having to remove
	 * more than one robot. 
	 *
	 * @param mapToRun the map to run on
	 * @param newChildList the list of children to try and place
	 * @param radius the radius to place them within
	 * 
	 * @return An op code corresponding to the success or failure of the run
	 */
	public int run(Coord parentLocation, ProbMap mapToRun, ArrayList<Child> newChildList, double radius){
		this.myMap = mapToRun;
		this.childList = newChildList;
		this.myDijk = new DijkstraAlgorithm();
		if (parentLocation.isValid()){
			this.parentLocation = parentLocation;
			this.parentLocation = this.parentLocation.scaleCoordDown(this.myMap.getCellConversion());
		}
		
		//Make a valid list of points in the map
		this.makeCompleteValidList();
			//If this list is empty, return map scanned
			if(completeValidList.size() == 0){
				return CoverageAlgorithm.NO_MORE_VALID_SPACES; //Return map scanned
			}
		
		this.convertToInternalUnits(); //Converts the appropriate fields to internal units
		this.setMinAndMaxScanRadius(); //Sets up the correct scan radius' for the CA
		this.setChildrenToRun(); //Sets all the placable flags to true in case they were false
			
		
		//While there are still robots to place
		while(this.arePlaceableRobots()){
			
			//Clear the temp list since this is a fresh run
			this.tempPlaceList.clear();
			
			//Sets all the relevant information to false before we try to run again
			this.resetChildrenForNewRun();
			
			//For Each Robot in the childList
			for (int robot = 0; robot < this.childList.size(); robot++){
				if(this.childList.get(robot).shouldTryToPlace()){
					//generate a list of all children besides
					ArrayList<Child> otherChildren= new ArrayList<Child>(this.childList);
					otherChildren.remove(robot);
					//create list of other children's locations in external units
					//this list will include the centerpoint location as well as all other
					//cells occupied by the physical body of the child
					ArrayList<Coord> otherChildrenLocations= new ArrayList<Coord>();
					for(Child child : otherChildren) {
						//add all points in the robot's previous scan radius as "locations" of the children
						//so that otherChildrenLocatinos also will contain the location of your children's children
						otherChildrenLocations.addAll(this.myMap.returnPointsInBubble(child.getCurrentLocation().scaleCoordUp(this.myMap.getCellConversion()), child.getScanRadius()));
					}
				
					//Check if we have done this before ... if no ...
					if (this.childList.get(robot).getCoverageAlgorithmShouldRunFlag()){
						
						//TODO P1 This is the method that should take in the list of points to avoid on the map so that they were there before Dijkstra's runs in the next line
						this.childList.get(robot).setDijkstraMap(new ProbMap(this.myMap), otherChildrenLocations);
				
						//TODO P1 Before running this line update the child's dijkstra map with robot occupied location
						//Run Dijkstras on that robot
						this.myDijk.run(this.childList.get(robot).getDijkstraMap(), this.childList.get(robot).getCurrentLocation());
					
						//Make a list of drivable valid points (in the child)
						this.makeRobotDrivableList(this.childList.get(robot));
				
						//Set the flag that we have done this before
						this.childList.get(robot).setShouldRunFlag(false);
					}
				
					//Make a new list of all points that are far enough from other robots who have already been placed on the map
					this.makeRobotPlaceableList(this.childList.get(robot), this.currentScanRadius);
					
					
					//If there are no points in the above list, 
					if(this.childList.get(robot).getPlaceableList().size() == 0){
						robot = this.childList.size(); //Break out of the For Each Robot For Loop
					}
		
					//Else
					else{
						//Get a value of each point in your list based off the scan radius
						//Pick the best points and make this your next location
						this.placeRobotAtBestPoint(this.childList.get(robot), this.currentScanRadius);
					}
			
				}	
			}//End of the For Each Robot Code 
			
			if (this.tempPlaceList.size() == this.numberOfRobotsTryingToPlace()){
				
				this.setChildNextLocations(); //THIS MUST BE CALLED FIRST
				this.moveIdleRobots(); //TODO uncomment this line
				this.convertToExternalUnits();
				
				for (Coord c : this.tempPlaceList){
					
					if (c.equals(this.parentLocation)){
						System.out.println("ERROR: Child at Same Location As Parent");
					}
				}
				
				return CoverageAlgorithm.WORK_TO_DO; //Return success
			}
	
			//If the current scan radius is a minmium
			if (this.currentScanRadius == this.internalMinScanRadius){
				System.out.flush();
				this.removeClosestRobot(); //Remove the closest robot
			}
			//Otherwise
			else {
				this.decrementScanRadius(); //Decrease the scan radius
			}
				
				
		}//End of the while robots to place loop
		
		//Return, nowhere left to place
		this.convertToExternalUnits();
		return CoverageAlgorithm.UNREACHABLE_SPACES;
	}

	/**
	 * Sets the min and max scan radius for the CA to use based on the children given
	 */
	private void setMinAndMaxScanRadius() {
		//If we were passed children
		if (!this.childList.isEmpty()){
			
			this.internalMaxScanRadius = this.childList.get(0).getMaxScanRadius() / this.myMap.getCellConversion();
			if (this.internalMaxScanRadius > this.myMap.getxSize()){
				this.internalMaxScanRadius = this.myMap.getxSize();
			}
			this.internalMinScanRadius = this.childList.get(0).getMinScanRadius() / this.myMap.getCellConversion();
			this.currentScanRadius = this.internalMaxScanRadius;
		}
		
	}

	/**
	 * Decrements the scan radius until a minimum value is reached
	 */
	private void decrementScanRadius() {
		int tempRadius;
		tempRadius = (int)(this.currentScanRadius - ((double)this.currentScanRadius / 10));
		
		//If the temp radius is greater than or equal to the minimum
		if (tempRadius >= this.internalMinScanRadius){
			this.currentScanRadius = tempRadius;
		}
		
		//Else (it was less than the minimum) then set the current to be the minimum
		else {
			this.currentScanRadius = this.internalMinScanRadius;
		}
		
		
	}

	/**
	 * Sets the children to run, used at the beginning of the CA
	 */
	private void setChildrenToRun() {
		//For Each Robot
		for (int robot = 0; robot < this.childList.size(); robot++){
			//Set the run dijkstra's flag to true
			this.childList.get(robot).setShouldRunFlag(true);
			
			//Set the should try to place flag to true
			this.childList.get(robot).setShouldTryToPlaceFlag(true);
			
			//Clear the old dijkstra List
			this.childList.get(robot).getFinalPath().clear();
		}
	}

	/**
	 * Resets the children in the list at the beginning of a new run of the CA
	 */
	private void resetChildrenForNewRun() {
		//For each robot	
		for (Child c : this.childList){
			//Set the move to be false
			c.setMove(false);
			//Set the scan to be false
			c.setScan(false);
			//Clear the next location if it was set
			c.setNextLocation(new Coord());
			//Clear the next possible location
			c.setPossibleNextLocation(new Coord());
			//Clear the Placeable List
			c.getPlaceableList().clear();
		}
	}

	/**
	 * This method is called before the class finishes running in order to set the next location for the robot to
	 * move to to be the same as the "temp" (now final) one that the algorithm found while running.
	 */
	private void setChildNextLocations() {
		
		//For each robot
		for (Child c : this.childList){
			
			//Set the next possible location to be the actual next location
			c.setNextLocation(c.getPossibleNextLocation().deepCopy());
			
			//Set that the child had to move and scan
			if (c.shouldTryToPlace()){
				c.setMove(true);
				c.setScan(true);
			}
		}
		
	}

	/**
	 * This method will move any idea robot away from the others so that they can scan their area
	 */
	private void moveIdleRobots() {
		// TODO Make the CA Move idle robots
		CopyOnWriteArraySet<Coord> unavailablePoints = new CopyOnWriteArraySet<Coord>();
		ArrayList<Child> placedByCA = new ArrayList<Child>();
		ArrayList<Child> idleRobots = new ArrayList<Child>();
		ArrayList<Coord> cannotDriveThruPoints= new ArrayList<Coord>();
		
		//Used by the idle robots
		int waveNumber = 1;
		
		placedByCA = this.getRobotsPlacedByCA();
		idleRobots = this.getIdleRobots();
		
		this.myPathPlanner.makeAndOptimizePaths(placedByCA, this.myMap);
		
		//If any of the robots were idle
		if (idleRobots.size() != 0){
			
			//For each robot placed by the CA
			for (Child c : placedByCA){
				//Add all the points within the child's scan radius to the list of points an idle robot cannot go
				unavailablePoints.addAll(this.myMap.expandArea(this.myMap.findVisiblePointsInBubble(c.getNextLocation().scaleCoordUp(this.myMap.getCellConversion()), 
																			  c.getScanRadius())));
		
				//Find all point in the paths used by the children
				myPathPlanner.getListPathAndRobotPoints(c, this.myMap);
				
				//Add the points in those paths to the locations that the idle robots cannot go
				unavailablePoints.addAll(c.getAllPointsRobotDrivesThru());
				//add child's current and endpoint locations to list of points we cannot drive through
				cannotDriveThruPoints.addAll(this.myMap.getRobotBodyCells(c.getCurrentLocation()));
				cannotDriveThruPoints.addAll(this.myMap.getRobotBodyCells(c.getNextLocation()));
			}
			
			//For each child that is idle
			for (Child c : idleRobots){
				//If the child needs to move from its current location
				if (this.childNeedsToBeMoved(c, unavailablePoints)){
					//Find a point for the robot to move to
					c.setNextLocation(this.placeAwayFromGivenList(c.getCurrentLocation(), unavailablePoints));
					
					//add new undrivable points to c's dijkstraMap
					c.setDijkstraMap(c.getDijkstraMap(), cannotDriveThruPoints);
					this.myDijk.run(c.getDijkstraMap(), c.getCurrentLocation());
					//Set the path for the robot
					this.myPathPlanner.makeAndOptimizePath(c, this.myMap);
					
					//Populate all the points that are in the child's path
					myPathPlanner.getListPathAndRobotPoints(c, this.myMap);
					
					
					//Set this child's wave to be the first / 2nd / nth before the robots moved by the CA
					c.setWave(waveNumber);
					
					//Increment the wave number
					waveNumber++;
					
					//Set that the child moved
					c.setMove(true); //NEEDS TO BE TRUE BEFORE WAVES IS CALLED
					
					//Add this robot to the list of robots moved by the CA, since it has now been moved
					placedByCA.add(c);
					//add where the child moved to the list of points we can't drive through
					//for the next child in the idleRobots list
					cannotDriveThruPoints.addAll(this.myMap.getRobotBodyCells(c.getNextLocation()));
					
				} //End of if the child needs to move				
			} //End of for each idle robot
		} //End of if there are idle robots
		
		//Assign waves to each robot
		this.myPathPlanner.assignWaves(placedByCA, this.myMap);
	}
	
	private ArrayList<Child> getIdleRobots() {
		ArrayList<Child>  tempList = new ArrayList<Child>();
		
		//For each robot in the child list
		for (Child c : this.childList){
			if (!c.shouldTryToPlace()){
				tempList.add(c);
			}
		}
		
		return tempList;
	}

	private ArrayList<Child> getRobotsPlacedByCA() {
		ArrayList<Child>  tempList = new ArrayList<Child>();
		
		//For each robot in the child list
		for (Child c : this.childList){
			if (c.shouldTryToPlace()){
				tempList.add(c);
			}
		}
		
		return tempList;
	}

	/**
	 * Determines if the child needs to be moved based on the current 
	 * state of the other children.
	 * 
	 * @param c child in question
	 * @return
	 */
	private boolean childNeedsToBeMoved(Child child, Collection<Coord> unavailablePoints) {
		CopyOnWriteArraySet<Coord> tempList = new CopyOnWriteArraySet<Coord>();
		for(Coord c : unavailablePoints) {
			tempList.add(c.scaleCoordDown(this.myMap.getCellConversion()));
		}
		if(tempList.contains(child.getCurrentLocation())){
			return true;
		}
		return false;
	}
	
	private Coord placeAwayFromGivenList(Coord currentLocation, Set<Coord> unavailablePoints) {
		int sStep = 1;
		int rLimit = Math.max(this.myMap.getxSize(), this.myMap.getySize());
		int currX = currentLocation.getX();
		int currY = currentLocation.getY();
		int x = currX;
		int y = currY;
		Coord tempCoord = new Coord();
		CopyOnWriteArraySet<Coord> tempList = new CopyOnWriteArraySet<Coord>();
		for(Coord c : unavailablePoints) {
			tempList.add(c.scaleCoordDown(this.myMap.getCellConversion()));
		}
		
		//Spiral outwards for 'nearest' point allowed to place at
		while(sStep < rLimit) {
			//Start traversing x
			while(x < currX + sStep) {
				x++;
				tempCoord = new Coord(x,y);
				//If cell is in range of the parent's map
				if(this.myMap.cellInMap(tempCoord.scaleCoordUp(this.myMap.getCellConversion()))){
					//And is drivable
					if(this.myMap.cellAreaIsDrivable(tempCoord)){
						//Is it allowed?
						if(!tempList.contains(tempCoord)) {
							return tempCoord;
						}
					}
				}
			}

			//Start traversing y
			while(y < currX + sStep) {
				y++;
				tempCoord = new Coord(x,y);
				//If cell is in range of the parent's map
				if(this.myMap.cellInMap(tempCoord.scaleCoordUp(this.myMap.getCellConversion()))){
					//And is drivable
					if(this.myMap.cellAreaIsDrivable(tempCoord)){
						//Is it allowed?
						if(!tempList.contains(tempCoord)) {
							return tempCoord;
						}
					}
				}
			}

			//Start traversing x left
			while(x > currX - sStep) {
				x--;
				tempCoord = new Coord(x,y);
				//If cell is in range of the parent's map
				if(this.myMap.cellInMap(tempCoord.scaleCoordUp(this.myMap.getCellConversion()))){
					//And is drivable
					if(this.myMap.cellAreaIsDrivable(tempCoord)){
						//Is it allowed?
						if(!tempList.contains(tempCoord)) {
							return tempCoord;
						}
					}
				}
			}

			//Start traversing y down
			while(y > currX - sStep) {
				y--;
				tempCoord = new Coord(x,y);
				//If cell is in range of the parent's map
				if(this.myMap.cellInMap(tempCoord.scaleCoordUp(this.myMap.getCellConversion()))){
					//And is drivable
					if(this.myMap.cellAreaIsDrivable(tempCoord)){
						//Is it allowed?
						if(!tempList.contains(tempCoord)) {
							return tempCoord;
						}
					}
				}
			}
			sStep += 1;
		}
		
		// Failure returns invalid Coord
		return new Coord();
	}

	/**
	 * This robot will compute which robot is closet to the others, and then 
	 * tell that robot not to run in the next iteration
	 */
	private void removeClosestRobot() {
		int closestRobot=-1;
		int tempDistance=-1;
		int closestDistance=-1;
		
		//For each robot
		for (int robot = 0; robot < this.childList.size(); robot++){
			if(this.childList.get(robot).shouldTryToPlace()){
			
				//For each robot to check
				for(int robotCheck = 0; robotCheck < this.childList.size(); robotCheck++){
					if (robot != robotCheck){
						
						//Sum the distances between this robot and every other robot
						tempDistance += this.myMap.distance(this.childList.get(robot).getCurrentLocation(),
															this.childList.get(robotCheck).getCurrentLocation());
					}
				}
				
				//If the sum of the distance is less than the current closest set of distances
				if (tempDistance < closestDistance || closestDistance == -1){
					closestRobot = robot;
					closestDistance = tempDistance;
				}
			}
			
		}
		
		//Tell the robot that is closest to all of the others to not try and place this time
		this.childList.get(closestRobot).setShouldTryToPlaceFlag(false);		
	}

	/**
	 * Converts everything going outside the CA to external units based on the maps scaling factor
	 */
	private void convertToExternalUnits() {

//		for (int robot = 0; robot < this.childList.size(); robot++){
			
		for (Child c : this.childList){
				
			//Convert the current location to external units
			Coord tempCoord1 = c.getCurrentLocation().deepCopy();
			tempCoord1.setX((tempCoord1.getX() * this.myMap.getCellConversion()) + (this.myMap.getCellConversion() / 2) );
			tempCoord1.setY((tempCoord1.getY() * this.myMap.getCellConversion()) + (this.myMap.getCellConversion() / 2) );
			c.setCurrentLocation(tempCoord1.deepCopy());
			

				
			//Convert the next location to external units
			Coord tempCoord2 = c.getNextLocation().deepCopy();
			tempCoord2.setX((tempCoord2.getX() * this.myMap.getCellConversion()) + (this.myMap.getCellConversion() / 2));
			tempCoord2.setY((tempCoord2.getY() * this.myMap.getCellConversion()) + (this.myMap.getCellConversion() / 2));
			c.setNextLocation(tempCoord2.deepCopy());
			
			//Convert the Radius to external units
			c.setScanRadius(this.currentScanRadius * this.myMap.getCellConversion());

						
		}
		
	}

	/**
	 * Sets the necessary starting information to internal units so that it can be used
	 */
	private void convertToInternalUnits() {
		for(int robot = 0; robot < this.childList.size(); robot++){
			
			//Set the robots current position to internal units
			Coord tempCord = this.childList.get(robot).getCurrentLocation().deepCopy();
			tempCord.setX(tempCord.getX() / this.myMap.getCellConversion());
			tempCord.setY(tempCord.getY() / this.myMap.getCellConversion());
			this.childList.get(robot).setCurrentLocation(tempCord.deepCopy());
						
		}
	}

	/**
	 * This method will return the number of robots we are currently trying to place with the coverage algorithm
	 * 
	 * @return int the number of robots we are trying to place
	 */
	private int numberOfRobotsTryingToPlace() {
		int numRobots = 0;
		
		//For each robot in the child list
		for(int robot = 0; robot < this.childList.size(); robot++){
			
			//Check if we should try to place this robot
			if (childList.get(robot).shouldTryToPlace()){
				numRobots++; //Increment the count
			}
		}
		
		return numRobots; //Return the count
	}

	/**
	 * This method finds the best place for the robot and then sets its next possible location to that point
	 * 
	 * @param child
	 * @param scanRadius
	 */
	private void placeRobotAtBestPoint(Child child, int scanRadius) {
		Coord tempCoord = new Coord();
		double currentBest = Double.NEGATIVE_INFINITY;
		double tempValue = 0;
		
		//For each point in the child's temp list
		for (int locPlaceableList = 0; locPlaceableList < child.getPlaceableList().size(); locPlaceableList++){
		
			//Compute the value of that cell
			tempValue = computeCellValue(this.myMap.computeCellsUncovered(child.getPlaceableList().get(locPlaceableList), 
			                 	   		 scanRadius), 
			                 	   		 child.getDijkstraMap().cellGetDistance(child.getPlaceableList().get(locPlaceableList)));
			
			//If its greater than the current best cell
			if (tempValue > currentBest){
				currentBest = tempValue; //Make the current best the same as temp
				tempCoord = child.getPlaceableList().get(locPlaceableList).deepCopy();//Make that point the temp best Coord
			}
		
		}
		
		//Set the child's next possible location to the best coord found
		child.setPossibleNextLocation(tempCoord.deepCopy());
		
		//Add that point to the temp placed list
		this.tempPlaceList.add(child.getPossibleNextLocation().deepCopy());
		
	}

	/**
	 * Makes a private list of points that the robot can actually be placed on the map
	 * 
	 * @param child
	 */
	private void makeRobotPlaceableList(Child child, int scanRadius) {
		
		//TODO PRIORITY 2 make this method also check if there is a barrier between two robots, if there is then they can be placed at that location
		
		boolean wasFarEnough = true;
		
		child.getPlaceableList().clear();

		//For each point in the drivable list
		for (int locDriveList = 0; locDriveList < child.getDrivableList().size(); locDriveList++){
			
			//Assume they were far enough apart
			wasFarEnough = true;
			
			//For each point in the temp list
			for (int locTempList = 0; locTempList < this.tempPlaceList.size(); locTempList++){
				
				//Check if the points are too close
				if (this.myMap.distanceLessThanOrEqual(child.getDrivableList().get(locDriveList), this.tempPlaceList.get(locTempList), scanRadius * 2)){
					wasFarEnough = false;
				}
				
			}
			
			//If they were far enough, add them to the child's placeable list
			if(wasFarEnough && !this.parentTooClose(child.getDrivableList().get(locDriveList))){
				//Add the point from the valid list into the placeable list
				child.getPlaceableList().add(child.getDrivableList().get(locDriveList).deepCopy());
			}
		}
		
	}

	/**
	 * This method checks to make sure that its parent is outside the current scan radius of the robot
	 * 
	 * If the parent is not outside the current scan radius then it will not place the robot at that point
	 * @param possibleLocation 
	 * @return
	 */
	private boolean parentTooClose(Coord possibleLocation) {
		
		//If the parent is on the map
		if (parentLocation.isValid()){
		
			//If the parent is less than or equal to your scan radius away
			if (this.myMap.distanceLessThanOrEqual(this.parentLocation, possibleLocation, this.currentScanRadius)){
				return true; // return true, they were too close
			}
			return false;
		}
		
		return false; // return false if the parents location is invalid, since they are not on the map you cannot be too close
	}
	
	
	/**
	 * Make the child's private list of valid, drivable robots
	 * 
	 * @param child
	 */
	private void makeRobotDrivableList(Child child) {
		
		//Clear the Driveable List
		child.getDrivableList().clear();
		
		//For each point in the valid list
		for (int locationInValidList = 0; locationInValidList < this.completeValidList.size(); locationInValidList++){
			//Check if that point is drivable for the robot
			if(child.getDijkstraMap().cellIsReachable(this.completeValidList.get(locationInValidList))){
				child.getDrivableList().add(this.completeValidList.get(locationInValidList).deepCopy()); //Add it to the child's list
			}
		}
		
	}

	/**
	 * Returns if there are robots that are placeable by the coverage algorithm
	 * 
	 * @return
	 */
	private boolean arePlaceableRobots() {
		//For each robot in the child list
		for (int robot = 0; robot < this.childList.size(); robot++){
			//Check if we should try to place that robot
			if(this.childList.get(robot).shouldTryToPlace()){
				return true; //Return true if even one robot is placeable
			}
		}
		return false; //Otherwise Return false
	}

	
	/**
	 * Makes the complete list of valid points in the map
	 */
	private void makeCompleteValidList() {
		
		this.completeValidList.clear();
		
		//For each cell in the map
		for (int x = myMap.getXBoundMin(); x < myMap.getXBoundMax(); x++){
			for(int y = myMap.getYBoundMin(); y < myMap.getYBoundMax(); y++){
				
				//Make a temp Coordinate
				Coord tempCoord = new Coord(x,y);
				
				//If that coordiante in the map is valid
				if (this.myMap.isCellValid(tempCoord)){
					
					this.completeValidList.add(tempCoord.deepCopy());//Add it to the list of all valid points
				}
			}
		}
		
	}
	
	/**
	 * This method returns the value of a given location to drive to
	 * 
	 * This method uses two constants defined at the top of this file to compute the value.
	 * It does the cells times a constant - the distance time a constant, which basically is
	 * a way of weighting "For every cell you drive you must uncover this many more cells.
	 * 
	 * @param cellsUncovered
	 * @param cellDistance
	 * @return
	 */
	private double computeCellValue(int cellsUncovered, int cellDistance) {
		
		return (cellsUncovered * CoverageAlgorithm.UNCOVEREDCONSTANT)
				-(cellDistance * CoverageAlgorithm.DISTANCECONSTANT);
	}
	
}
