package workerCode;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;
import java.util.Vector;

import communication.*;
import dijkstra.DijkstraAlgorithm;

public abstract class Robot implements Runnable, Observer {
	protected Vector<SwarmPacket> commandQ;
	protected int numChildren;
	protected Map<Integer, Child> childList;
	protected Locomotion robotLocomotion;
	protected Comm robotComm;
	protected CoverageAlgorithm robotCoverageAlgorithm;
	protected ProbMap robotMap;
	protected int scanRadiusMin;
	protected int scanRadiusMax;
	private boolean isRunning;
	private DijkstraAlgorithm myDijk = new DijkstraAlgorithm();
	private PathPlanner myPathPlanner = new PathPlanner();	

	public Robot(int scanMax, int scanMin, Comm robotComm, int numChildren, Locomotion robotLocomotion) {
		this.scanRadiusMax= scanMax;
		this.scanRadiusMin= scanMin;
		this.robotComm= robotComm;
		this.numChildren= numChildren;
		this.robotLocomotion= robotLocomotion;
		this.childList= new HashMap<Integer, Child>(numChildren);
		this.commandQ= new Vector<SwarmPacket>();
		this.robotComm.addObserver(this);
		this.isRunning= true;
	}
	//Command Thread methonds
	//should run always be overridden by implementing class?
	/**
	 * The run method for the main command thread of the robot.  This method
	 * assumes that all commands are placed into a command queue and this
	 * thread is notified of the placement.
	 */
	@Override
	synchronized public void run() {
		//wait for children to start up
		this.waitForChildren();
		this.allChildrenReady();
		
//		System.out.println(this + " has " + this.childList.size() + " children");
		
		while(this.isRunning) {
			//while there are no commands to execute, sleep
			while(commandQ.isEmpty()) {
				try {
					this.wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			
			//pass command to command handler
			SwarmPacket command= this.commandQ.remove(0);
			this.commandHandler(command);
		}
		this.robotComm.kill();
	}
	
	/**
	 * Handle a command packet.  If the packet is a downstream
	 * packet the ID of the sender must be the parent so there is
	 * no need to include it again here.  This method is responsible for
	 * determining the type of command and passing the message along to the
	 * appropriate handler.
	 * @param command The command packet to process.
	 */
	private void commandHandler(SwarmPacket command) {
		if(command instanceof PathPacket) {
			this.pathCommandHandler((PathPacket) command);
		}
		else if(command instanceof ScanPacket) {
			this.scanCommandHandler((ScanPacket)command);
		}
		else if(command instanceof MapPacket) {
			this.mapCommandHandler((MapPacket)command);
		}
		else if(command instanceof KillPacket) {
			this.killCommandHandler((KillPacket)command);
		}
		else {
			return;
		}
	}
	
	//Command handlers
	/**
	 * Handle a scan command.  This function is called whenever a scan command
	 * needs to be processed.  This function must either divide the work among
	 * the children or perfrom the scan itself.
	 */
	protected void scanCommandHandler(ScanPacket command) {
		this.scan(command.getRadius());
	}
	
	/**
	 * This abstract method must be overridden by each robot based on how
	 * they scan.  This method is responsible for sending scan results back
	 * to the parent.
	 * @param radius The radius to scan.
	 */
	protected abstract void scan(int radius);
	
	/**
	 * Handle a path command.  This function is called whenever a packet
	 * command needs to be processed.  This function is responsible for moving
	 * both the robot itself as well as any children to keep them contained
	 * inside the minimum bubble size.
	 */
	protected void pathCommandHandler(PathPacket command) {
		Coord tempParentLastLocation = this.robotLocomotion.getCurrentLocation();
		this.robotLocomotion.drivePath(command.getPath());
		
		//Move the children with the parent to the next location
		this.moveChildrenWithMe(command.getPath(), tempParentLastLocation);
		
		this.robotComm.send(Comm.PARENT_ID, new ReachedDestinationPacket(this.robotLocomotion.getCurrentLocation()));
	}
	
	private void moveChildrenWithMe(ArrayList<Coord> parentPath, Coord tempParentLastLocation) {
//		System.out.println("moving children with me");
		ArrayList<Child> tempChildList = new ArrayList<Child>();
		ArrayList<Coord> cannotDriveThruPoints= new ArrayList<Coord>();
		int distanceIncrement = 30; //The distance to stop from the worker in cm (will be incremented later)
		int backoffDistance= distanceIncrement;
		
		//First Dijkstra's needs to be run on the child so it can determine how to travel from where
		//It is now to the current location of the worker
//		System.out.println("The parent is at: " + this.robotLocomotion.getCurrentLocation());
		for (Child c : this.childList.values()){
			if(this.robotMap.distanceLessThan(parentPath.get(parentPath.size() - 1), c.getCurrentLocation(), this.scanRadiusMin)) {
//				System.out.println("not moving child!");
				c.setMove(false);
				continue;
			}
			else {
//				System.out.println("parent at "+parentPath.get(parentPath.size() - 1)+"child at "+c.getCurrentLocation()+"min radius is "+this.scanRadiusMin);
			}
			//Set the first point that the child wants to get to as the workers current location
			c.setNextLocation(tempParentLastLocation.scaleCoordDown(this.robotMap.getCellConversion())); 
			c.setDijkstraMap(new ProbMap(this.robotMap), cannotDriveThruPoints);
			
			//TODO P1 Add in a c.setDijkstra's Map to take in the now changed robot positions you want dijkstra to path around
			
			
			//TODO P1 the robot positions must be set in the map before this line is run
			
//			System.out.println(c + "'s unavailable points list: ");
			for(Coord coord : cannotDriveThruPoints) {
//				System.out.println(this + ": " + coord);
			}
//			System.out.println("===============================\n\n");
//			System.out.println(Thread.currentThread().getName()+" map limits "+this.robotMap.getXBoundMin()+" minX "+this.robotMap.getXBoundMax()+" maxX "+this.robotMap.getYBoundMin()+" minyY "+this.robotMap.getYBoundMax()+ " maxY");
//			this.myDijk.run(c.getDijkstraMap(), c.getCurrentLocation().scaleCoordDown(this.robotMap.getCellConversion()));
			tempChildList.add(c);
//
//			
//			this.myPathPlanner.generateInitialPath(c);
//			for(Coord coord : c.getFinalPath()) {
//				System.out.println(this + ": " + coord);
//			}
//			System.out.println("\n------------------");
//			
//			myPathPlanner.makeAndOptimizePath(c, this.robotMap);
			ArrayList<Coord> teleportPath= new ArrayList<Coord>();
			teleportPath.add(c.getCurrentLocation().deepCopy());
			teleportPath.add(tempParentLastLocation.deepCopy());
			c.setFinalPath(teleportPath);
//			System.out.println(c.getFinalPath());
			myPathPlanner.addToPath(c.getFinalPath(), parentPath);
			
//			System.out.println("cihld final path pre backoff!");
//			System.out.println(c.getFinalPath());
			myPathPlanner.findNewEndpointDistanceFromEnd(c, backoffDistance, this.robotMap);
			if(c.getFinalPath().size()!= 0) c.setMove(true);
			backoffDistance+= distanceIncrement;
//			System.out.println("child final path post backoff");
//			System.out.println(c.getFinalPath());
			//add where the child ended up as a point the next child in the list cannot drive through
			cannotDriveThruPoints.addAll(this.robotMap.getRobotBodyCells(c.getNextLocation().scaleCoordUp(this.robotMap.getCellConversion())));
			//in order for dijkstra's to run properly the old position of the worker has
			//to not be in the list of points you can't drive through.  therefore we
			//manually remove those points here
			ArrayList<Coord> parentBodyCoords= this.robotMap.getRobotBodyCells(tempParentLastLocation);
			cannotDriveThruPoints.removeAll(parentBodyCoords);
		}
		
		//assigning waves to children
		myPathPlanner.assignWaves(tempChildList, this.robotMap);
		
		//TODO Most of these comments
		// At this point in the code, every child should be moving to the
		// parent's location and then following the parents path
		// Since they are all going to what was the parents original location,
		// each child should be in a different wave number
		// This should automatically ensure that each child will not drive
		// simultaneously since they are all going to the same point
		// forcing a higher wave number for each child.
		
		//The next step is to determine a new spot for each child to stop, and then make that the new end point to the path
		//Although in theory the first child should be wave 1 and every child should go up by one after that one, I will still
		//step through the list by wave number instead of just going through the list
		
		ArrayList<Child> childSendList = new ArrayList<Child>();
			for (int currentWave = 1; currentWave <= this.findNumWaves(); currentWave++){	
			childSendList.clear();
			for(Child c : this.childList.values()) {
				if(c.isMove() && c.getWave() == currentWave) {
//					System.out.println("I am at: " + this.robotLocomotion.getCurrentLocation() + " child at: " + c.getFinalPath().get(c.getFinalPath().size() - 1) + " also " + c.getNextLocation());
					childSendList.add(c);
				}
			}
			
			//TODO P1 Up until this point c.getCurrentLocation is where the robot is currently sitting.  After this next line in the code you dont' know if the robot has responded with a reached destination packet changing its current location
			this.sendPathToChildren(childSendList);
			childSendList.clear();
			//wait for children to be ready to scan
			this.waitForChildren();
		}
		
	}
	
	protected void mapCommandHandler(MapPacket command) {
//		System.out.println("Got a Map");
		this.robotMap= command.getMap();
	}
	
	protected void killCommandHandler(KillPacket command) {
		this.kill();
	}
	
	/**
	 * Wait until all children are ready to run.  This method will wait until
	 * all children have checked in with their information and are now ready
	 * to accept commands.
	 */
	protected void waitForChildren() {
		while(!this.checkAllChildrenReady()) {
			try {
				this.wait();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * This method is called after all children are ready.  This method should
	 * send a child information packet to the parent.
	 */
	protected abstract void allChildrenReady();
	
	//Upstream Thread methods
	/**
	 * Update callback from observed objects.  This update is called whenever
	 * any observed object notifies its observers and this function is
	 * responsible for determining who triggered the update and passing the
	 * information to the appropriate handler.
	 */
	@Override
	public void update(Observable arg0, Object arg1) {
		//If the Comm is what updated us
		if(arg0 == this.robotComm) {
			this.commCallback((SwarmCommunication)arg1);
		}
		
	}
	
	/**
	 * This function is called whenever a communication is received.  It must
	 * determine if the packet is upstream or downstream and pass to the
	 * correct handler.
	 * @param swarmComm The packet/sender id received.
	 */
	private void commCallback(SwarmCommunication swarmComm) {
		//downstream packets should be added to the queue
		if(swarmComm.getID() == Comm.PARENT_ID) {
			this.commandQ.add(swarmComm.getPacket());
			//notify the command thread
			synchronized(this) {
				this.notifyAll();
			}
		}
		else {
			//upstream packets
			this.handleUpstreamPacket(swarmComm.getID(), swarmComm.getPacket());
		}
	}
	
	/**
	 * Handle an upstream packet.  This method is called whenever a packet is
	 * received from a child.  This function determines the type of packet and
	 * passes it to the correct packet handler.
	 * @param childID The id of the child which sent this packet.
	 * @param packet The packet that was sent.
	 */
	synchronized private void handleUpstreamPacket(int childID, SwarmPacket packet) {
		//switch on packet
		if(packet instanceof ChildInformationPacket) {
			this.childInformationPacketHandler(childID, (ChildInformationPacket)packet);
		}
		
		else if(packet instanceof AddPointToMapPacket) {
			this.addPointToMapPacketHandler(childID, (AddPointToMapPacket)packet);
		}
		
		else if(packet instanceof ScanCompletedKnownShadowingPacket) {
			this.scanCompletedKnownShadowingPacketHandler(childID, (ScanCompletedKnownShadowingPacket)packet);
		}
		
		else if(packet instanceof ScanCompletedUnknownShadowingPacket) {
			this.scanCompletedUnknownShadowingPacketHandler(childID, (ScanCompletedUnknownShadowingPacket)packet);
		}
		
		else if(packet instanceof ReachedDestinationPacket) {
			this.reachedDestinationPacketHandler(childID, (ReachedDestinationPacket)packet);
		}
	}
	
	//upstream handlers
	/**
	 * Handle a child information packet.  This method is called whenever an
	 * information packet is received from a child.  This will store the
	 * information of the child so that it may be later recalled.
	 * @param childID The id of the child which sent this packet.
	 * @param packet The child information packet itself.
	 */
	protected void childInformationPacketHandler(int childID, ChildInformationPacket packet) {
		Child newChild= new Child(childID, packet.getLocation(),
				                  packet.getMinScanRadius(), packet.getMaxScanRadius(), packet.checkRequiresMap());
		this.addChild(newChild);
		newChild.setBusy(false);
		
		//notify waiting threads of this update
		this.notifyAll();
	}
	
	/**
	 * Handle adding a single point to the map.  This method will fill in
	 * the probability map based on the information contained in the packet.
	 * @param childID The id of the child who sent this packet.
	 * @param packet The packet itself.
	 */
	//TODO PRIORITY 1 Make this method abstract and change it for the queen
	protected void addPointToMapPacketHandler(int childID, AddPointToMapPacket packet) {
//		System.out.println(this + " got an AddPointToMapPacket from " + childID);
//		System.out.println(this+ ": packet data: "+packet.getLocation()+" "+packet.getStdDev());
//		if(packet.isObsticle()){
		
		Coord tempCoord =  packet.getLocation().scaleCoordDown(this.robotMap.getCellConversion());
		
		
//			System.out.println(this + " got a packet from " + childID + "\n " +
//							   "X: " + this.robotMap.getXBoundMin() + " to " + this.robotMap.getXBoundMax() + "\n " +
//							   "Y: " + this.robotMap.getYBoundMin() + " to " + this.robotMap.getYBoundMax() + "\n " +
//							   "Saw a point at: " + packet.getLocation().scaleCoordDown(this.robotMap.getCellConversion()));
		
		if (tempCoord.getX() < this.robotMap.getXBoundMin() || 
			tempCoord.getX() >= this.robotMap.getXBoundMax() ||
			tempCoord.getY() < this.robotMap.getYBoundMin() || 
			tempCoord.getY() >= this.robotMap.getYBoundMax()){}
		else{
//			System.out.println("setting cell to be blocked");
			this.robotMap.updateCellAreaBlockValue(packet.getLocation(), packet.getStdDev());
			//this.robotMap.cellSetBlockValue(packet.getLocation().scaleCoordDown(this.robotMap.getCellConversion()), 1);
		}
//		}
		
		//Send the same packet up to the parent
		if (!(this instanceof Queen)){
//			System.out.println("sending to parent");
			this.robotComm.send(Comm.PARENT_ID, packet);
		}
//		System.out.println("packet been handled!");
	}
	
	/**
	 * Handle a scan completed packet.  When a child informs a parent that the
	 * bubble has been completely scanned this packet will contain a list of
	 * cells which were unable to be scanned and any cell not in this list is
	 * assumed to have been scanned.
	 * 
	 * Worker -> Queen
	 * 
	 * @param childID The id of the child who sent this packet.
	 * @param packet The packet itself.
	 */
	protected void scanCompletedKnownShadowingPacketHandler(int childID, ScanCompletedKnownShadowingPacket packet) {
		Child sender= this.getChild(childID);
		this.robotMap.updateCompletedScanAreaUnseenSpaces(sender.getCurrentLocation(), sender.getScanRadius(), packet.getShadowedPoints());
		//set busy status to false
		sender.setBusy(false);
		//notify waiting threads of this update
		this.notifyAll();
	}
	
	/**
	 * Handle a scan complete with unknown shadowing packet.  When a child
	 * sends this packet to a parent the child is indicating that it has
	 * completed a scan and that all remaining cells cannot be scanned.
	 * However, the child is unaware of which cells could not be scanned so it
	 * is up to the parent to make an educated guess at which cells were
	 * "in shadow".
	 * 
	 * Scout -> Worker / Anything
	 * 
	 * @param childID The id of the child who sent this packet.
	 * @param packet The packet itself.
	 */
	protected void scanCompletedUnknownShadowingPacketHandler(int childID, ScanCompletedUnknownShadowingPacket packet) {
//		System.out.println(this+" got a scan completed w/unknown shadowing packet frome "+childID);
		Child sender= this.getChild(childID);
		this.robotMap.updateCompletedScanAreaUnknownShadowing(sender.getCurrentLocation(), sender.getScanRadius());
		//set busy status to false
		sender.setBusy(false);
		//notify waiting threads of this update
		this.notifyAll();
	}
	
	/**
	 * Handle a reached destination packet.  This packet informs a parent that
	 * child has arrived at the final destination and is ready to perform
	 * another task.
	 * @param childID The id of the child which sent this packet.
	 * @param packet The packet itself (this packet is empty and is included
	 *               here as a parameter to support potential future expandability).
	 */
	protected void reachedDestinationPacketHandler(int childID, ReachedDestinationPacket packet) {
		Child sender= this.getChild(childID);
		//set move status to false
		sender.setMove(false);
		//set child busy status to false
		sender.setBusy(false);
		//update location of child
		sender.setCurrentLocation(packet.getDestination());
		//notify waiting threads of this update
		this.notifyAll();
	}
	
	
	
	//accessor methods
	/**
	 * Add a child to the list of children belonging to this robot.  The child
	 * is later accessible again with it's ID.
	 */
	protected void addChild(Child c) {
		this.childList.put(new Integer(c.getID()), c);
	}
	
	/**
	 * Get a child with a given ID.
	 * @param id The id of the child.
	 * @return The child with the given id.
	 */
	protected Child getChild(int id) {
		return this.childList.get(new Integer(id));
	}
	
	/**
	 * Check the ready status of all children.
	 * @return true if all children are ready, false otherwise.
	 */
	protected boolean checkAllChildrenReady() {
		int readyCount= 0;
		//for each child in the list
		for(Child c : this.childList.values()) {
			if(!c.isBusy()) readyCount++;
		}
		
		if(readyCount == this.numChildren) {
			return true;
		}
		
		else return false;
	}
	
	
	/**
	 * Update the map with your current location and the locations of your children.
	 */
	protected void updateMapWithCurrentLocation() {
		if(this.robotLocomotion.getCurrentLocation().isValid()) {
//			this.robotMap.cellScan(this.robotLocomotion.getCurrentLocation().scaleCoordDown(this.robotMap.getCellConversion()));
			//Update the area around where the robot is
			this.updateRobotOccupiedArea(this.robotLocomotion.getCurrentLocation());
		}
		for(Child c : this.childList.values()) {
			if(c.getCurrentLocation().isValid()) {
//				this.robotMap.cellScan(c.getCurrentLocation().scaleCoordDown(this.robotMap.getCellConversion()));
				this.updateRobotOccupiedArea(c.getCurrentLocation()); 
			}
		}
	}
	
	private void updateRobotOccupiedArea(Coord currentLocation) {
		Coord scaledLocation = currentLocation.scaleCoordDown(this.robotMap.getCellConversion());
//		System.out.println("\n In Update Robot Occupied Area: \n" +
//						   "X: " + this.robotMap.getXBoundMin() + " to " + this.robotMap.getXBoundMax() + " \n" +
//						   "Y: " + this.robotMap.getYBoundMin() + " to " + this.robotMap.getYBoundMax() + " \n" +
//						   this + "'s Child: CLoc: " + currentLocation + "\n");
		
		this.robotMap.cellScan(scaledLocation);
		
		if (this.robotMap.CUpIsInTheMap(scaledLocation)){
			this.robotMap.cellScan(this.robotMap.CUp(scaledLocation));
		}
		
		if (this.robotMap.CDownIsInTheMap(scaledLocation)){
			this.robotMap.cellScan(this.robotMap.CDown(scaledLocation));
		}
		
		if (this.robotMap.CLeftIsInTheMap(scaledLocation)){
			this.robotMap.cellScan(this.robotMap.CLeft(scaledLocation));
			
			if (this.robotMap.CUpIsInTheMap(this.robotMap.CLeft(scaledLocation))){
				this.robotMap.cellScan(this.robotMap.CUp(this.robotMap.CLeft(scaledLocation)));
			}
			
			if (this.robotMap.CDownIsInTheMap(this.robotMap.CLeft(scaledLocation))){
				this.robotMap.cellScan(this.robotMap.CDown(this.robotMap.CLeft(scaledLocation)));
			}
			
		}
		
		if (this.robotMap.CRightIsInTheMap(scaledLocation)){
			this.robotMap.cellScan(this.robotMap.CRight(scaledLocation));
			
			if (this.robotMap.CUpIsInTheMap(this.robotMap.CRight(scaledLocation))){
				this.robotMap.cellScan(this.robotMap.CUp(this.robotMap.CRight(scaledLocation)));
			}
			
			if (this.robotMap.CDownIsInTheMap(this.robotMap.CRight(scaledLocation))){
				this.robotMap.cellScan(this.robotMap.CDown(this.robotMap.CRight(scaledLocation)));
			}
		}
		
		
		
	}
	/**
	 * Send the path of each child to every child.
	 */
	protected void sendPathToChildren(Collection<Child> children) {
		for(Child c : children) {
			//FIXME
			if (!c.getFinalPath().isEmpty()) {
			c.setBusy(true);
	//		System.out.println(this+" sending "+c+" to "+c.getFinalPath().get(c.getFinalPath().size() - 1)+" map limits "+this.robotMap.getXBoundMin()+" minX "+this.robotMap.getXBoundMax()+" maxX "+this.robotMap.getYBoundMin()+" minyY "+this.robotMap.getYBoundMax()+ " maxY");
			this.robotComm.send(c.getID(), new PathPacket(c.getFinalPath()));
			}
		}
	}
	
	/**
	 * Send the map to each child if they require it.
	 */
	protected void sendMapToChildren(Collection<Child> children) {
		for(Child c : children) {
			this.robotComm.send(c.getID(), new MapPacket(this.robotMap.getMapToSend(c.getCurrentLocation(),
						                                                                c.getScanRadius())));
		}
	}
	
	/**
	 * Send a scan to each child.
	 */
	protected void sendScanToChildren(Collection<Child> children) {
		for(Child c : children) {
			c.setBusy(true);
			this.robotComm.send(c.getID(), new ScanPacket(c.getScanRadius()));
		}
	}
	
	/**
	 * Kill this robot and all children of this robot.
	 */
	public void kill() {
		for(Child c : this.childList.values()) {
			this.robotComm.send(c.getID(), new KillPacket());
		}
		this.isRunning= false;
	}
	
	protected int findNumWaves(){
		int tempMaxWave = 0;
		for (Child c : this.childList.values()){
			if (c.getWave() > tempMaxWave){
				tempMaxWave = c.getWave();
			}
		}
		return tempMaxWave;
	}
	
	/**
	 * This method is specifically for updating the Simulations
	 * @return
	 */
	public ProbMap getMap() {
		//TODO PRIORITY SIMULATION Delete once the simulation holds its own map
		return this.robotMap;
	}
	public Coord getCurrentLocation() {
		return this.robotLocomotion.getCurrentLocation();
	}
}
