/*
 * command.c
 *
 *  Created on: Dec 19, 2010
 *      Author: hsrmqp
 */

#include "command.h"
#include "localization.h"
#include "locomotion.h"
#include "communications.h"
#include "includes.h"
#include "scanner.h"
#include "rangefinder.h"
#include "servo.h"
#include "mechanical_properties.h"
#include "utils/emb_malloc.h"
#include <math.h>

void command_task(void* pd) {
   //on startup, send back an information packet
   information_reply_packet_t infoReply;
   pose_t startingPose;
   getPose(&startingPose);
   infoReply.location.x= startingPose.x;
   infoReply.location.y= startingPose.y;
   infoReply.maxScanRadius= MAX_SCAN_RADIUS;
   infoReply.minScanRadius= MIN_SCAN_RADIUS;
   infoReply.requiresMap= false;
   for(int i= 0; i < 1; i++) {
      send(PARENT_ID, INFORMATION_REPLY_PACKET_OPCODE, &infoReply);
   }

   //identify scout ID to indicate joined to swarm
   PORTK.OUT= ~(SCOUT_LED_ID << 4);

   uint8_t err;
   swarm_communication_t* nextCommand;
   //all tasks must wait in a loop forever
   for(;;) {
      //grab a queue'd message
      nextCommand= OSQPend(command_q, 0, &err);
      uint8_t opcode= nextCommand->opcode;
      //do command based on opcode
      if(opcode == SCAN_PACKET_OPCODE) {
         scan_packet_t* packet;
         packet= nextCommand->packet;
         scan(packet->radius);
         emb_free(packet);
      }
      else if(opcode == PATH_PACKET_OPCODE) {
         path_packet_t* packet;
         packet= nextCommand->packet;
         drivePath(packet->waypoints, packet->numWaypoints);
         emb_free(packet->waypoints);
         emb_free(packet);
      }

      emb_free(nextCommand);
   }
}
/**
 * Perform a scan.
 * @param radius The maximum radius to scan.  Only points seen within
 *               the specified radius will be recorded as being seen.
 */
void scan(uint8_t radius) {
   object_t observedObject;
   int8_t retVal;
   //turn on the servo and the IR sensors
   servoEnable(MAIN_SERVO, true);
   rangeFinderEnable(IR_FRONT, true);
   rangeFinderEnable(IR_REAR, true);
   for(int8_t angle= -90; angle <= 90; angle+= DEG_STEP) {
      add_point_map_packet_t packet;
      //scan with front IR
      retVal= scanAngle(angle, IR_FRONT, radius, &observedObject);
      if(retVal == 0) {
         //send back to parent
         packet.location.x= observedObject.x;
         packet.location.y= observedObject.y;
         packet.propertiesBitmask= observedObject.properties_bm;
         packet.stdDeviation= observedObject.stdDeviation;
         send(PARENT_ID, ADD_POINT_MAP_PACKET_OPCODE, &packet);
      }
      //otherwise nothing was seen so do nothing

      //Scan with rear ir
      retVal= scanAngle(angle, IR_REAR, radius, &observedObject);
      if(retVal == 0) {
         //send back to parent
         packet.location.x= observedObject.x;
         packet.location.y= observedObject.y;
         packet.propertiesBitmask= observedObject.properties_bm;
         packet.stdDeviation= observedObject.stdDeviation;
         send(PARENT_ID, ADD_POINT_MAP_PACKET_OPCODE, &packet);
      }
      //otherwise nothing was seen so do nothing
   }
   //turn off the servo and the IR sensors
   servoEnable(MAIN_SERVO, false);
   rangeFinderEnable(IR_FRONT, false);
   rangeFinderEnable(IR_REAR, false);
   //All done!
   map_complete_unknown_shadow_t packet;
   //packet has no body so no need to fill in fields
   send(PARENT_ID, MAP_COMPLETE_UNKNOWN_SHADOW_OPCODE, &packet);
}

/**
 * Drive along a given series of waypoints.
 */
void drivePath(coordinate_t* waypoints, uint8_t numWaypoints) {
   pose_t nextPose;
   getPose(&nextPose); //initialize nextPose to be where we are now
                       //so if we get an empty packet list we
                       //respond with where we are now
   bool noCrash;
   for(uint8_t i= 0; i < numWaypoints; i++) {
      //set values in pose
      nextPose.x= waypoints[i].x;
      nextPose.y= waypoints[i].y;
      nextPose.theta= NAN; //use NAN to indicate we don't care about the angle of arrival
      noCrash= driveToPose(&nextPose);
      if(!noCrash) {
         break;
      }
   }
   if(noCrash) {
      reached_destination_packet packet;
      packet.destination.x= round(nextPose.x);
      packet.destination.y= round(nextPose.y);
      //packet has no body so no need to fill in fields
      send(PARENT_ID, REACHED_DESTINATION_PACKET_OPCODE, &packet);
   }
   else {
      //send a "we crashed" packet
   }
}

/**
 * Respond to a request information packet.
 */
void requestInformationResponse() {
   information_reply_packet_t packet;
   pose_t currentPose;
   getPose(&currentPose);
   packet.location.x= currentPose.x;
   packet.location.y= currentPose.y;
   packet.maxScanRadius= MAX_SCAN_RADIUS;
   packet.minScanRadius= MIN_SCAN_RADIUS;
   send(PARENT_ID, INFORMATION_REPLY_PACKET_OPCODE, &packet);
}


