/*
 * communications.c
 *
 *  Created on: Dec 19, 2010
 *      Author: hsrmqp
 */

#include "communications.h"
#include "xbee.h"
#include "utils/serialize.h"
#include "utils/emb_malloc.h"
#include "includes.h"

static uint8_t transmitBuffer[XBEE_MAX_PAYLOAD];

//Static prototypes
uint8_t serializeScanPacket(scan_packet_t* packet, uint8_t serialDataBuf[]);
void deSerializeScanPacket(scan_packet_t* packet, uint8_t serialData[], uint8_t dataLen);
uint8_t serializePathPacket(path_packet_t* packet, uint8_t serialDataBuf[]);
void deSerializePathPacket(path_packet_t* packet, uint8_t serialData[], uint8_t dataLen);
uint8_t serializeInformationReplyPacket(information_reply_packet_t* packet, uint8_t serialDataBuf[]);
void deSerializeInformationReplyPacket(information_reply_packet_t* packet, uint8_t serialData[], uint8_t dataLen);
uint8_t serializeAddPointMapPacket(add_point_map_packet_t* packet, uint8_t serialDataBuf[]);
void deSerializeAddPointMapPacket(add_point_map_packet_t* packet, uint8_t serialData[], uint8_t dataLen);
uint8_t serializeMapCompleteUnknownShadowPacket(map_complete_unknown_shadow_t* packet, uint8_t serialDataBuf[]);
void deSerializeMapCompleteUnknownShadowPacket(map_complete_unknown_shadow_t* packet, uint8_t serialData[], uint8_t dataLen);
uint8_t serializeReachedDestinationPacket(reached_destination_packet* packet, uint8_t serialDataBuf[]);
void deSerializeReachedDestinationPacket(reached_destination_packet* packet, uint8_t serialData[], uint8_t dataLen);

void send(uint8_t dst, uint8_t opcode, void* pkt) {
   uint8_t numBytes= 0;
   switch(opcode) {
   case INFORMATION_REPLY_PACKET_OPCODE:
      numBytes= serializeInformationReplyPacket(pkt, transmitBuffer);
      break;
   case ADD_POINT_MAP_PACKET_OPCODE:
      numBytes= serializeAddPointMapPacket(pkt, transmitBuffer);
      break;
   case MAP_COMPLETE_UNKNOWN_SHADOW_OPCODE:
      numBytes= serializeMapCompleteUnknownShadowPacket(pkt, transmitBuffer);
      break;
   case REACHED_DESTINATION_PACKET_OPCODE:
      numBytes= serializeReachedDestinationPacket(pkt, transmitBuffer);
   default:
      //Can't send this packet, do nothing
      break;
   }
   uint8_t sendStatus;
   do {
      sendStatus= xbee_send(dst, transmitBuffer, numBytes);
      if(sendStatus != XBEE_SEND_SUCCESS) {
         _delay_ms(10); //delay slightly before retry
      }
   }while(sendStatus!= XBEE_SEND_SUCCESS);

}

void remote_RXDriverCallback(uint8_t src, uint8_t* data, uint8_t len) {
   swarm_communication_t* swarmComm;
   swarmComm= emb_malloc(sizeof(swarm_communication_t));
   uint8_t opcode= data[0];
   swarmComm->id= src;
   swarmComm->opcode= opcode;
   switch(src) {
   case PARENT_ID:
      //message from a parent
      if(opcode == PATH_PACKET_OPCODE) {
         path_packet_t* pathPak;
         pathPak= emb_malloc(sizeof(path_packet_t));
         pathPak->waypoints= emb_malloc(sizeof(coordinate_t)*24); //shouldn't ever have more than 24 waypoints
         swarmComm->packet= pathPak;
         deSerializePathPacket(pathPak, data, len);
      }

      else if(opcode == SCAN_PACKET_OPCODE) {
         scan_packet_t* scanPak;
         scanPak= emb_malloc(sizeof(scan_packet_t));
         swarmComm->packet= scanPak;
         deSerializeScanPacket(scanPak, data, len);
      }

      else {
         break;
      }
      //add swarm communication to queue of commands
      OSQPost(command_q, swarmComm);
      break;
   case INVALID_SENDER:
      //message from invalid sender
      //only thing here should be new parent adopting this scout
      break;
   case BROADCAST_SENDER:
      //packet was a broadcast packet
      //scout probably has no use for this
      break;
   default:
      //scout has no children so do nothing
      break;
   }
}


/**
 * Serialize a scan packet.  This will convert the scan packet into an array
 * of bytes and store it in the given buffer.  The buffer must ensure that
 * it is large enough to hold the entire packet.  All data is sent big endian.
 * @param packet The scan packet to serialize.
 * @param serialDataBuf An array of bytes to store the serialized data.
 * @return The number of bytes written to the buffer.
 */
uint8_t serializeScanPacket(scan_packet_t* packet, uint8_t serialDataBuf[]) {
   //copy out the fields of the scan packet
   serialDataBuf[0]= SCAN_PACKET_OPCODE;
   uint8_t size= 1; //length of serial packet
   size+= serializeInt16(packet->radius, &serialDataBuf[1]);

   return size;
}

/**
 * Deserialize a scan packet.  This will convert an array of bytes into
 * a scan packet.  The first byte of the serialData array MUST be equal
 * to the SCAN_PACET_OPCODE otherwise the function will return immediately.
 * The array must be big endian.
 * @param packet Location to store the output packet.
 * @param serialData The serialized data.
 * @param dataLen The number of bytes in the serialized data.
 */
void deSerializeScanPacket(scan_packet_t* packet, uint8_t serialData[], uint8_t dataLen) {
   //verify that the first byte is the correct opcode.
   if(serialData[0]!= SCAN_PACKET_OPCODE) {
      return;
   }
   uint8_t offset= 1; //offset into the serial data array;
   //first byte is correct opcode, continue execution
   deSerializeInt16(&(packet->radius), &serialData[offset]);
}

/**
 * Serialize a path packet.  This will convert a path
 * packet into an array of bytes.  The size of the buffer
 * must be large enough to hold the entire packet.  The data will be stored
 * big endian.
 * @param packet The packet to serialize.
 * @param serialDataBuf Location to store the serialized data.
 * @return The number of bytes written to the serial data buffer.
 */
uint8_t serializePathPacket(path_packet_t* packet, uint8_t serialDataBuf[]) {
   //set opcode
   serialDataBuf[0]= PATH_PACKET_OPCODE;
   uint8_t size= 1;
   //for each waypoint in the packet
   for(uint8_t i= 0; i < packet->numWaypoints; i++) {
      //serialize the x and y component
      size+= serializeInt16(packet->waypoints[i].x, &serialDataBuf[size]);
      size+= serializeInt16(packet->waypoints[i].y, &serialDataBuf[size]);
   }

   return size;
}

/**
 * Deserialize a path packet.  This will convert an array of bytes into
 * a path packet.  The first byte of the serialData array MUST be equal
 * to the PATH_PACKET_OPCODE otherwise the function will return immediately.
 * The array must be big endian.
 * @param packet Location to store the output packet.
 * @param serialData The serialized data.
 * @param dataLen The number of bytes in the serialized data.
 */
void deSerializePathPacket(path_packet_t* packet, uint8_t serialData[], uint8_t dataLen) {
   //confirm correct opcode
   if(serialData[0] != PATH_PACKET_OPCODE) {
      return;
   }
   //continue with execution
   uint8_t offset= 1; //offset into serial data array
   //for each waypoint, copy out of the serial data array
   packet->numWaypoints= 0;
   for(uint8_t i= 0; offset < dataLen; i++) {
      offset+= deSerializeInt16(&(packet->waypoints[i].x), &serialData[offset]);
      offset+= deSerializeInt16(&(packet->waypoints[i].y), &serialData[offset]);
      (packet->numWaypoints)++;
   }
}

/**
 * Serialize an information reply packet.  This will convert an information
 * reply packet into an array of bytes.  The size of the buffer
 * must be large enough to hold the entire packet.  The data will be stored
 * big endian.
 * @param packet The packet to serialize.
 * @param serialDataBuf Location to store the serialized data.
 * @return The number of bytes written to the serial data buffer.
 */
uint8_t serializeInformationReplyPacket(information_reply_packet_t* packet, uint8_t serialDataBuf[]) {
   //set opcode
   serialDataBuf[0]= INFORMATION_REPLY_PACKET_OPCODE;
   uint8_t size= 1;
   //serialize the x component of the location
   size+= serializeInt16(packet->location.x, &serialDataBuf[size]);
   //serialize the y component of the location
   size+= serializeInt16(packet->location.y, &serialDataBuf[size]);
   //serialize the max scan radius
   size+= serializeInt16(packet->maxScanRadius, &serialDataBuf[size]);
   //serialize the max scan radius
   size+= serializeInt16(packet->minScanRadius, &serialDataBuf[size]);
   //do not need a map
   size+= serializeInt8(packet->requiresMap, &serialDataBuf[size]);

   return size;
}

/**
 * Deserialize an information reply packet.  This will convert an array of bytes into
 * an information reply packet.  The first byte of the serialData array MUST be equal
 * to the INFORMATION_REPLY_PACKET_OPCODE otherwise the function will return immediately.
 * The array must be big endian.
 * @param packet Location to store the output packet.
 * @param serialData The serialized data.
 * @param dataLen The number of bytes in the serialized data.
 */
void deSerializeInformationReplyPacket(information_reply_packet_t* packet, uint8_t serialData[], uint8_t dataLen) {
   if(serialData[0]!= INFORMATION_REPLY_PACKET_OPCODE) {
      return;
   }

   //continue with execution
   uint8_t offset= 1; //offset into serial data array
   //deserialize x component of location
   offset+= deSerializeInt16(&(packet->location.x), &serialData[offset]);
   //deserialize y component of location
   offset+= deSerializeInt16(&(packet->location.y), &serialData[offset]);
   //deserialize the max scan radius
   offset+= deSerializeInt16(&(packet->maxScanRadius), &serialData[offset]);
   //deserialize the min scan radius
   offset+= deSerializeInt16(&(packet->minScanRadius), &serialData[offset]);
   //deserialize if map is required
   offset+= deSerializeInt8(&(packet->requiresMap), &serialData[offset]);
}

/**
 * Serialize an add point to map packet.  This will convert an add point to
 * map packet into an array of bytes.  The size of the buffer
 * must be large enough to hold the entire packet.  The data will be stored
 * big endian.
 * @param packet The packet to serialize.
 * @param serialDataBuf Location to store the serialized data.
 * @return The number of bytes written to the serial data buffer.
 */
uint8_t serializeAddPointMapPacket(add_point_map_packet_t* packet, uint8_t serialDataBuf[]) {
   //set opcode
   serialDataBuf[0]= ADD_POINT_MAP_PACKET_OPCODE;
   uint8_t size= 1;
   //serialize x location
   size+= serializeInt16(packet->location.x, &serialDataBuf[size]);
   //serialize y location
   size+= serializeInt16(packet->location.y, &serialDataBuf[size]);
   //serialize standard deviation
   size+= serializeInt16(packet->stdDeviation, &serialDataBuf[size]);
   //serialize properties bitmask
   size+= serializeInt8(packet->propertiesBitmask, &serialDataBuf[size]);

   return size;
}

/**
 * Deserialize an add point to map packet.  This will convert an array of bytes into
 * an add point to map packet.  The first byte of the serialData array MUST be equal
 * to the ADD_POINT_MAP_PACKET_OPCODE otherwise the function will return immediately.
 * The array must be big endian.
 * @param packet Location to store the output packet.
 * @param serialData The serialized data.
 * @param dataLen The number of bytes in the serialized data.
 */
void deSerializeAddPointMapPacket(add_point_map_packet_t* packet, uint8_t serialData[], uint8_t dataLen) {
   if(serialData[0]!= ADD_POINT_MAP_PACKET_OPCODE) {
      return;
   }
   uint8_t offset= 1;

   offset+= deSerializeInt16(&(packet->location.x), &serialData[offset]);
   offset+= deSerializeInt16(&(packet->location.y), &serialData[offset]);
   offset+= deSerializeInt16(&(packet->stdDeviation), &serialData[offset]);
   offset+= deSerializeInt8(&(packet->propertiesBitmask), &serialData[offset]);

}

/**
 * Serialize a map complete unknown shadowing packet.  This will convert a map complete
 * unknown shadowing packet into an array of bytes.  The size of the buffer
 * must be large enough to hold the entire packet.  The data will be stored
 * big endian.
 * @param packet The packet to serialize.
 * @param serialDataBuf Location to store the serialized data.
 * @return The number of bytes written to the serial data buffer.
 */
uint8_t serializeMapCompleteUnknownShadowPacket(map_complete_unknown_shadow_t* packet, uint8_t serialDataBuf[]) {
   //set opcode
   serialDataBuf[0]= MAP_COMPLETE_UNKNOWN_SHADOW_OPCODE;
   uint8_t size= 1;
   //packet has no body so we're done
   return size;
}

/**
 * Deserialize a map complete unknown shadowing packet.  This will convert an array of bytes into
 * a map complete unknown shadowing packet.  The first byte of the serialData array MUST be equal
 * to the MAP_COMPLETE_UNKNOWN_SHADOW_OPCODE otherwise the function will return immediately.
 * The array must be big endian.
 * @param packet Location to store the output packet.
 * @param serialData The serialized data.
 * @param dataLen The number of bytes in the serialized data.
 */
void deSerializeMapCompleteUnknownShadowPacket(map_complete_unknown_shadow_t* packet, uint8_t serialData[], uint8_t dataLen) {
   //verify that the first byte is the correct opcode.
   if(serialData[0]!= MAP_COMPLETE_UNKNOWN_SHADOW_OPCODE) {
      return;
   }
   //packet has no body so we're done
}

/**
 * Serialize a ready packet.  This will convert a ready
 * packet into an array of bytes.  The size of the buffer
 * must be large enough to hold the entire packet.  The data will be stored
 * big endian.
 * @param packet The packet to serialize.
 * @param serialDataBuf Location to store the serialized data.
 * @return The number of bytes written to the serial data buffer.
 */
uint8_t serializeReachedDestinationPacket(reached_destination_packet* packet, uint8_t serialDataBuf[]) {
   //set opcode
   serialDataBuf[0]= REACHED_DESTINATION_PACKET_OPCODE;
   uint8_t size= 1;
   //serialize destination
   size+= serializeInt16(packet->destination.x, &serialDataBuf[size]);
   size+= serializeInt16(packet->destination.y, &serialDataBuf[size]);
   return size;
}

/**
 * Deserialize a ready packet.  This will convert an array of bytes into
 * a ready packet.  The first byte of the serialData array MUST be equal
 * to the READY_PACKET_OPCODE otherwise the function will return immediately.
 * The array must be big endian.
 * @param packet Location to store the output packet.
 * @param serialData The serialized data.
 * @param dataLen The number of bytes in the serialized data.
 */
void deSerializeReachedDestinationPacket(reached_destination_packet* packet, uint8_t serialData[], uint8_t dataLen) {
   //verify that the first byte is the correct opcode.
   if(serialData[0]!= REACHED_DESTINATION_PACKET_OPCODE) {
      return;
   }
   uint8_t offset= 1;
   //deserialize the x,y
   offset+= deSerializeInt16(&(packet->destination.x), &serialData[offset]);
   offset+= deSerializeInt16(&(packet->destination.y), &serialData[offset]);
}

