/*************************************************************************
 *
 *
 *              Task1: Dynamic clock switching
 *
 *
 *************************************************************************/

// Include header files for GCC/IAR
#include "avr_compiler.h"

// Include Clock system driver from application note AVR1003
#include "clksys_driver.h"

#define SWITCHPORT PORTF
#define LEDPORT PORTE

int clockTestMain1(void)
{
	// Configure switches
	PORTCFG.MPCMASK = 0xff; // Configure several PINxCTRL registers at the same time
	SWITCHPORT.PIN0CTRL = (SWITCHPORT.PIN0CTRL & ~PORT_OPC_gm) | PORT_OPC_PULLUP_gc; //Enable pull-up to get a defined level on the switches

	SWITCHPORT.DIRCLR = 0xff; // Set port as input

	// Configure LEDs
	PORTCFG.MPCMASK = 0xff; // Configure several PINxCTRL registers at the same time
	LEDPORT.PIN0CTRL = PORT_INVEN_bm; // Invert input to turn the leds on when port output value is 1

	LEDPORT.DIRSET = 0xff; 	// Set port as output
	LEDPORT.OUT = 0xa1;  // Set initial value

	unsigned int counter = 0;

	while (1)
	{

		/*  Enable internal 2 MHz rc oscillator and wait until it's
		 *  stable. Set the 2 MHz rc oscillator as the main clock source.
		 *  Then disable other oscillators.
		 */
		if ( (SWITCHPORT.IN & 0x01) == 0 )
		{
			CLKSYS_Enable( OSC_RC2MEN_bm );
			do {} while ( CLKSYS_IsReady( OSC_RC2MRDY_bm ) == 0 );
			CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_RC2M_gc );
			CLKSYS_Disable( OSC_RC32MEN_bm | OSC_RC32KEN_bm );

			LEDPORT.OUTCLR = 0x0f;
			LEDPORT.OUTSET = 0x01;
		}

		/*  Enable internal 32 MHz ring oscillator and wait until it's
		 *  stable. Set the 32 MHz ring oscillator as the main clock source.
		 *  Then disable the other oscillators.
		 */
		if ( (SWITCHPORT.IN & 0x02) == 0 )
		{
			CLKSYS_Enable( OSC_RC32MEN_bm );
			do {} while ( CLKSYS_IsReady( OSC_RC32MRDY_bm ) == 0 );
			CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_RC32M_gc );
			CLKSYS_Disable( OSC_RC2MEN_bm | OSC_RC32KEN_bm );


			LEDPORT.OUTCLR = 0x0f;
			LEDPORT.OUTSET = 0x02;
		}

		/*  Enable internal 32 kHz calibrated oscillator and check for
		 *  it to be stable. Set the 32 kHz oscillator as the main clock source.
		 *  Then disable the other oscillators.
		 */
		if ( (SWITCHPORT.IN & 0x04) == 0 )
		{
			CLKSYS_Enable( OSC_RC32KEN_bm );
			do {} while ( CLKSYS_IsReady( OSC_RC32KRDY_bm ) == 0 );
			CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_RC32K_gc );
			CLKSYS_Disable( OSC_RC2MEN_bm | OSC_RC32MEN_bm );

			LEDPORT.OUTCLR = 0x0f;
			LEDPORT.OUTSET = 0x04;
		}

		// Blink LEDS after the loop has been run several times
		// This gives us visual feedback on the MCU speed
		counter++;
		if (counter > 20000 )
		{
			LEDPORT.OUTTGL = 0xe0;
			counter=0;
		}


	}
}

/*************************************************************************
 *
 *
 *              Task2: Phase Locked Loop (PLL)
 *
 *
 *************************************************************************/

int clockTestMain2(void)
{
	// Configure switches
	PORTCFG.MPCMASK = 0xff; // Configure several PINxCTRL registers at the same time
	SWITCHPORT.PIN0CTRL = (SWITCHPORT.PIN0CTRL & ~PORT_OPC_gm) | PORT_OPC_PULLUP_gc; //Enable pull-up to get a defined level on the switches

	SWITCHPORT.DIRCLR = 0xff; // Set port as input

	// Configure LEDs
	PORTCFG.MPCMASK = 0xff; // Configure several PINxCTRL registers at the same time
	LEDPORT.PIN0CTRL = PORT_INVEN_bm; // Invert input to turn the leds on when port value is 1

	LEDPORT.DIRSET = 0xff; 	// Set port as output
	LEDPORT.OUT = 0xa0;  // Set initial value

	unsigned int counter = 0;

	while (1)
	{

		/*  Change the clock source to the internal 2MHz. Disable PLL.
		 *  Configure PLL with the 2 MHz RC oscillator as source and
		 *  multiply by 15 to get 30 MHz PLL clock and enable it. Wait
		 *  for it to be stable and set prescaler C to divide by two
		 *  to set the CPU clock to 15 MHz
		 */
		if ( (SWITCHPORT.IN & 0x01) == 0 )
		{
			CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_RC2M_gc );

			CLKSYS_Disable( OSC_PLLEN_bm );

			CLKSYS_PLL_Config( OSC_PLLSRC_RC2M_gc, 15 );
			CLKSYS_Enable( OSC_PLLEN_bm );
			CLKSYS_Prescalers_Config( CLK_PSADIV_1_gc, CLK_PSBCDIV_1_2_gc );
			do {} while ( CLKSYS_IsReady( OSC_PLLRDY_bm ) == 0 );
			CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_PLL_gc );

			LEDPORT.OUTCLR = 0x0f;
			LEDPORT.OUTSET = 0x01;
		}

		/*  Configure PLL with the 2 MHz RC oscillator as source and
		 *  multiply by 15 to get 30 MHz PLL clock and enable it. Wait
		 *  for it to be stable and set prescaler B and C to divide by four
		 *  to set the CPU clock to 7.5 MHz
		 */
		if ( (SWITCHPORT.IN & 0x02) == 0 )
		{
			CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_RC2M_gc );

			CLKSYS_Disable( OSC_PLLEN_bm );

			CLKSYS_PLL_Config( OSC_PLLSRC_RC2M_gc, 15 );
			CLKSYS_Enable( OSC_PLLEN_bm );
			CLKSYS_Prescalers_Config( CLK_PSADIV_1_gc, CLK_PSBCDIV_2_2_gc );
			do {} while ( CLKSYS_IsReady( OSC_PLLRDY_bm ) == 0 );
			CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_PLL_gc );

			LEDPORT.OUTCLR = 0x0f;
			LEDPORT.OUTSET = 0x02;
		}


		/*  Configure PLL with the 2 MHz RC oscillator as source and
		 *  multiply by 5 to get 10 MHz PLL clock and enable it. Wait
		 *  for it to be stable and set prescaler C to divide by two
		 *  to set the CPU clock to 5 MHz
		 */
		if ( (SWITCHPORT.IN & 0x04) == 0 )
		{
			CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_RC2M_gc );

			CLKSYS_Disable( OSC_PLLEN_bm );

			CLKSYS_PLL_Config( OSC_PLLSRC_RC2M_gc, 5 );
			CLKSYS_Enable( OSC_PLLEN_bm );
			CLKSYS_Prescalers_Config( CLK_PSADIV_1_gc, CLK_PSBCDIV_1_2_gc );
			do {} while ( CLKSYS_IsReady( OSC_PLLRDY_bm ) == 0 );
			CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_PLL_gc );

			LEDPORT.OUTCLR = 0x0f;
			LEDPORT.OUTSET = 0x04;
		}


		/*  Dynamically change prescaler:
		 *  Set prescaler A to divide by 2, prescaler B with no division,
		 *  and prescaler C to divide by 2. This will give the CPU a freqency
		 *  of 1/4th of the input freqency. Input freqency will depend on the
		 *  previous switch that has been pressed.
		 */
		if ( (SWITCHPORT.IN & 0x08) == 0 )
		{
			CLKSYS_Prescalers_Config( CLK_PSADIV_2_gc, CLK_PSBCDIV_1_2_gc );
			LEDPORT.OUTSET = 0x08;
		}


		/*
		 *  Add code here for generating 62 MHz to clkper4 and 15.5 MHz to the CPU
		 */
		if ( (SWITCHPORT.IN & 0x10) == 0 )
		{

		}


		// Blink LEDS after the loop has been run several times
		// This gives us visual feedback on the MCU speed
		counter++;
		if (counter > 20000 )
		{
			LEDPORT.OUTTGL = 0xe0;
			counter=0;
		}


	}
}

/*************************************************************************
 *
 *
 *              Task3: Runtime calibration with DFLL
 *
 *
 *************************************************************************/


int clockTestMain3(void)
{
	// Configure switches
	PORTCFG.MPCMASK = 0xff; // Configure several PINxCTRL registers at the same time
	SWITCHPORT.PIN0CTRL = (SWITCHPORT.PIN0CTRL & ~PORT_OPC_gm) | PORT_OPC_PULLUP_gc; //Enable pull-up to get a defined level on the switches

	SWITCHPORT.DIRCLR = 0xff; // Set port as input

	// Configure LEDs
	PORTCFG.MPCMASK = 0xff; // Configure several PINxCTRL registers at the same time
	LEDPORT.PIN0CTRL = PORT_INVEN_bm; // Invert input to turn the leds on when port output value is 1

	LEDPORT.DIRSET = 0xff; 	// Set port as output
	LEDPORT.OUT = 0xa0;  // Set initial value

	unsigned int counter = 0;

	// Enable the 32KHz internal oscillator which is used as reference for the DFLL
	CLKSYS_Enable( OSC_RC32KEN_bm );

	while (1)
	{

		/*  Enable automatic calibration of the internal 2MHz oscillator.
		 *  We do not use the external crystal connected to the TOSC pins,
		 *  but the internal 32kHz oscillator as reference.
		 */
		if ( (SWITCHPORT.IN & 0x01) == 0 )
		{
			CLKSYS_AutoCalibration_Enable( OSC_RC2MCREF_bm, false );

		/*  Due to an errata on ATxmega128A1 we need to enable the 32MHz internal RC oscillator
		 *  and the Autocalibration for the 32MHz internal RC oscillator in order for the calibration
		 *  to work correctly on the 2 MHz internal RC oscillator*/

		 	CLKSYS_Enable( OSC_RC32MEN_bm );
			do {} while ( CLKSYS_IsReady( OSC_RC32MRDY_bm ) == 0 );
			CLKSYS_AutoCalibration_Enable( OSC_RC32MCREF_bm, false );

			LEDPORT.OUTCLR = 0x0f;
			LEDPORT.OUTSET = 0x01;
		}

		/*  Disable the automatic calibration of the internal 2MHz oscillator.
		 */
		if ( (SWITCHPORT.IN & 0x02) == 0 )
		{
			CLKSYS_AutoCalibration_Disable( DFLLRC2M );


			//Disable the auto calibration of the 32MHz internal RC oscillator and the 32MHz internal RC oscillator
			CLKSYS_AutoCalibration_Disable( DFLLRC32M );
			CLKSYS_Disable( OSC_RC32MEN_bm );

			LEDPORT.OUTCLR = 0x0f;
			LEDPORT.OUTSET = 0x02;
		}



		// Blink LEDS after the loop has been run several times
		// This gives us visual feedback on the MCU speed
		counter++;
		if (counter > 20000 )
		{
			LEDPORT.OUTTGL = 0xe0;
			counter=0;
		}


	}
}

/*************************************************************************
 *
 *
 *              Task4: Real Time Counter
 *
 *
 *************************************************************************/

#define RTC_HOURS_PER_DAY 24
#define RTC_MINUTES_PER_HOUR 60
#define RTC_SECONDS_PER_MINUTE 60
#define RTC_TICKS_PER_SECOND 10  // Change this if RTC prescaling changes.

volatile uint8_t	RTC_days;     // Holds day of week.
volatile uint8_t    RTC_hours;    // Holds hour of day.
volatile uint8_t    RTC_minutes;  // Holds minute of hour.
volatile uint8_t    RTC_seconds;  // Holds second of minute.
volatile uint8_t 	RTC_ticks;    // Holds tick count. Tick granularity depends on prescaler setting.


ISR( RTC_OVF_vect )
{
	// Count ticks or wrap back to zero.
	if (RTC_ticks < (RTC_TICKS_PER_SECOND - 1))
	{
		RTC_ticks++;
	} else
	{
		RTC_ticks = 0;

		// Count seconds or wrap back to zero.
		if (RTC_seconds < (RTC_SECONDS_PER_MINUTE - 1))
		{
			RTC_seconds++;
		} else
		{
			RTC_seconds = 0;

			// Count minutes or wrap back to zero.
			if (RTC_minutes < (RTC_MINUTES_PER_HOUR - 1))
			{
				RTC_minutes++;
			} else
			{
				RTC_minutes = 0;

				// Count hours or wrap back to zero.
				if (RTC_hours < (RTC_HOURS_PER_DAY - 1))
				{
					RTC_hours++;
				} else
				{
					RTC_hours = 0;

					// Count days
					RTC_days++;

					// Date processing could be added here
				}
			}
		}
	}
}

int clockTestMain4(void)
{
	// Configure switches: Enable pull-up to get a defined level on the switches
	// and invert the input to make the switches go high when pressed
	PORTCFG.MPCMASK = 0xff; // Configure several PINxCTRL registers at the same time
	SWITCHPORT.PIN0CTRL = (SWITCHPORT.PIN0CTRL & ~PORT_OPC_gm) | PORT_OPC_PULLUP_gc | PORT_INVEN_bm;

	SWITCHPORT.DIRCLR = 0xff; // Set port as input

	// Configure LEDs
	PORTCFG.MPCMASK = 0xff; // Configure several PINxCTRL registers at the same time
	LEDPORT.PIN0CTRL = PORT_INVEN_bm; // Invert input to turn the leds on when port output value is 1

	LEDPORT.DIRSET = 0xff; 	// Set port as output
	LEDPORT.OUT = 0xa0;  // Set initial value

	// Configure RTC
	// Set Internal 32kHz oscillator prescaled to 1kHz as source for RTC
	CLKSYS_RTC_ClockSource_Enable( CLK_RTCSRC_RCOSC_gc );
	// Enable Low level interrupts for the RTC Overflow
	RTC.INTCTRL = ( RTC.INTCTRL & ~RTC_OVFINTLVL_gm) | RTC_OVFINTLVL_LO_gc;

	// Enable low level interrupts in the programmable interrupt controller
	PMIC.CTRL |= PMIC_LOLVLEN_bm;

	// Set initial period for RTC and start the RTC with no prescaler
	RTC.PER = 100;
	RTC.CTRL |= RTC_PRESCALER_DIV1_gc;

	// Enable interrupts.
	sei();

	while (1)
	{
		switch ( SWITCHPORT.IN )
		{
			case 0x01:
				// Show 1/10th seconds
				LEDPORT.OUT = RTC_ticks;
				break;
			case 0x02:
			default:
				// Show seconds
				LEDPORT.OUT = RTC_seconds;
				break;
			case 0x04:
				// Show minutes
				LEDPORT.OUT = RTC_minutes;
				break;
			case 0x08:
				// Show hours
				LEDPORT.OUT = RTC_hours;
				break;

		}
	}
}



