/*
 * dmaTest.c
 *
 *  Created on: Nov 16, 2010
 *      Author: hsrmqp
 */

/**************************************************************************
 *
 *
 *                            Task1: DMAC Memory Copy
 *
 *
 **************************************************************************/


#include "dma_driver.h"
#include "avr_compiler.h"

#define LEDPORT PORTE

// Size of memory block must not
// exceed 64k for this demo application.
#define MEM_BLOCK_SIZE (100)

// Memory blocks A and B for testing
uint8_t memoryBlockA[MEM_BLOCK_SIZE];
uint8_t memoryBlockB[MEM_BLOCK_SIZE];

// Global declared status for interrupt routine
volatile bool gInterruptDone;
volatile bool gTransferError = false;

/*  AsyncMemCopy
 *
 *  Block size 0 = 64k. Enable the DMA channel to use first.
 *  Then setup channel for copying data, increasing addresses,
 *  no address pointer reload, with 8-byte bursts.
 *
 * \note This function is asynchronous and DOES NOT wait for
 *       completion.This must be handled by the program
 *       calling this function.
 *       In this example, an interrupt.
 *
 */
void AsyncMemCopy( const void * src,
                   void * dest,
                   uint16_t blockSize,
                   DMA_CH_t * dmaChannel )
{
	DMA_EnableChannel( dmaChannel );

	DMA_SetupBlock( dmaChannel,
	                src,
	                DMA_CH_SRCRELOAD_NONE_gc,
	                DMA_CH_SRCDIR_INC_gc,
	                dest,
	                DMA_CH_DESTRELOAD_NONE_gc,
	                DMA_CH_DESTDIR_INC_gc,
	                blockSize,
	                DMA_CH_BURSTLEN_8BYTE_gc,
	                0,
	                false );

	DMA_StartTransfer( dmaChannel );
}



/*  MemCopy
 *
 *  Block size 0 = 64k. Enable the DMA channel to use first and the channel
 *  will be disabled automatically. A parameter check to avoid illegal values.
 *  Setup channel for copying data, increasing addresses, no address pointer
 *  reload, with 8-byte bursts.
 *
 * \note This function wait until the transfer is complete or an error occurs
 *       before returning which makes this function blocking or synchronous.
 *
 * \retval true  if failure.
 * \retval false if success.
 */


bool MemCopy( const void * src, void * dest, uint16_t blockSize,
                          DMA_CH_t * dmaChannel )
{
    uint8_t flags = 0;

	DMA_EnableChannel( dmaChannel );
	DMA_SetupBlock( dmaChannel,
	                src,
	                DMA_CH_SRCRELOAD_NONE_gc,
	                DMA_CH_SRCDIR_INC_gc,
	                dest,
	                DMA_CH_DESTRELOAD_NONE_gc,
	                DMA_CH_DESTDIR_INC_gc,
	                blockSize,
	                DMA_CH_BURSTLEN_8BYTE_gc,
	                0,
	                false ); // no repeat

	DMA_StartTransfer( dmaChannel );

    // Wait until the completion or error flag is set. The flags
	// must be cleared manually.
    do {
        flags = DMA_ReturnStatus_non_blocking( dmaChannel );
    } while ( flags == 0 );

	// Clear flags
    DMA.CH0.CTRLB |= ( flags );

    // Check if error flag is set
    if ( ( flags & DMA_CH_ERRIF_bm ) != 0x00 )
    {
        return true;
    } else
    {
        return false;
    }
}


//  The code using the two example functions to do block memory copy
//  The first test use a polling function to wait until the transfer
//  completes or get an error. The second test use an interrupt to
//  know when the transfer completes or get an error.
int dmaTestMain1( void )
{
	uint16_t index;
    LEDPORT.DIR = 0xFF;
    LEDPORT.OUT = 0xFF;

	DMA_CH_t * Channel;
	Channel = &DMA.CH0;

	DMA_Enable();

    // Test 1: Copy using polling.

    // Fill memory block A with example data.
	for ( index = 0; index < MEM_BLOCK_SIZE; ++index ) {
		memoryBlockA[index] = ( (uint8_t) index & 0xff );
	}

	// Copy data using channel 0.
	gTransferError = MemCopy(
							memoryBlockA,
                            memoryBlockB,
                            MEM_BLOCK_SIZE,
                            Channel );

	// Compare memory blocks if status is true
	if ( gTransferError )
    {
        // Signal Error on LEDs by turning them all on
        LEDPORT.OUT = 0x00;
	}
	else
	{
		// Signal success on one LED
		LEDPORT.OUT = ~0x01;
	}



	//  Test 2: Copy using interrupts.

	//  Enable LO interrupt level for the complete transaction and
	//  error flag on DMA channel 0.
	DMA_SetIntLevel( Channel, DMA_CH_TRNINTLVL_LO_gc, DMA_CH_ERRINTLVL_LO_gc );
    PMIC.CTRL |= PMIC_LOLVLEN_bm;

    // Fill memory block A with example data again.
    for ( index = 0; index < MEM_BLOCK_SIZE; ++index )
    {
        memoryBlockA[index] = 0xff - ( (uint8_t) index & 0xff );
    }
    // Set intDone to false to know when it has been executed.
    gInterruptDone = false;
	gTransferError = false;

	// Clear pending interrupts
	DMA.INTFLAGS |= DMA_CH0TRNIF_bm|DMA_CH0ERRIF_bm;

    // Enable interrupts
    sei();


    // Copy data using channel 0.
    AsyncMemCopy(  memoryBlockA,
	                memoryBlockB,
	                MEM_BLOCK_SIZE,
	                Channel);

    do
    {
        // Do something here while waiting for the
        // interrupt routine to signal completion.
		LEDPORT.OUTTGL = 0x10;
		nop();
    } while ( gInterruptDone == false );


	if ( gTransferError )
    {
        // Signal Error on LEDs by turning them all on
        LEDPORT.OUT = 0x00;

        do
        {
			nop();
            // Completed with failure
		} while (1);
	} else
    {
        // Signal success on one LED
        LEDPORT.OUT = ~0x02;
	    do
        {
			nop();
            // Completed with no error
        } while (1);
    }
}


// DMA CH0 Interrupt service routine. Clear interrupt flags after check
ISR(DMA_CH0_vect)
{
    if (DMA.INTFLAGS & DMA_CH0TRNIF_bm)
    {
        DMA.INTFLAGS |= DMA_CH0TRNIF_bm;
        gTransferError = false;
    } else
    {
        DMA.INTFLAGS |= DMA_CH0ERRIF_bm;
        gTransferError = true;
    }
    gInterruptDone = true;
}

/**************************************************************************
 *
 *
 *                            Task2: Recorder
 *
 *
 **************************************************************************/

#define SWITCHPORT PORTF

#define SAMPLE_COUNT 300 // How many switch state samples to store.
uint8_t samples[SAMPLE_COUNT]; // Store switch state samples here.

// The DMA channel is triggered by the Timer Overflow Interrupt Flag,
// and will be continuously re-triggered as long as the flag is set.
// We only want one single data transfer triggered for every overflow.
// Therefore we need to enable interrupts in order to automatically
// clear the Timer Overflow Interrupt Flag when the interrupt handler is
// called. However, we do not have to do anything inside the handler itself.
// That is why we need an empty handler below.


ISR(TCC0_OVF_vect)
{
    nop();
}


// Setup parameters for the channel that read switch states. The channel is
// configured to reload original destination address when the data block is finished.
// Single-shot is enabled so that only one byte gets copyed for every timer overflow.
void SetupReadChannel( DMA_CH_t * dmaChannel )
{
	DMA_SetupBlock(
                    dmaChannel,
			        (void const *) &(SWITCHPORT.IN),
                    DMA_CH_SRCRELOAD_NONE_gc,
                    DMA_CH_SRCDIR_FIXED_gc,
			        samples,
                    DMA_CH_DESTRELOAD_BLOCK_gc,
                    DMA_CH_DESTDIR_INC_gc,
			        SAMPLE_COUNT,
                    DMA_CH_BURSTLEN_1BYTE_gc,
                    0,
                    false
                );

    DMA_EnableSingleShot( dmaChannel );
	DMA_SetTriggerSource( dmaChannel, DMA_CH_TRIGSRC_TCC0_OVF_gc ); // TCC0 Overflow as trigger source.
}


// Setup parameters for the channel that writes LED values. The channel is
// configured to reload orignal source address when the data block is finished.
// It is also configured to repeat the block copy operation indefinitely, by setting
// the Repeat Counter to 0 initially.
// Single-shot is enabled so that only one byte gets copyed for every timer overflow.
void SetupWriteChannel( DMA_CH_t * dmaChannel  )
{
	DMA_SetupBlock(
                    dmaChannel,
			        samples,
                    DMA_CH_SRCRELOAD_BLOCK_gc,
                    DMA_CH_SRCDIR_INC_gc,
                    (void *) &(LEDPORT.OUT),
                    DMA_CH_DESTRELOAD_NONE_gc,
                    DMA_CH_DESTDIR_FIXED_gc,
                    SAMPLE_COUNT,
                    DMA_CH_BURSTLEN_1BYTE_gc,
                    0,
                    true
                );
	DMA_EnableSingleShot( dmaChannel );
	DMA_SetTriggerSource( dmaChannel, DMA_CH_TRIGSRC_TCC0_OVF_gc ); // TCC0 Overflow as trigger source.
}


// Setup a timer to give overflow interrupts at the desired sample rate.
// The default setup is to have no prescaling and full period time, which
// means a sample rate of 2MHz / 2^16 = approx. 30Hz.
void SetupSampleTimer( void )
{
	TCC0.CTRLA = TC_CLKSEL_DIV1_gc;
	TCC0.PER = 0xffff;
	TCC0.INTCTRLA = TC_OVFINTLVL_LO_gc;
	PMIC.CTRL |= PMIC_LOLVLEN_bm;
}


// Provide a visible indication to the user that the application is ready
// to record, finished with recording etc.
void BlinkLEDs( void )
{
	LEDPORT.OUT = 0x00;
	delay_us( 1000000 ); // Wait one second
	LEDPORT.OUT = 0xff;
}


int dmaTestMain2( void )
{
    // The DMA channel to use for reading switch state.
    DMA_CH_t * ReadChannel = &DMA.CH0;
    // The DMA channel to use for writing data to LEDs.
    DMA_CH_t * WriteChannel = &DMA.CH1;

    // Setup IO ports.
	PORTCFG.MPCMASK = 0xff;
	SWITCHPORT.PIN0CTRL = PORT_OPC_PULLUP_gc;

    SWITCHPORT.DIR = 0x00;
    LEDPORT.DIR = 0xff;

    // Prepare DMA and sample timer tick.
    DMA_Enable();
    SetupReadChannel(ReadChannel);
    SetupWriteChannel(WriteChannel);
    SetupSampleTimer();
    // Enable interrupts;
    sei();

    while(1)
    {
		// Blink, wait for user to press a switch, blink again.
		BlinkLEDs();
		do {} while (SWITCHPORT.IN == 0xff);
		BlinkLEDs();

		// Record switch state until buffer is full, then blink LEDs.
		DMA_EnableChannel( ReadChannel );
		DMA_ReturnStatus_blocking( ReadChannel );
		BlinkLEDs();

		// Wait for user to press a switch, then blink LEDs.
		do {} while (SWITCHPORT.IN == 0xff);
		BlinkLEDs();

		// Play back switch samples on LEDs over and over again, until
		// user presses a switch.
		DMA_EnableChannel( WriteChannel );
		do {} while (SWITCHPORT.IN == 0xff);
		DMA_DisableChannel( WriteChannel );

		// Setup write channel again, in case (very likely) the playback
		// was interrupted by the user in the middle of the buffer, which would
		// leave the channel in an unknown state.
		SetupWriteChannel(WriteChannel);
	}
}

