

/****************************************************************************
 *
 *                           -= XMEGA USART Task 1 =-
 *
 *        Task 1: USART in polling mode
 *
 *        The application sends and receives TEST_CHARS bytes, in a loop-
 *        back connection. Latch the TX(PD3) to RX(PD2) and start having fun.
 *
 *                                 -= XMEGA =-
 *
 *****************************************************************************/

#include <avr/io.h>
#include "usart_driver.h"
#include "avr_compiler.h"
#include "dma_driver.h"
#include <stdbool.h>

#define BSCALE_VALUE  0
#define BSEL_VALUE   12

#define TEST_CHARS  100

#define LEDPORT PORTE

#define USART USARTD0
#define USART_PORT PORTD

/*  Example USART application.
 *
 *  This example configures USARTD0 with the parameters:
 *      - 8 bit character size
 *      - No parity
 *      - 1 stop bit
 *      - 9600 Baud
 *
 *  This function will send the values 0-100 and test if the received data is
 *  equal to the sent data. The code can be tested by connecting PD3 to PD2. If
 *  the variable 'TransferError' is false and LEDs are lit up at the end of
 *  the function, the bytes have been successfully sent and received by USARTD0.
 */

int usartTestMain1(void)
{

	LEDPORT.DIR = 0xFF;
	LEDPORT.OUT = 0xFF;

	uint8_t i;

	char Tx_Buf[TEST_CHARS];
	char Rx_Buf[TEST_CHARS];

	bool TransferError = false;

	USART_PORT.DIRSET   = PIN3_bm;   // Pin 3 (TXD0) as output.
	USART_PORT.DIRCLR   = PIN2_bm;   // Pin 2 (RXD0) as input.

	// USARTD0, 8 Data bits, No Parity, 1 Stop bit
	USART.CTRLC = (uint8_t) USART_CHSIZE_8BIT_gc | USART_PMODE_DISABLED_gc | false;

   /* Target: Internal RC 2MHz (default)
    * Example (9600bps)  :   - I/O clock = 2MHz
 	*                        - 9600bps can be acheived by 9600bps / 2^0
 	*                        - UBRR = 2MHz/(16*9600)-1 = 12.02
 	*                        - ScaleFactor = 0
	*/

	USART.BAUDCTRLA = BSEL_VALUE;

	// Enable both RX and TX
	USART.CTRLB |= USART_RXEN_bm;
	USART.CTRLB |= USART_TXEN_bm;

	i = 0;

    while(i != TEST_CHARS)
    {
        Tx_Buf[i] = i;
        // Transmit a character, first wait for the previous to be sent
        while( (USART.STATUS & USART_DREIF_bm) == 0 ) {}
       	  // Now, transmit the character
        USART.DATA = Tx_Buf[i];

        // Wait until the data is received
        while( (USART.STATUS & USART_RXCIF_bm) == 0 ) {}

        // Read out the received data
        Rx_Buf[i] = USART.DATA;

        if(Rx_Buf[i] != Tx_Buf[i])
        {
            TransferError = true;
        }

        i++;
    }

    while(1)
    {
  	    if(TransferError)
	    {
           // No light
            LEDPORT.OUT = 0xFF;
	    }
	    else
	    {
            // Light
            LEDPORT.OUT = 0x00;
	    }
    }
}



/****************************************************************************
 *
 *                           -= XMEGA =-
 *
 *        Task 2: USART Using driver and ring buffer in interrupt mode
 *
 *        The application sends NUM_BYTES bytes, then waits for NUM_BYTES
 *        bytes to be received. The transmission is done using one USART,
 *        USARTD0 (TX/PD3 and RX/PD2). The code can be tested by
 *        connecting PD2 to PD3. If the variable 'TransferError' is false
 *        at the end of the file, the NUM_BYTES bytes have been successfully
 *        transmitted and received. The task will also give insight in the
 *        ring buffer that is implemented in the driver.
 *
 *                           -= XMEGA =-
 *
 *****************************************************************************/

// Number of bytes to send in this task
#define NUM_BYTES  3

// USART data struct used in task
USART_data_t USART_data;

// Test data to send
uint8_t sendArray[NUM_BYTES] = {0x55, 0xaa, 0xf0};

// Array to put received data in
uint8_t receiveArray[NUM_BYTES];

// Variable to save errors
bool TransferError = false;


//
//  This code in this task configures USARTD0 with these parameters:
//      - 8 bit character size
//      - No parity
//      - 1 stop bit
//      - 9600 Baud

int usartTestMain2(void)
{
    // counter variable
    uint8_t i;

	// LEDS as output
	LEDPORT.DIR=0xff;
	LEDPORT.OUT=0xff;

    // PD3 (TX) as output
    USART_PORT.DIRSET   = PIN3_bm;
    // PD2 (RX) as input
    USART_PORT.DIRCLR   = PIN2_bm;

    // Use USARTD0 and initialize buffers
    USART_InterruptDriver_Initialize(&USART_data, &USART, USART_DREINTLVL_LO_gc);

    // USARTD0, 8 Data bits, No Parity, 1 Stop bit
    USART_Format_Set(USART_data.usart, USART_CHSIZE_8BIT_gc,
                     USART_PMODE_DISABLED_gc, false);

    // Enable RXC interrupt
	USART_RxdInterruptLevel_Set(USART_data.usart, USART_RXCINTLVL_LO_gc);

	// Set Baudrate to 9600 bps:
	// Use the default I/O clock frequency that is 2 MHz.
	// Do not use the baudrate scale factor
	//
	// Baudrate select = (1/(16*(((I/O clock frequency)/Baudrate)-1)
	//                 = 12

	USART_Baudrate_Set(&USART, 12 , 0);

	/* Enable both RX and TX. */
	USART_Rx_Enable(USART_data.usart);
	USART_Tx_Enable(USART_data.usart);

	// Enable PMIC interrupt level low
	PMIC.CTRL |= PMIC_LOLVLEX_bm;

	// Enable global interrupts
	sei();

	// Send sendArray
	i = 0;
	while (i < NUM_BYTES)
    {
		bool byteToBuffer;

		byteToBuffer = USART_TXBuffer_PutByte(&USART_data, sendArray[i]);
        if(byteToBuffer)
        {
			i++;
		}
	}

	// Fetch received data as it is received
	i = 0;
    while (i < NUM_BYTES)
    {
        if (USART_RXBufferData_Available(&USART_data))
        {
			receiveArray[i] = USART_RXBuffer_GetByte(&USART_data);
			i++;
		}
	}

	// Test to see if sent data equals received data
	// Assume no Errors first
	TransferError = false;
	for(i = 0; i < NUM_BYTES; i++)
    {
		// Check that each element is received correctly
		if (receiveArray[i] != sendArray[i])
        {
			TransferError = true;
		}
	}

	// Show status of the transfer
    while(1)
    {
  	    if(TransferError)
	    {
           // No light
            LEDPORT.OUT = 0xFF;
	    }
	    else
	    {
            // Light
            LEDPORT.OUT = 0x00;
	    }
    }

}


//  Receive complete interrupt service routine.
//  Calls the common receive complete handler with pointer to the correct USART
//  as argument.
/*
ISR( USARTD0_RXC_vect )
{
    USART_RXComplete( &USART_data );
}


//  Data register empty  interrupt service routine.
//  Calls the common data register empty complete handler with pointer to the
//  correct USART as argument.

ISR( USARTD0_DRE_vect )
{
    USART_DataRegEmpty( &USART_data );
}*/


/******************************************************************************
 *                      Task 3: USART using DMAC
 *
 * This task will show how to set up the USART with the DMA controller.
 * SetupTransmitChannel and SetupReceiveChannel are using the dma_driver.c
 * The driver provides a fast setup of the DMAC. SetupTransmitChannel
 * is setting up the Tx_Buf buffer to feed the USART data transfer register.
 * At the receiver, another DMAC channel will receive the bytes and
 * start filling the Rx_Buf buffer.
 *****************************************************************************/

// 2MHz internal RC oscillator, 9600Baud:
#define BSEL_VALUE   12
#define BSCALE_VALUE  0

//#define TEST_CHARS  20

#define DMA_TX_Channel &DMA.CH0
#define DMA_RX_Channel &DMA.CH1

// Buffers to send/receive
static char Tx_Buf[TEST_CHARS];
static char Rx_Buf[TEST_CHARS];

void SetupTransmitChannel( void )
{
   DMA_SetupBlock(
                    DMA_TX_Channel,
                    Tx_Buf,
                    DMA_CH_SRCRELOAD_NONE_gc,
                    DMA_CH_SRCDIR_INC_gc,
                    (void *) &USARTD0.DATA,
                    DMA_CH_DESTRELOAD_NONE_gc,
                    DMA_CH_DESTDIR_FIXED_gc,
                    TEST_CHARS,
                    DMA_CH_BURSTLEN_1BYTE_gc,
                    0, // Perform once
                    false
                   );
    DMA_EnableSingleShot( DMA_TX_Channel);
    // USART Trigger source, Data Register Empty, 0x4C
    DMA_SetTriggerSource( DMA_TX_Channel, DMA_CH_TRIGSRC_USARTD0_DRE_gc);
}

void SetupReceiveChannel( void )
{
    DMA_SetupBlock(
                    DMA_RX_Channel,
                    (void *) &USARTD0.DATA,
                    DMA_CH_SRCRELOAD_NONE_gc,
                    DMA_CH_SRCDIR_FIXED_gc,
                    Rx_Buf,
                    DMA_CH_DESTRELOAD_NONE_gc,
                    DMA_CH_DESTDIR_INC_gc,
                    TEST_CHARS,
                    DMA_CH_BURSTLEN_1BYTE_gc,
                    0, // Perform once
                    false
                   );

    DMA_EnableSingleShot( DMA_RX_Channel );
    // USART Trigger source, Receive complete
    DMA_SetTriggerSource( DMA_RX_Channel, DMA_CH_TRIGSRC_USARTD0_RXC_gc);
}

void SetupUsart( void )
{

    // Place a jumper to connect Pin3 and Pin2
    // PD3 (TXD0) as output
    PORTD.DIRSET   = PIN3_bm; // (TXD0) as output.

    // PD2 (RXD0) as input
    PORTD.DIRCLR   = PIN2_bm; // (RXD0) as input.
    // USARTD0; 8 Data bits, No Parity, 1 Stop bit
    USARTD0.CTRLC = (uint8_t) USART_CHSIZE_8BIT_gc | USART_PMODE_DISABLED_gc | false;

    // Enable both TX on PORTC and RX on PORTD
    USARTD0.CTRLB |= USART_TXEN_bm | USART_RXEN_bm;

    // Target: Internal RC 2MHz (default), 9600baud
    USARTD0.BAUDCTRLA = BSEL_VALUE;
}

int usartTestMain3(void)
{
    uint16_t i = 0;
    bool success;

    SetupUsart();

    // Initialize LEDs
    LEDPORT.DIR = 0xFF;
    LEDPORT.OUT = 0xFF;

    // Initialise DMAC
    DMA_Enable();
    SetupTransmitChannel();
    SetupReceiveChannel();

    // Initialise transmit buffer
    for(i=0; i < TEST_CHARS; i++)
    {
        // filling in a,b,c,...,t
        Tx_Buf[i] = 'a'+ i;
    }

    // Assume that everything is OK.
    success = true;

    // The receiving DMA channel will wait for characters
    // and write them to Rx_Buf
    DMA_EnableChannel( DMA_RX_Channel );

    // Start sending the data bits from Tx_Buf on the
    // transmit dma channel.
    // Note that DRE is considered empty at this point,
    // thus triggering the initial transfer.
    DMA_EnableChannel( DMA_TX_Channel );

    // While DMA is handling the USART, we can do something else
	// In this case something as simple as counting upwards
    i=0;
	while ( !DMA_ReturnStatus_non_blocking( DMA_RX_Channel ))
	{
		i++;
	}

	// DMA has completed
	// Show the lower bits of the counter on the LEDs
	LEDPORT.OUT = i & 0xFF;

	// Verify the received characters
    for(i=0; i<TEST_CHARS; i++)
    {
        if(Tx_Buf[i] != Rx_Buf[i])
        {
            // Something went wrong..
            success = false;
        }
    }

    while(1)
    {

        if(success)
        {
            if(i > TEST_CHARS)
            {
                i = 0;
            }
            else
            {
                i++;
            }
            // Note that the repeat is on and
            // that the result will be rewritten
            // Output the result to the LEDs
            _delay_ms(200);
            LEDPORT.OUT = Rx_Buf[i];
        }
        else
        {
            // No lights means failure!
            LEDPORT.OUT = 0xFF;
        }
    }
}

