/*
 * encoder_counter.c
 *
 *  Created on: Nov 28, 2010
 *      Author: hsrmqp
 */

#include "encoder_counter.h"
#include "utils/serialize.h"

static const encoder_counter_setup_t* sSetup= 0;

static int32_t lastEncoderCount[NUM_MOTORS]= {0, 0}; //the previous encoder count

//Static Prototypes
static int32_t getEncCntrReg(MOTOR_t motor, uint8_t reg);
static void setEncCntrReg(MOTOR_t motor, uint8_t reg, uint8_t val);
static void clrEncCntrReg(uint8_t encoder, uint8_t reg);
static uint8_t formEncCntrHeader(uint8_t opCode, uint8_t reg);
static bool SPI_SendBytes(uint8_t* tx, uint8_t* rx, uint8_t numBytes, uint8_t sensor);

void init_encoder_counter(encoder_counter_setup_t* setup) {
   sSetup= setup;
   //send out initial SPI communications to setup the chips
   //set up with 4x quadrature count free running mode and all
   //unused flags/index stuff disabled
   setEncCntrReg(MOTOR0, ENC_CNTR_MDR0, 0x03);
   setEncCntrReg(MOTOR0, ENC_CNTR_MDR1, 0x00);

   setEncCntrReg(MOTOR1, ENC_CNTR_MDR0, 0x03);
   setEncCntrReg(MOTOR1, ENC_CNTR_MDR1, 0x00);

   clrEncCntrReg(MOTOR0, ENC_CNTR_CNTR);
   clrEncCntrReg(MOTOR1, ENC_CNTR_CNTR);

}

/**
 * Get the difference in encoder counts since the last reading.
 * @param motor The motor to retrieve an encoder count off of (MOTOR0 or MOTOR1)
 * @return The difference in encoder counts.
 */
int32_t getEncDeltaCount(MOTOR_t motor) {
   int32_t retVal; //delta count
   int32_t currentCount= getEncRunningCount(motor); //current count

   retVal= currentCount - lastEncoderCount[motor];
   lastEncoderCount[motor]= currentCount; //update last count
   return retVal;
}


/**
 * Read the encoder count off of a given motor.  This will return a signed
 * number representing the total running count of the encoder since startup.
 * @param motor The motor to retrieve an encoder count off of (MOTOR0 or MOTOR1)
 * @return The 32 bit signed encoder count.
 */
int32_t getEncRunningCount(MOTOR_t motor) {
   //get the counter register
   return getEncCntrReg(motor, ENC_CNTR_CNTR);
}

/**
 * Read a register in the encoder counter chip.  The return value is a
 * signed 32 bit integer because the encoder tick counting registers are 32 bit.
 * @param encoder The which encoder to read off
 * @param reg The register to read.
 * @return The value stored in the register.
 */
static int32_t getEncCntrReg(uint8_t encoder, uint8_t reg) {
   //formulate a packet
   uint8_t packet[5];
   packet[0]= formEncCntrHeader(ENC_CNTR_RD, reg); //other elements of packet not used

   uint8_t regValue[5];
   uint8_t numBytes;
   //set the number of bytes to transcive.  MDR0, MDR1, and STR are 1 byte
   //and all others are 4 bytes.  numBytes is 1 greater than the size of the
   //register to account
   if((reg == ENC_CNTR_MDR0) || (reg == ENC_CNTR_MDR1) ||
      (reg == ENC_CNTR_STR)) {
      numBytes= 2;
   }
   else {
      numBytes= 5;
   }
   SPI_SendBytes(packet, regValue, numBytes, encoder);

   //convert the array of bytes to 32 bit number
   uint32_t retVal;
   deSerializeInt32(&retVal, &regValue[1]);

   return (int32_t)retVal;
}
/**
 * Set a register in the encoder counter chip
 * @param encoder Which encoder to set.
 * @param reg The register to set.
 * @param val The value to set the register to.
 */
static void setEncCntrReg(uint8_t encoder, uint8_t reg, uint8_t val) {
   //formulate a packet
   uint8_t packet[2];
   packet[0]= formEncCntrHeader(ENC_CNTR_WR, reg);
   packet[1]= val;

   uint8_t dummyRcv[2];

   //do spi communication
   SPI_SendBytes(packet, dummyRcv, 2, encoder);

}

/**
 * Clear a register in the encoder counter chip
 * @param encoder Which encoder to clear.
 * @param reg The register to clear.
 */
static void clrEncCntrReg(uint8_t encoder, uint8_t reg) {
   //formulate a packet
   uint8_t packet[1];
   packet[0]= formEncCntrHeader(ENC_CNTR_CLR, reg);

   uint8_t dummyRcv[1];

   //do spi communication
   SPI_SendBytes(packet, dummyRcv, 1, encoder);
}

/**
 * Form the header for an encoder counter SPI transmission.
 * @param opCode The opcode for the packet (read/write/clear/load)
 * @param reg The register to set.
 * @return The 1 byte header to perform opCode on register reg.
 */
static uint8_t formEncCntrHeader(uint8_t opCode, uint8_t reg) {
   uint8_t retVal= 0; //return value

   retVal|= opCode<<6;
   retVal|= reg<<3;

   return retVal;
}

/**
 * Send a series of bytes out to a given mouse sensor over the spi line.
 * @param tx Pointer to transmit byte array
 * @param rx Pointer to receive byte array (must be already allocated)
 * @param numBytes The number of bytes to send
 * @param sensor The sensor to send to
 * @return True on success
 */
static bool SPI_SendBytes(uint8_t* tx, uint8_t* rx, uint8_t numBytes, uint8_t sensor) {
   //First make sure SPI is setup properly
   sSetup->SPIinitFunc(sSetup->spiDevice);

   //create packet
   SPI_DataPacket_t spiPacket;
   SPI_MasterCreateDataPacket(&spiPacket, tx, rx, numBytes,
                              sSetup->ssPort[sensor], sSetup->ssPinMask[sensor]);
   //send packet
   return SPI_MasterTransceivePacket(sSetup->spiDevice, &spiPacket);
}
