/*
 * localization.c
 *
 *  Created on: Dec 18, 2010
 *      Author: hsrmqp
 */

#include "localization.h"
#include "encoder_counter.h"
#include "mouse_sensor.h"
#include "mechanical_properties.h"
#include <math.h>
#include "utils/system_time.h"


volatile static pose_t sPose; //holds the most updated position of the robot
static float sRightWheelVelocity; //in cm/s
static float sLeftWheelVelocity;  //in cm/s

static uint64_t previousSysTime= 0;  //used for tracking time difference
static uint64_t sDeltaTime_us;       //time difference in us
static float sDeltaTime_s;           //time difference in seconds


static int32_t sDeltaEnc0;
static int32_t sDeltaEnc1;

static float total_distance= 0; //the total distance driven by the robot
static uint16_t std_deviation= 0; //the expected deviation from the calculated pose in cm

//Static prototypes
static void readSensors(void);
static void updateWheelVelocity();
static void timeStamp(void);

void updateState(void) {
   timeStamp();
   readSensors();
   updatePose();
   updateWheelVelocity();
}



/**
 * Get the x,y position and orientation of the robot.
 * @param p The location to store the pose.
 */
void getPose(pose_t* p) {
   uint8_t err;
   OSMutexPend(current_pose_mutex, 0, &err);
   p->x= sPose.x;
   p->y= sPose.y;
   p->theta= sPose.theta;
   OSMutexPost(current_pose_mutex);

}

/**
 * Get the x,y position and orientation of the robot as well as the expected
 * deviation from the pose.
 * @param p The location to store the pose.
 * @param deviation The expected deviation from the specified pose.  Passing a
 * null pointer here causes the function to ignore this argument.
 */
void getPose_dev(pose_t* p, uint16_t* deviation) {
   uint8_t err;
   OSMutexPend(current_pose_mutex, 0, &err);
   p->x= sPose.x;
   p->y= sPose.y;
   p->theta= sPose.theta;
   if(deviation != 0) {
      *deviation= std_deviation;
   }
   OSMutexPost(current_pose_mutex);

}

/**
 * Get the velocity (in cm/s) of the given wheel.
 * 0= left
 * 1= right
 * NOTE: This function is not thread safe.
 */
float getWheelVelocity(uint8_t wheel) {
   switch(wheel) {
   case 0:
      return sLeftWheelVelocity;
      break;
   case 1:
      return sRightWheelVelocity;
      break;
   default:
      return 0;
      break;
   }
}

/**
 * Get the delta time stamp in microseconds.
 * NOTE: This function is not thread safe.
 * @return The number of microseconds since the previous timestamp.
 */
uint64_t getTimeStamp_us(void) {
   return sDeltaTime_us;
}

/**
 * Get the delta time stamp in seconds.
 * NOTE: This function is not thread safe.
 * @return The number of seconds since the previous timestamp.
 */
float getTimeStamp_s(void) {
   return sDeltaTime_s;
}

/**
 * Manually set the current pose of the robot.  This should be used with care.
 * @param p The pointer to a pose struct containing the values to set.
 */
void setPose(pose_t* p) {
   uint8_t err;
   OSMutexPend(current_pose_mutex, 0, &err);
   sPose.x= p->x;
   sPose.y= p->y;
   sPose.theta= p->theta;
   OSMutexPost(current_pose_mutex);
}

/**
 * Update the position of the robot.
 * NOTE: This function is not thread safe.
 * @param deltaTime_s The time difference between this running and the previous
 *                    one in seconds.
 */
void updatePose(void) {

   float RForward_enc;   //right wheel forward progress seen by encoder
   float LForward_enc;   //left wheel forward progress seen by encoder
   float deltaForward_enc;    //Forward progress of the robot as seen by the encoders
   float deltaTheta_enc; //Angular rotation as of the robot as seen by the encoders

   float deltaForward;  //Forward progress of robot
   float deltaTheta;    //Change in robot angle

   //get encoder readings
   //one of the encoder readings will probably be backwards (negative in value) so
   //it should be flipped so "forwards" is the same direction for both in future calculations
   LForward_enc= ((float)sDeltaEnc0/Q_COUNTS_PER_REV)*2*M_PI*WHEEL_RADIUS*ENC0_POLARITY;
   RForward_enc= ((float)sDeltaEnc1/Q_COUNTS_PER_REV)*2*M_PI*WHEEL_RADIUS*ENC1_POLARITY;

   //calculate how far the robot has moved forwards and how far
   //the robot has turned according to the encoders
   deltaForward_enc= (RForward_enc + LForward_enc) / 2;

   deltaTheta_enc= (RForward_enc - LForward_enc)/(WHEEL_BASE);

   //average encoders and mouse sensors
   deltaForward= deltaForward_enc;

   deltaTheta= deltaTheta_enc;

   //lock mutex
   uint8_t err;
   OSMutexPend(current_pose_mutex, 0, &err);
   sPose.theta+= deltaTheta;
   //keep theta between -pi and pi
   while(sPose.theta < -M_PI) {
      sPose.theta+= 2*M_PI;
   }
   while(sPose.theta > M_PI) {
      sPose.theta-= 2*M_PI;
   }

   //update x and y position
   float dx= deltaForward*cos(sPose.theta);
   float dy= deltaForward*sin(sPose.theta);
   sPose.x+= dx;
   sPose.y+= dy;

   total_distance+= hypot(dx, dy);
   std_deviation= total_distance*DRIFT_PER_CM;
   //unlock mutex
   OSMutexPost(current_pose_mutex);

}

/**
 * Update the sensor readings.  This should be called after obtaining a new timestamp
 */
static void readSensors(void) {

   sDeltaEnc0= getEncDeltaCount(MOTOR0);
   sDeltaEnc1= getEncDeltaCount(MOTOR1);

}

/**
 * Update the wheel velocities.
 */
static void updateWheelVelocity(void) {
   sRightWheelVelocity= ((float)sDeltaEnc1/Q_COUNTS_PER_REV)*2*M_PI*WHEEL_RADIUS*ENC1_POLARITY/sDeltaTime_s;
   sLeftWheelVelocity= ((float)sDeltaEnc0/(float)Q_COUNTS_PER_REV)*2*M_PI*WHEEL_RADIUS*ENC0_POLARITY/sDeltaTime_s;

}

/**
 * Get a time stamp for the sensor readings
 */
static void timeStamp(void) {
   //get a timestamp
   uint64_t currentTime= getTime_us();
   sDeltaTime_us= (currentTime - previousSysTime);
   sDeltaTime_s= (float)sDeltaTime_us/US_PER_S;
   previousSysTime= currentTime;
}
