/*
 * locomotion.c
 *
 *  Created on: Dec 19, 2010
 *      Author: hsrmqp
 */

#include "locomotion.h"
#include "localization.h"
#include "includes.h"
#include "hbridge.h"
#include "mechanical_properties.h"
#include "utils/pid.h"
#include <math.h>
#include "utils/simplified_math.h"

#include "utils/system_time.h"
#include "stdio.h"

static float sRightWheelVelSetpoint; //setpoint for the right wheel velocity in cm/s
static float sLeftWheelVelSetpoint;  //setpoint for the left wheel velocity in cm/s
volatile static pose_t setEndpoint;   //The position we're driving towards
static pid_f_t leftMotorPID;
static pid_f_t rightMotorPID;
static bool crashFlag= false;

static void setRobotVelocitySetpoint(float v, float omega);
static v_omega_pair_t calculatePath(float k_row, float k_alpha, float k_beta,
                          pose_t* current_pose, pose_t* set_pose);
static bool checkEndpointArrival(pose_t* current_pose, pose_t* set_pose,
                                 float translationTolerance, float angleTolerance);

void locomotion_task(void* pd) {
   uint8_t err;
   float v;     //set forward velocity
   float omega; //set angular velocity
   v_omega_pair_t pathVelocities; //used for return value of calculatePath
   pose_t temp_endpoint; //temp storage for the current endpoint
   pose_t currentPoint;  //Where we are now
   setupPID_f(&leftMotorPID, 0, K_P, K_I, K_D);
   setupPID_f(&rightMotorPID, 0, K_P, K_I, K_D);
   //Hardcode in starting location
   currentPoint.theta= M_PI/2;
   currentPoint.x= 220;
   currentPoint.y= 0;
   setPose(&currentPoint);
   //Manually set endpoint to be where we are now
   //so we don't start driving right away.
   temp_endpoint= currentPoint;
   temp_endpoint.theta= NAN;
   setEndpointPose(&temp_endpoint);

   //all tasks must sit in a loop forever
   for(;;) {
      updateState();
      //update current pose
      getPose(&currentPoint);
      //get current endpoint
      getEndpointPose(&temp_endpoint);
      if(checkEndpointArrival(&currentPoint, &temp_endpoint, POSITIONAL_TOLERANCE, ANGULAR_TOLERANCE)) {
         v= 0;
         omega= 0;
         //post a flag if we've arrived at endpoint
         OSFlagPost(locomotion_driving_flag, DRIVING_FLAG, OS_FLAG_SET, &err);
      }
      else {
         //calculate the robot forward/rotational speed
         pathVelocities= calculatePath(K_ROW, K_ALPHA, K_BETA, &currentPoint, &temp_endpoint);
         v= pathVelocities.v;
         omega= pathVelocities.omega;
      }
      //cap v and omega to stop setpoints from being set too high
      if(fabs(v) > ROBOT_MAX_FORWARD_VELOCITY) {
         if( v < 0)
            v= -ROBOT_MAX_FORWARD_VELOCITY;
         else
            v= ROBOT_MAX_FORWARD_VELOCITY;
      }
      if(fabs(omega) > ROBOT_MAX_ANGULAR_VELOCITY) {
         if( omega < 0) {
            omega= -ROBOT_MAX_ANGULAR_VELOCITY;
            v= 0;
         }
         else {
            omega= ROBOT_MAX_ANGULAR_VELOCITY;
            v= 0;
         }
      }
      //use resulting v/omega from calculatePath to set v and omega setpoints
      setRobotVelocitySetpoint(v, omega);
      //special case for 0 setpoints
      if(sLeftWheelVelSetpoint == 0 && sRightWheelVelSetpoint == 0) {
         setMotorPWM(0, 0);
         setMotorPWM(1, 0);
      }
      else {
         //run pid
         float deltaLeft;
         float deltaRight;
         setErrorPID_f(&leftMotorPID, sLeftWheelVelSetpoint - getWheelVelocity(0));
         setErrorPID_f(&rightMotorPID, sRightWheelVelSetpoint - getWheelVelocity(1));
         deltaLeft= runPID_f(&leftMotorPID, getTimeStamp_s());
         deltaRight= runPID_f(&rightMotorPID, getTimeStamp_s());
         //set motors based on pid
         setMotorPWM(0, getMotorPWM(0) + deltaLeft*MOTOR0_POLARITY);
         setMotorPWM(1, getMotorPWM(1) + deltaRight*MOTOR1_POLARITY);
      }
      //Delay to run at ~100Hz
      OSTimeDly(1);
   }
}

/**
 * Drive to a given pose.
 * @param p The pose to drive to.
 * @return true If we arrived at the destination, false otherwise
 */
bool driveToPose(pose_t* p) {
   uint8_t err;
   setEndpointPose(p);
   //wait until we arrive at the desired location
   OS_FLAGS flags;
   flags= OSFlagPend(locomotion_driving_flag, DRIVING_FLAG + COLLISION_FLAG, OS_FLAG_WAIT_SET_ANY + OS_FLAG_CONSUME, 0, &err);
   if(flags & COLLISION_FLAG) {
      //we crashed! return false
      return false;
   }
   else {
      return true;
   }

}

/**
 * Perform an emergency stop.  This function will forcibly stop the motors and cause
 * the current path to abort.
 */
void emergencyStop(void) {
   //SHIT!!!!!
   setMotorPWM(0, 0);
   setMotorPWM(1, 0);
   //set the collision flag
   crashFlag= true;
}

/**
 * Set the forward and rotational velocity setpoints of the robot. If the values given
 * are outside of what the robot can physically achieve, it will get as close
 * as it can to the values given.  Generally speaking, it is best to try and
 * keep the values within physical limits.
 * @param v The setpoint forward velocity (in cm/s).
 * @param omega The rotational velocity (in radians/s)
 */
static void setRobotVelocitySetpoint(float v, float omega) {
   //Vr= V + omega*r
   //Vl= V - omega*r
   sRightWheelVelSetpoint= v + omega*WHEEL_BASE/2;
   sLeftWheelVelSetpoint= v - omega*WHEEL_BASE/2;
}

/**
 * Calculate the required v and omega setpoints to drive from the given current
 * position to the given endpoint within a given tolerance.  If the specified
 * ending angle in set_pose is NAN then it is assumed that the angle of arrival
 * doesn't matter.
 */
static v_omega_pair_t calculatePath(float k_row, float k_alpha, float k_beta,
                          pose_t* current_pose, pose_t* set_pose) {
   v_omega_pair_t retVal;

   //calculate the distance to the endpoint
   float row= hypot(set_pose->y - current_pose->y, set_pose->x - current_pose->x);

   //calculate alpha
   float row_angle= atan2(set_pose->y - current_pose->y, set_pose->x - current_pose->x);
   float alpha= row_angle - current_pose->theta;

   //wraparound
   if(alpha > M_PI) {
      alpha-= 2*M_PI;
   }
   else if(alpha < -M_PI) {
      alpha+= 2*M_PI;
   }
   //calculate beta
   float beta= set_pose->theta - row_angle;

   retVal.v= k_row*row;
   if(isnan(beta)) {
      retVal.omega= k_alpha*alpha;
   }
   else {
      retVal.omega= k_alpha*alpha + k_beta*beta;
   }
   return retVal;
}

/**
 * Check if the robot has arrived at a location within some tolerance.
 */
static bool checkEndpointArrival(pose_t* current_pose, pose_t* set_pose,
                                 float translationTolerance, float angleTolerance) {
   float dx= fabs(set_pose->x - current_pose->x);
   float dy= fabs(set_pose->y - current_pose->y);
   float dtheta= fabs(set_pose->theta - current_pose->theta);


   if((dx < translationTolerance) && (dy < translationTolerance) && (isnan(set_pose->theta))) {
      return true;
   }

   else if((dx < translationTolerance) && (dy < translationTolerance) && (dtheta < angleTolerance)) {
      return true;
   }

   return false;
}

/**
 * Set the setpoint pose of the robot.
 * @param newEndpoint The new setpoint pose of the robot.
 */
void setEndpointPose(pose_t* newEndpoint) {
   uint8_t err;
   OSMutexPend(setpoint_pose_mutex,0, &err);
   setEndpoint.x= newEndpoint->x;
   setEndpoint.y= newEndpoint->y;
   setEndpoint.theta= newEndpoint->theta;
   //clear flags to be safe
   OSFlagPost(locomotion_driving_flag, DRIVING_FLAG + COLLISION_FLAG, OS_FLAG_CLR, &err);
   OSMutexPost(setpoint_pose_mutex);
}

/**
 * Get the endpoint pose of the robot.
 * @param endpoint The location to store the endpoint of the robot.
 */
void getEndpointPose(pose_t* endpoint) {
   uint8_t err;
   OSMutexPend(setpoint_pose_mutex,0, &err);
   endpoint->x= setEndpoint.x;
   endpoint->y= setEndpoint.y;
   endpoint->theta= setEndpoint.theta;
   OSMutexPost(setpoint_pose_mutex);
}

