/*
 * mouse_sensor.c
 *
 *  Created on: Dec 2, 2010
 *      Author: hsrmqp
 */

#include "mouse_sensor.h"
#include <util/delay.h>
#include <stdio.h>

#define NCS_HIGH(_sensor) sSetup->ssPort[_sensor]->OUTSET= sSetup->ssPinMask[_sensor]
#define NCS_LOW(_sensor) sSetup->ssPort[_sensor]->OUTCLR= sSetup->ssPinMask[_sensor]

static const mouse_sensor_setup_t* sSetup= 0;

//Static Function Prototypes
static bool SPI_SendBytes(uint8_t* tx, uint8_t* rx, uint8_t numBytes, uint8_t sensor);
static bool writeRegister(uint8_t reg, uint8_t data, uint8_t sensor);
static bool readRegister(uint8_t reg, uint8_t* data, uint8_t sensor);
static bool readMotionBurst(uint8_t* retBytes, uint8_t numBytes, uint8_t sensor);

void init_mouse_sensor(mouse_sensor_setup_t* setup) {
   sSetup= setup;

   sSetup->SPIinitFunc(sSetup->spiDevice);
   //Do power up init.  This is stupid and miserable but oh well
   uint8_t dummyRcv[MAX_MOTION_BURST];
   for(uint8_t i= 0; i < NUM_MOUSE_SENSORS; i++) {
      //First toggle nCS
      NCS_HIGH(i);
      NCS_LOW(i);
      NCS_HIGH(i);
      writeRegister(0x3a, 0x5a, i);
      _delay_us(MOUSE_FRAME);
      writeRegister(OBSERVATION_REGISTER, 0x00, i);
      _delay_us(MOUSE_FRAME);
      readRegister(OBSERVATION_REGISTER, dummyRcv, i);
      //we can do if they aren't...
      readMotionBurst(dummyRcv, MAX_MOTION_BURST, i);
      writeRegister(0x3C, 0x27, i);
      writeRegister(0x22, 0x0A, i);
      writeRegister(0x21, 0x01, i);
      writeRegister(0x3C, 0x32, i);
      writeRegister(0x23, 0x20, i);
      writeRegister(0x3C, 0x05, i);
      writeRegister(0x37, 0xB9, i);
      //End of power on stuff
      readRegister(0x00, dummyRcv, i);

      //set at max resolution
      writeRegister(H_RESOLUTION_REGISTER, HI_RES_2000, i);

      //set at super high burn your eyes out power!
      writeRegister(LASER_CTRL0_REGISTER, CURRENT_RANGE_4_10, i);
      writeRegister(LASER_CTRL1_REGISTER, CURRENT_RANGE_4_10_COMP, i);
      writeRegister(LSRPWR_CFG0, LASER_PWR_100, i);
      writeRegister(LSRPWR_CFG1, LASER_PWR_100_COMP, i);

   }
}

/**
 * Get the displacement from the mouse sensor.
 * @param reading Pointer to a struct to store the results of the read.
 * @param sensor The sensor to read off of.
 */
void getDisplacement(mouse_sensor_reading_t* reading, uint8_t sensor) {
   uint8_t bytes[4];
   readMotionBurst(bytes, 4, sensor);
   //check motion byte?
   //get low bytes of x and y
   uint16_t x= bytes[1];
   uint16_t y= bytes[2];
   //get the high bits (XXXXYYYY)
   x|= ((uint16_t)(bytes[3] & (0xF0))) << 4;
   y|= ((uint16_t)(bytes[3] & (0x0F))) << 8;

   //do a kind of weird conversion to convert from 12 bit 2's complement
   //to a 16 bit 2's complement by checking if MSB is a 1, then inverting
   //and adding one to get the absolute value of the number, then multiplying
   //by negative one to get the sign correct.  If the MSB isn't a 1, we're all
   //set so don't do anything
   if(x & (1u << 11)) {
      x= ~x +1;
      x&= 0x0FFF; //as a result of the inversion there will be some 1's in
                  //the highest 4 bits that we don't want.
      x*=-1;
   }

   if(y & (1u << 11)) {
      y= ~y +1;
      y&= 0x0FFF;
      y*= -1;
   }

   reading->x= x;
   reading->y= y;
}

/**
 * Read from the motion burst register.
 * @param retBytes Array of bytes to store motion read burst.
 * @param numBytes The number of bytes to read off the burst.
 * @param sensor   The sensor to read off of
 * @return true on success, false on failure
 */
static bool readMotionBurst(uint8_t* retBytes, uint8_t numBytes, uint8_t sensor) {
   sSetup->SPIinitFunc(sSetup->spiDevice);
   bool retVal;

   uint8_t bytes[MAX_MOTION_BURST];
   bytes[0]= MOTION_BURST_REGISTER;
   NCS_LOW(sensor);
   retVal= SPI_SendBytes(bytes, 0, 1, sensor);
   _delay_us(MOUSE_TSRAD);
   //send dummy bytes and read back
   retVal= SPI_SendBytes(bytes, retBytes, numBytes, sensor);
   NCS_HIGH(sensor);
   _delay_us(MOUSE_TSRW); //delay for subsequent read/write

   return retVal;
}

static bool writeRegister(uint8_t reg, uint8_t data, uint8_t sensor) {
   sSetup->SPIinitFunc(sSetup->spiDevice);
   bool retVal;

   reg|= MOUSE_RW_bm; //set MSB of reg to be 1 to indicate write
   uint8_t bytes[2]= {reg, data};
   NCS_LOW(sensor);
   retVal= SPI_SendBytes(bytes, 0, 2, sensor);
   NCS_HIGH(sensor);

   _delay_us(MOUSE_TSWR); //delay for subsequent read/write

   return retVal;
}

static bool readRegister(uint8_t reg, uint8_t* data, uint8_t sensor) {
   sSetup->SPIinitFunc(sSetup->spiDevice);
   bool retVal;
   reg&= ~(MOUSE_RW_bm);
   NCS_LOW(sensor);
   retVal= SPI_SendBytes(&reg, 0, 1, sensor);
   _delay_us(MOUSE_TSRAD); //delay while ADNS looks up address
   retVal= SPI_SendBytes(&reg, data, 1, sensor);
   NCS_HIGH(sensor);
   _delay_us(MOUSE_TSRW); //delay for subsequent read/write

   return retVal;
}

/**
 * Send a series of bytes out to a given mouse sensor over the spi line.  The
 * control of the SS line must be handled independently.
 * @param tx Pointer to transmit byte array
 * @param rx Pointer to receive byte array (must be already allocated)
 *           if a null is passed recieved data will be thrown out.
 * @param numBytes The number of bytes to send
 * @param sensor The sensor to send to
 * @return True on success
 */
static bool SPI_SendBytes(uint8_t* tx, uint8_t* rx, uint8_t numBytes, uint8_t sensor) {
   //First make sure SPI is setup properly
   sSetup->SPIinitFunc(sSetup->spiDevice);

   //create packet
   SPI_DataPacket_t spiPacket;
   SPI_MasterCreateDataPacket(&spiPacket, tx, rx, numBytes,
                              0, 0);
   //send packet
   return SPI_MasterTransceivePacket(sSetup->spiDevice, &spiPacket);
}
