/*
 * scanner.c
 *
 *  Created on: Dec 19, 2010
 *      Author: hsrmqp
 */

#include "servo.h"
#include "beacondetector.h"
#include "rangefinder.h"
#include "scanner.h"
#include "mechanical_properties.h"
#include "utils/transformation_matrix.h"
#include "localization.h"
#include "math.h"

static const float irXoff[]= {IR0_X_OFFSET, IR1_X_OFFSET};
static const float irYoff[]= {IR0_Y_OFFSET, IR1_Y_OFFSET};

/**
 * This function will rotate the turret to a desired angle and perform a scan.
 * @param angle The angle to rotate the turret to before scanning. (In degrees
 *              from -90 to 90).  0 Is defined as pointing directly forwards.
 * @param irSel The IR scanner to use.
 * @param range The range to scan
 * @param o The location to store the result of the scan.
 * @return 0 If an object is seen, -1 if no object is seen.
 */
int8_t scanAngle(int8_t servoAngle, uint8_t irSel, uint8_t range, object_t* o) {
   //set the servo angle
   setServoAngle(MAIN_SERVO, servoAngle);

   //use the rangefinder
   uint8_t rawDist;
   rawDist= rangefinderGetDist(irSel);
   //if the distance is out of range, ignore it
   if(rawDist >= range) {
      return -1;
   }
   //o->properties_bm= irledDetectorGet(irSel);
   o->properties_bm= 0; //no candle/beacon detector
   //perform a frame transformation on the distance to get the coordinates
   //create matrices
   matrix_t T_0_robot; //transformation matrix from home to robot frame
   matrix_t T_robot_sensor; //transformation from robot to sensor frame
   matrix_t r_sensor; //vector to the object in the sensor's frame
   matrix_t r_global;  //the global coordinates of the object
   matrix_create(&T_0_robot, 3, 3);
   matrix_create(&T_robot_sensor, 3, 3);
   matrix_create(&r_sensor, 3, 1);
   matrix_create(&r_global, 3, 1);
   //get the robots current position
   pose_t robotPosition;
   uint16_t deviation;
   getPose_dev(&robotPosition, &deviation);
   o->stdDeviation= deviation;
   tf_matrix_set(&T_0_robot, robotPosition.x, robotPosition.y, robotPosition.theta);
   //set up the transformation from robot to IR sensor.  note that for
   //the back IR sensor the angle will be off by 180 degrees
   float irAngle= servoAngle*M_PI/180;

   if(irSel == IR_REAR) {
      irAngle-= M_PI;
      //make angle wrap around
      if(irAngle > M_PI) {
         irAngle-=2*M_PI;
      }
      else if(irAngle < -M_PI) {
         irAngle+= 2*M_PI;
      }
   }
   tf_matrix_set(&T_robot_sensor, irXoff[irSel], irYoff[irSel], irAngle);
   //set up vector
   tf_vector_set(&r_sensor, rawDist, 0);
   //now put the two transformation frames into an array to make a chain of them
   matrix_t* tfChain[3]= {&T_0_robot, &T_robot_sensor, 0};
   tf_transform_chain(tfChain, &r_sensor, &r_global);

   o->x= tf_vector_getx(&r_global);
   o->y= tf_vector_gety(&r_global);

   matrix_delete(&T_0_robot);
   matrix_delete(&T_robot_sensor);
   matrix_delete(&r_sensor);
   matrix_delete(&r_global);

   return 0;
}
