/*
 * OS_Resources.c
 *
 * This file contains the operating system resources such as the stack size,
 * semaphores, mutexes.
 *
 *  Created on: Dec 16, 2010
 *      Author: hsrmqp
 */

#include "includes.h"

/*****************************************************************************
 *                                TASK STACKS
 *****************************************************************************/
OS_STK locomotion_task_stack[LOCOMOTION_TASK_STACK_SIZE];
OS_STK rx_remote_task_stack[RX_REMOTE_TASK_STACK_SIZE];
OS_STK command_task_stack[COMMAND_TASK_STACK_SIZE];

/******************************************************************************
 *                             MEMORY PARTITIONS
 *****************************************************************************/
OS_MEM* small_mem;
OS_MEM* medium_mem;
OS_MEM* large_mem;

//allocate one extra byte than needed because uC/OS likes to have even
//memory addresses but the XMEGA doesn't actually care.  If we just
//allocate extra bytes then we can chose even/odd address
#if SMALL_MEM_NUM_BLOCKS > 0 && SMALL_MEM_BLOCK_SIZE > 0
INT8U small_mem_part[SMALL_MEM_NUM_BLOCKS][SMALL_MEM_BLOCK_SIZE+1];
#endif
#if MEDIUM_MEM_NUM_BLOCKS > 0 && MEDIUM_MEM_BLOCK_SIZE > 0
INT8U medium_mem_part[MEDIUM_MEM_NUM_BLOCKS][MEDIUM_MEM_BLOCK_SIZE+1];
#endif
#if LARGE_MEM_NUM_BLOCKS > 0 && LARGE_MEM_BLOCK_SIZE > 0
INT8U large_mem_part[LARGE_MEM_NUM_BLOCKS][LARGE_MEM_BLOCK_SIZE+1];
#endif

/*****************************************************************************
 *                                 SEMAPHORES
 *****************************************************************************/
OS_EVENT* xbee_rx_data_sem;   //wakes up rx thread when data enters the usart rx buffer

/*****************************************************************************
 *                                  MUTEXES
 *****************************************************************************/
OS_EVENT* current_pose_mutex; //controls access to the current pose of the robot
OS_EVENT* setpoint_pose_mutex; //controls access to the setpoint pose of the robot

/*****************************************************************************
 *                                 MAILBOXES
 *****************************************************************************/


/*****************************************************************************
 *                               MESSAGE QUEUES
 *****************************************************************************/
void* command_q_storage[COMMAND_Q_SIZE];
OS_EVENT* command_q;

/*****************************************************************************
 *                                   FLAGS
 *****************************************************************************/
OS_FLAG_GRP* locomotion_driving_flag;
OS_FLAG_GRP* xbee_tx_status_flag;

/**
 * Initialize all resources needed by uC/OS.  This sets up any semaphores,
 * mutexes, memory partitions, mailboxes, message queues, and event flags.
 * This function does not create the tasks themselves.
 */
void OSResourceInit() {
   uint8_t err;
   //Init memory partitions
   uint8_t* small_mem_part_aligned= (uint8_t*)small_mem_part;
   if (((INT16U)small_mem_part & (sizeof(void *) - 1u)) != 0u) {
      small_mem_part_aligned++;
   }

   uint8_t* medium_mem_part_aligned= (uint8_t*)medium_mem_part;
   if (((INT16U)medium_mem_part & (sizeof(void *) - 1u)) != 0u) {
      medium_mem_part_aligned++;
   }

   uint8_t* large_mem_part_aligned= (uint8_t*)large_mem_part;
   if (((INT16U)large_mem_part & (sizeof(void *) - 1u)) != 0u) {
      large_mem_part_aligned++;
   }
   small_mem= OSMemCreate(small_mem_part_aligned, SMALL_MEM_NUM_BLOCKS, SMALL_MEM_BLOCK_SIZE, &err);
   medium_mem= OSMemCreate(medium_mem_part_aligned, MEDIUM_MEM_NUM_BLOCKS, MEDIUM_MEM_BLOCK_SIZE, &err);
   large_mem= OSMemCreate(large_mem_part_aligned, LARGE_MEM_NUM_BLOCKS, LARGE_MEM_BLOCK_SIZE, &err);

   //Init semaphores
   xbee_rx_data_sem= OSSemCreate(0);

   //Init Mutexes
   current_pose_mutex= OSMutexCreate(CURRENT_POSE_MUTEX_PRIO, &err);
   setpoint_pose_mutex= OSMutexCreate(SETPOINT_POSE_MUTEX_PRIO, &err);

   //Init Message Queues
   command_q= OSQCreate(command_q_storage, COMMAND_Q_SIZE);

   //Init Flags
   locomotion_driving_flag= OSFlagCreate(0, &err);
   xbee_tx_status_flag= OSFlagCreate(0, &err);
}
