/*
 * matrix.c
 *
 *  Created on: Dec 9, 2010
 *      Author: hsrmqp
 */

#include "utils/matrix.h"
#include "utils/emb_malloc.h"


/**
 * Create a matrix.  The initial values in the matrix are essentially random
 * and should be set later.
 * @param m An instance of a matrix struct.  The struct must exist in memory
 * however space for the data in the matrix is allocated here.
 * @param rows The number of rows in the matrix.
 * @param cols The number of columns in the matrix.
 */
void matrix_create(matrix_t* m, uint8_t rows, uint8_t cols) {
   m->numRows= rows;
   m->numCols= cols;
   m->data= emb_malloc(rows*cols*sizeof(float));
}

/**
 * Create a matrix similar to matrix_create however this allows the user to
 * specify a valid memory location to store the data rather than using
 * malloc/free.  Because the data for this function is allocated elsewhere,
 * matrix_delete should not be called on a matrix created with
 * matrix_create_preallocated.
 * @param m An instance of a matrix struct.  The struct must exist in memory
 * however needs no initialization prior to this funciton.
 * @param rows The number of rows in the matrix.
 * @param cols The number of columns in the matrix.
 * @param data A contiguous stream of bytes large enough to hold all values in
 *             the matrix.
 */
void matirx_create_preallocated(matrix_t* m, uint8_t rows, uint8_t cols, float* data) {
   m->numRows= rows;
   m->numCols= cols;
   m->data= data;
}

/**
 * Delete A matrix.  This will set the matrix to be a 0x0 matrix and free any
 * allocated data.  This function can only delete matrices created with
 * matrix_create.
 * @param An instance of a matrix struct to delete.
 */
void matrix_delete(matrix_t* m) {
   m->numRows= 0;
   m->numCols= 0;
   emb_free(m->data);
}

/**
 * Get the element at the ith row and jth column.  Rows and collumns are
 * ZERO INDEXED.  If the user tries to access an element which is out of bounds
 * this function returns 0;
 * @param m An instance of a matrix.
 * @param i The row containing the element.
 * @param j The column containing the element.
 */
float matrix_get(matrix_t* m, uint8_t i, uint8_t j) {
   if(i>= m->numRows || j >= m->numCols) return 0;
   return m->data[(i * m->numCols) + j];
}

/**
 * Set the element at the ith row and jth column.  Rows and collums are
 * ZERO INDEXED.
 * @param m An instance of a matrix.
 * @param i The row containing the element to set.
 * @param j The column containing the element to set.
 * @param value The value to store in m[i,j]
 */
void matrix_set(matrix_t* m, uint8_t i, uint8_t j, float value) {
   m->data[(i * m->numCols) + j]= value;
}

/**
 * Add a matrix and a scalar value.  The output matrix must be the same
 * dimensions as the input matrix
 * @param m An instance of a matrix to add.
 * @param s The scalar value to add
 * @param output The location to store the output of the addition
 */
void matrix_add_scalar(matrix_t* m, float s, matrix_t* output) {
   if((m->numCols != output->numCols) || (m->numRows != output->numRows)) {
      return;
   }
   //add to each value in the matrix data array
   for(uint16_t i= 0; i < (m->numRows * m->numCols); i++) {
      output->data[i]= s + m->data[i];
   }
}

/**
 * Multiply a matrix by a scalar number.  The output matrix must be the same
 * dimensions as the input matrix
 * @param m An instance of a matrix to multiply
 * @param s The scalar number to multiply.
 */
void matrix_multiply_scalar(matrix_t* m, float s, matrix_t* output) {
   if((m->numCols != output->numCols) || (m->numRows != output->numRows)) {
      return;
   }
   //multiply to each value in the matrix data array
   for(uint16_t i= 0; i < (m->numRows * m->numCols); i++) {
      output->data[i]= s * m->data[i];
   }
}

/**
 * Add two matrices.  The two input matrices and the output matrix must have
 * the same dimensions.
 * @param A The first matrix to add.
 * @param B The second matrix to add.
 * @param output A + B
 */
void matrix_add(matrix_t* A, matrix_t* B, matrix_t* output) {
   if((A->numCols != B->numCols) || (A->numRows != B->numRows)) {
      return;
   }
   if((A->numCols != output->numCols) || (A->numRows != output->numRows)) {
      return;
   }

   //allow addition of different sized matricies if one of them
   //is actually a scalar matrix.
   if(matrix_isScalar(A)) {
      matrix_add_scalar(B, matrix_get(A, 0, 0), output);
   }
   else if(matrix_isScalar(B)) {
      matrix_add_scalar(A, matrix_get(B, 0, 0), output);
   }

   else {
      for(uint16_t i= 0; i < (output->numRows * output->numCols); i++) {
         output->data[i]= A->data[i] + B->data[i];
      }
   }
}

/**
 * Multiply two matrices.  The number of columns in the first matrix must be
 * the same as the number of rows in the second matrix.  The output is
 * stored in the address given.  The dimensions of the output matrix
 * are (rows in A * columns in B)
 * output= A*B
 * @param A The first matrix to multiply.
 * @param B The second matrix to multiply.
 * @param output An instance of a matrix to store the output (A*B)
 */
void matrix_multiply(matrix_t* A, matrix_t* B, matrix_t* output) {
   //error check arguments
   if(A->numCols != B->numRows) {
      return;
   }
   if((A->numRows != output->numRows) || (B->numCols != output->numCols)) {
      return;
   }

   //allow multiplication of different sized matrices if one of them
   //is actually a scalar matrix.
   if(matrix_isScalar(A)) {
      matrix_multiply_scalar(B, matrix_get(A, 0, 0), output);
   }
   else if(matrix_isScalar(B)) {
      matrix_multiply_scalar(A, matrix_get(B, 0, 0), output);
   }
   else {
      uint8_t p= A->numCols; //could also use number of rows in B
      float sum;
      for(uint8_t i= 0; i < output->numRows; i++) {
         for(uint8_t j= 0; j < output->numCols; j++) {
            sum= 0;
            for(uint8_t k= 0; k < p; k++) {
               sum+= matrix_get(A, i, k) * matrix_get(B, k, j);
            }
            matrix_set(output, i, j, sum);
         }
      }
   }
}

/**
 * Determine if a matrix is a scalar number.  That is, determine if the matrix
 * is a 1x1 matrix.
 * @param m The matrix to test.
 * @return True if the matrix is 1x1, false otherwise.
 */
bool matrix_isScalar(matrix_t* m) {
   if((m->numRows == 1) && (m->numCols == 1)) {
      return true;
   }

   else return false;
}

/**
 * Find the determinant of a matrix.
 * @param m The matrix to find the determinant of.
 * @return The determinant of m.
 */
float matrix_determinant(matrix_t* m) {
   float retVal= 0;

   if(m->numRows != m->numCols) {
      return 0;
   }
   //if matrix is 3x3 or 2x2, use rule of sarrus for more efficient computation
   if((m->numRows <= 3) && (m->numRows > 1)) {
      float tmpProduct1;
      float tmpProduct2;
      for(uint8_t i= 0; i < m->numRows; i++) {
         //reset tmpProduct before accumulating inside inner loop
         tmpProduct1= 1;
         tmpProduct2= 1;
         for(uint8_t j= 0; j < m->numCols; j++) {
            tmpProduct1*= matrix_get(m, j, (i+j) % m->numCols);
            tmpProduct2*= matrix_get(m, m->numRows - 1 - j, (i+j) % m->numCols);
            retVal+= tmpProduct1 - tmpProduct2;
         }
      }
      return retVal;
   }

   //otherwise use full recursive method for nxn matrix

   //stopping condition
   if(m->numRows == 1) {
      return matrix_get(m, 0, 0);
   }
   uint8_t bm= 1; //bitmask for checking even/odd later
   matrix_t minor; //storage for minor of m
   matrix_create(&minor, m->numRows - 1, m->numCols - 1);
// The recursive definition is :
//   det( m ) = Summation( i = 0 to order ) [ (-1)^i * m[0][i] * det( minor( m[0][i] ))]
   for(uint8_t i= 0; i < m->numRows; i++) {
      matrix_minor(m, 0, i, &minor);
      //this is a more efficient way of saying (-1)^i
      if(i & bm) {
         retVal+= -1*matrix_get(m, 0, i) * matrix_determinant(&minor);
      }
      else {
         retVal+= matrix_get(m, 0, i) * matrix_determinant(&minor);
      }
   }
   //free back memory
   matrix_delete(&minor);

   return retVal;
}

/**
 * Get the minor of a matrix.  The minor of a matrix is the matrix which results
 * from deleting row i and column j.
 * @param m The matrix to find the minor of.
 * @param i The row to "delete"
 * @param j The column to "delete
 * @param output The location to store the output minor matrix
 */
void matrix_minor(matrix_t* m, uint8_t i, uint8_t j, matrix_t* output) {
   uint8_t numRows= m->numRows;
   uint8_t numCols= m->numCols;
   uint8_t skipRow= 0;
   uint8_t skipCol= 0;
   for(uint8_t row= 0; row < numRows - 1; row++) {
      skipCol= 0;
      for(uint8_t col= 0; col < numCols - 1; col++) {
         if(row == i) {
            skipRow= 1;
         }
         if(col == j) {
            skipCol= 1;
         }
         matrix_set(output, row, col,
                    matrix_get(m, row + skipRow, col+ skipCol));
      }
   }
}


/**
 * Get a cofactor matrix from the matrix m.
 * @param m The matrix to obtain a cofactor matrix from.
 * @param transpose True to return a transpose cofactor matrix, false to return
 *                  a standard cofactor matrix.
 * @param output An instance of a matrix to store the output. Must have the
 *               dimensions as m.
 */
void matrix_cofactor(matrix_t* m, bool transpose, matrix_t* output) {
   if((m->numCols != output->numCols) || (m->numRows != output->numRows)) {
      return;
   }
   float tmpVal; //temporary element storage
   matrix_t minor; //temporary minor matrix storage
   matrix_create(&minor, m->numRows - 1, m->numCols - 1);
   for(uint8_t i= 0; i < output->numRows; i++) {
      for(uint8_t j= 0; j < output->numCols; j++) {
         matrix_minor(m, i, j, &minor);
         tmpVal= matrix_determinant(&minor);
         //check if the negative sign must be added
         //if i+ j is odd...
         if(((i + j) % 2)) {
            tmpVal*= -1;
         }
         //if in transpose mode then use i and j the opposite of what seems right
         if(transpose) {
            matrix_set(output, j, i, tmpVal);
         }
         //otherwise set normally
         else {
            matrix_set(output, i, j, tmpVal);
         }

      }
   }
   matrix_delete(&minor);
}

/**
 * Get the transpose of a matrix.  The input and output matrices must have the
 * same dimensions.
 * @param m The matrix to transpose.
 * @param output A location to store the output matrix.
 */
void matrix_transpose(matrix_t* m, matrix_t* output) {
   if((m->numCols != output->numCols) || (m->numRows != output->numRows)) {
      return;
   }
   for(uint8_t i= 0; i < output->numRows; i++) {
      for(uint8_t j= 0; j < output->numCols; j++) {
         matrix_set(output, j, i, matrix_get(m, i, j));
      }
   }
}

/**
 * Invert a matrix.
 * @param m The matrix to invert
 * @param output The output of the inversion.
 */
void matrix_invert(matrix_t* m, matrix_t* output) {
   //first store the transpose cofactor matrix in the output.
   matrix_cofactor(m, true, output);
   //calculate the determinante of m
   float detInv= 1/matrix_determinant(m);
   //multiply all elements in cofactor matrix by 1/(det(m))
   for(uint8_t i= 0; i < (output->numRows * output->numCols); i++) {
      output->data[i]*= detInv;
   }
}
