/*
 * pid.c
 *
 *  Created on: Dec 17, 2010
 *      Author: hsrmqp
 */

#include "utils/pid.h"
#include "utils/system_time.h"

//Static prototypes
static float sumBuffer_f(pid_f_buffer_t* pidBuf);
static void addBuffer_f(pid_f_buffer_t* pidBuf, float val);
static uint32_t sumBuffer_i(pid_i_buffer_t* pidBuf);
static void addBuffer_i(pid_i_buffer_t* pidBuf, uint32_t val);


/**
 * Run a single pid loop.
 * @param pidStruct An instance of a pid struct setup to use.
 *                  This must be initialized and have some error
 *                  set inside it.
 * @param deltaTime_s The time between the last running of PID and this one
 *                    measured in seconds
 * @return The result of the PID loop
 */
float runPID_f(pid_f_t* pidStruct, float deltaTime_s) {
   float retVal;

   //proportional error
   retVal= pidStruct->kp*pidStruct->err;

   //integrate the error
   if(pidStruct->prevErr) {
      //add error to buffer
      addBuffer_f(pidStruct->prevErr, pidStruct->err);
      retVal+= sumBuffer_f(pidStruct->prevErr)*deltaTime_s*pidStruct->ki;
   }

   //differentiate the error
   retVal+= ((pidStruct->err - pidStruct->lastErr)*deltaTime_s)*pidStruct->kd;
   pidStruct->lastErr= pidStruct->err;

   return retVal;
}

/**
 * Setup a pid struct.
 * @param pidStruct An instance of a pid struct to initialize.
 * @param pidBuf An instance of a PID buffer.  This can be ignored
 *               if the integral component is not used
 * @param kp Proportional constant.
 * @param ki Integral constant.
 * @param kd Derivative constant.
 */
void setupPID_f(pid_f_t* pidStruct, pid_f_buffer_t* pidBuf, float kp, float ki, float kd) {
   pidStruct->kp= kp;
   pidStruct->ki= ki;
   pidStruct->kd= kd;

   pidStruct->prevErr= pidBuf;
   if(pidBuf) {
      pidStruct->prevErr->bufIndex= 0;
      for(uint8_t i= 0; i < PID_BUFFER_SIZE; i++) {
         pidStruct->prevErr->errors[i]= 0;
      }
   }
   pidStruct->err= 0;
}

/**
 * Set the error in a pid struct.
 * @param pidStruct An instance of a pid struct.
 * @param err The new value of the error.
 */
void setErrorPID_f(pid_f_t* pidStruct, float err) {
   pidStruct->err= err;
}

/**
 * Run a single pid loop using only integer math.
 */
int32_t runPID_i(pid_i_t* pidStruct, uint64_t deltaTime_us) {
   int32_t retVal;

   //calculating the result is basically the same but the error and all
   //constants are scaled up by some factor
   //proportional error
   //(k*scale)*(err*scale)= k*err*scale^2
   retVal= pidStruct->kp*pidStruct->err;

   //integrate the error if the user is using integral control
   //check for null pointer
   if(pidStruct->prevErr) {
      //add error to buffer
      //(err*scale)*(time*1000000)/(1000000)= intErr*scale
      //(k*scale)*(intErr*scale)= k*err*scale^2
      addBuffer_i(pidStruct->prevErr, pidStruct->err);
      retVal+= (uint64_t)sumBuffer_i(pidStruct->prevErr)*deltaTime_us*pidStruct->ki/(US_PER_S);
   }

   //differentiate the error
   //(derErr*scale)*(k*scale)= k*derErr*scale^2
   //(deltaError)*(number of us in a s)*kd/(time difference in us) = kd*(dError/dt)
   retVal+= (uint64_t)((pidStruct->err - pidStruct->lastErr)*US_PER_S)*pidStruct->kd/(deltaTime_us);
   pidStruct->lastErr= pidStruct->err;

   //retVal now equals (kp*err + ki*intErr + kd*derErr)*scale^2

   //divide to return as (normal output)*scale

   return retVal/INTEGER_PID_PRECISION;
}

/*
 * Setup a PID struct for integer math.  The integer PID loop scales up all
 * the numbers before using calculations.  However, for ease of use the
 * pid struct is initialized as if it was done with floating point.
 * @param pidStruct Instance of the PID struct to initialize.
 * @param pidBuf Instance of error buffer.  This can be ignored if the integer
 *        component isn't being used
 * @param kp The proportional constant
 * @param ki The integral constant
 * @param kd The derivative constant.
 */
void setupPID_i(pid_i_t* pidStruct, pid_i_buffer_t* pidBuf, float kp, float ki, float kd) {
   pidStruct->kp= kp*INTEGER_PID_PRECISION;
   pidStruct->ki= ki*INTEGER_PID_PRECISION;
   pidStruct->kd= kd*INTEGER_PID_PRECISION;

   pidStruct->prevErr= pidBuf;
   if(pidBuf) {
      pidStruct->prevErr->bufIndex= 0;
      for(uint8_t i= 0; i < PID_BUFFER_SIZE; i++) {
         pidStruct->prevErr->errors[i]= 0;
      }
   }
   pidStruct->err= 0;
}

/*
 * Set the error of an integer math pid struct.  The error will be entered
 * should be scaled by INTEGER_PID_PRECISION otherwise you'll end up with
 * some roundoff errors.
 */
void setErrorPID_i(pid_i_t* pidStruct, uint32_t err) {
   pidStruct->err= err;
}

/**
 * Run a single PID using a sort of hybrid between integer and floating point
 * math.  Most of the math is done through integer calculations but the
 * minimum needed is done as floating point.  From the outside perspective,
 * it should appear to work exactly as the pure floating point PID with the
 * only exception of being rounded off after log(INTEGER_PID_PRECISION) places.
 */
float runPID(pid_t* pidStruct, uint64_t deltaTime_us) {
   int32_t retVal_i; //retVal as integer
   float retVal;

   //calculating the result is basically the same but the error and all
   //constants are scaled up by some factor
   //proportional error
   //(k*scale)*(err*scale)= k*err*scale^2
   retVal_i= pidStruct->kp*pidStruct->err;

   //integrate the error if the user is using integral control
   //check for null pointer
   if(pidStruct->prevErr) {
      //add error to buffer
      //(err*scale)*(time*1000000)/(1000000)= intErr*scale
      //(k*scale)*(intErr*scale)= k*err*scale^2
      addBuffer_i(pidStruct->prevErr, pidStruct->err);
      retVal_i+= (int64_t)sumBuffer_i(pidStruct->prevErr)*deltaTime_us*pidStruct->ki/(US_PER_S);
   }

   //differentiate the error
   //(derErr*scale)*(k*scale)= k*derErr*scale^2
   //(deltaError)*(number of us in a s)*kd/(time difference in us) = kd*(dError/dt)
   retVal_i+= ((int64_t)(pidStruct->err - pidStruct->lastErr)*US_PER_S)*pidStruct->kd/(deltaTime_us);
   pidStruct->lastErr= pidStruct->err;

   //retVal_i now equals (kp*err + ki*intErr + kd*derErr)*scale^2

   //divide by scale^2 and cast as a float to get floating point output

   retVal= ((float)retVal_i)/(INTEGER_PID_PRECISION*INTEGER_PID_PRECISION);

   return retVal;
}

/**
 * Setup PID for hybrid integer/float version.  All integer math is done externally
 * so treat kp, ki, and kd normally.
 * If you aren't using integer control, set pidBuf to a null pointer.
 */
void setupPID(pid_t* pidStruct, pid_i_buffer_t* pidBuf, float kp, float ki, float kd) {
   pidStruct->kp= kp*INTEGER_PID_PRECISION;
   pidStruct->ki= ki*INTEGER_PID_PRECISION;
   pidStruct->kd= kd*INTEGER_PID_PRECISION;

   pidStruct->prevErr= pidBuf;
   if(pidBuf) {
      pidStruct->prevErr->bufIndex= 0;
      for(uint8_t i= 0; i < PID_BUFFER_SIZE; i++) {
         pidStruct->prevErr->errors[i]= 0;
      }
   }
   pidStruct->err= 0;
}

/**
 * Set the error for a hybrid pid struct.
 */
void setErrorPID(pid_t* pidStruct, float val) {
   pidStruct->err=val*INTEGER_PID_PRECISION;
}

static float sumBuffer_f(pid_f_buffer_t* pidBuf) {
   float retVal= 0;

   for(uint8_t i= 0; i < PID_BUFFER_SIZE; i++) {
      retVal+= pidBuf->errors[i];
   }

   return retVal;
}

static void addBuffer_f(pid_f_buffer_t* pidBuf, float val) {
   pidBuf->errors[pidBuf->bufIndex]= val;
   (pidBuf->bufIndex)++;
   pidBuf->bufIndex &= PID_BUFFER_MASK;
}

static uint32_t sumBuffer_i(pid_i_buffer_t* pidBuf) {
   uint32_t retVal= 0;

   for(uint8_t i= 0; i < PID_BUFFER_SIZE; i++) {
      retVal+= pidBuf->errors[i];
   }

   return retVal;
}

static void addBuffer_i(pid_i_buffer_t* pidBuf, uint32_t val) {
   pidBuf->errors[pidBuf->bufIndex]= val;
   (pidBuf->bufIndex)++;
   pidBuf->bufIndex &= PID_BUFFER_MASK;
}
