/*
 * serialize.c
 *
 *  Created on: Jan 10, 2011
 *      Author: Eric
 */
#include "utils/serialize.h"

/* Convert a 16 bit integer into an array of bytes.  The array of bytes
* will be big endian.
* @param integer The integer to convert.
* @param serialDataBuf Location to store the output array.
* @return the number of bytes written to the buffer, i.e. sizeof(integer).
*/
uint8_t serializeInt8(uint8_t integer, uint8_t serialDataBuf[]) {
  serialDataBuf[0]= integer;
  return sizeof(uint8_t);
}

/**
* Convert an array of bytes into a 16 bit integer.  The array of bytes must
* be big endian.
* @param integer Pointer to the location to store the resulting conversion.
* @param serialData The serialized array of bytes.
* @return The number of bytes read off the data array, i.e. sizeof(*integer).
*/
uint8_t deSerializeInt8(uint8_t* integer, uint8_t serialData[]) {
  *integer= serialData[0];
  return sizeof(uint8_t);
}

/**
 * Convert a 16 bit integer into an array of bytes.  The array of bytes
 * will be big endian.
 * @param integer The integer to convert.
 * @param serialDataBuf Location to store the output array.
 * @return the number of bytes written to the buffer, i.e. sizeof(integer).
 */
uint8_t serializeInt16(uint16_t integer, uint8_t serialDataBuf[]) {
   serialDataBuf[0]= (integer & 0xFF00) >> 8;
   serialDataBuf[1]= (integer & 0x00FF);
   return sizeof(uint16_t);
}

/**
 * Convert an array of bytes into a 16 bit integer.  The array of bytes must
 * be big endian.
 * @param integer Pointer to the location to store the resulting conversion.
 * @param serialData The serialized array of bytes.
 * @return The number of bytes read off the data array, i.e. sizeof(*integer).
 */
uint8_t deSerializeInt16(uint16_t* integer, uint8_t serialData[]) {
   uint16_t retVal= 0;
   retVal|= (uint16_t)serialData[0] << 8;
   retVal|= (uint16_t)serialData[1];

   *integer= retVal;
   return sizeof(uint16_t);
}

/**
 * Convert a 32 bit integer into an array of bytes.  The array of bytes
 * will be big endian.
 * @param integer The integer to convert.
 * @param serialDataBuf Location to store the output array.
 * @return the number of bytes written to the buffer, i.e. sizeof(integer).
 */
uint8_t serializeInt32(uint32_t integer, uint8_t serialDataBuf[]) {
   serialDataBuf[0]= (integer & 0xFF000000) >> 24;
   serialDataBuf[1]= (integer & 0x00FF0000) >> 16;
   serialDataBuf[2]= (integer & 0x0000FF00) >> 8;
   serialDataBuf[3]= (integer & 0x000000FF);
   return sizeof(uint32_t);
}

/**
 * Convert an array of bytes into a 32 bit integer.  The array of bytes must
 * be big endian.
 * @param integer Pointer to the location to store the resulting conversion.
 * @param serialData The serialized array of bytes.
 * @return The number of bytes read off the data array, i.e. sizeof(*integer).
 */
uint8_t deSerializeInt32(uint32_t* integer, uint8_t serialData[]) {
   uint32_t retVal= 0;
   retVal|= (uint32_t)serialData[0] << 24;
   retVal|= (uint32_t)serialData[1] << 16;
   retVal|= (uint32_t)serialData[2] << 8;
   retVal|= (uint32_t)serialData[3];

   *integer= retVal;
   return sizeof(uint32_t);
}

/**
 * Convert a 64 bit integer into an array of bytes.  The array of bytes
 * will be big endian.
 * @param integer The integer to convert.
 * @param serialDataBuf Location to store the output array.
 * @return the number of bytes written to the buffer, i.e. sizeof(integer).
 */
uint8_t serializeInt64(uint64_t integer, uint8_t serialDataBuf[]) {
   serialDataBuf[0]= (integer & 0xFF00000000000000) >> 56;
   serialDataBuf[1]= (integer & 0x00FF000000000000) >> 48;
   serialDataBuf[2]= (integer & 0x0000FF0000000000) >> 40;
   serialDataBuf[3]= (integer & 0x000000FF00000000) >> 32;
   serialDataBuf[4]= (integer & 0x00000000FF000000) >> 24;
   serialDataBuf[5]= (integer & 0x0000000000FF0000) >> 16;
   serialDataBuf[6]= (integer & 0x000000000000FF00) >> 8;
   serialDataBuf[7]= (integer & 0x00000000000000FF);
   return sizeof(uint64_t);
}

/**
 * Convert an array of bytes into a 32 bit integer.  The array of bytes must
 * be big endian.
 * @param integer Pointer to the location to store the resulting conversion.
 * @param serialData The serialized array of bytes.
 * @return The number of bytes read off the data array, i.e. sizeof(*integer).
 */
uint8_t deSerializeInt64(uint64_t* integer, uint8_t serialData[]) {
   uint64_t retVal= 0;
   retVal|= (uint64_t)serialData[0] << 56;
   retVal|= (uint64_t)serialData[1] << 48;
   retVal|= (uint64_t)serialData[2] << 40;
   retVal|= (uint64_t)serialData[3] << 32;
   retVal|= (uint64_t)serialData[4] << 24;
   retVal|= (uint64_t)serialData[5] << 16;
   retVal|= (uint64_t)serialData[6] << 8;
   retVal|= (uint64_t)serialData[7];

   *integer= retVal;
   return sizeof(uint64_t);
}
