/*
 * simplified_math.c
 *
 *  Created on: Feb 19, 2011
 *      Author: Eric
 */
#include "utils/simplified_math.h"
#include <stdint.h>
#include <math.h>


//Static prototypes
static uint8_t num_binary_digits(uint32_t x);

/**
 * This function is designed to be a simpler alternative to the hypot function
 * defined in math.h.  If superior accuracy is not required this function
 * is acceptable.  The usage is intended for places where there are not
 * sufficient resources to use the real hypot function.
 * @param a Parameter a
 * @param b Parameter b
 * @return sqrt(a^2 + b^2)
 */
float hypot_simple(float a, float b) {
   float a_2= a*a;
   float b_2= b*b;
   return sqrt_simple(a_2 + b_2);
}

/**
 * This function is designed to be a simpler alternative to the sqrt function
 * defined in math.h. It is less accurate but is acceptable for places when
 * the conservation of resources is more important than accuracy. The function
 * uses the Babylonian sqrt method and iterates BABYLONIAN_SQRT_ITERATIONS
 * times. NOTE: x must be positive or NAN is returned.
 * @param x The input to the sqare root.
 * @return sqrt(x)
 */
float sqrt_simple(float x) {
   //guess/improved estimations for sqrt(x)
   float sqrt_x_guess;
   float sqrt_x_improved;

   //can't do square root of negative number
   if (x < 0)
      return NAN;
   //start with an initial guess for sqrt(x).
   //sqrt(x)~= 2^(N/2) where N is number of binary digits
   uint32_t x_int= x;
   sqrt_x_guess= 1 << (num_binary_digits(x_int)/2);

   //iterate through improving the square root estimate
   //x_n+1= 1/2 * (x_n + S/x_n) where x_n+1 is a better approximation
   //than x_n and S is the original number to take the square root of.
   for(uint8_t i= 0; i < BABYLONIAN_SQRT_ITERATIONS; i++) {
      sqrt_x_improved= (sqrt_x_guess + x/sqrt_x_guess)/2;
      sqrt_x_guess= sqrt_x_improved;
   }

   return sqrt_x_improved;
}

/**
 * Calculate the number of binary digits in a number. The number 0 is
 * considered to have zero digits.
 * This is a helper to square root function.
 * @param x Number to count number of digits.
 * @return The number of binary digits in x.
 */
static uint8_t num_binary_digits(uint32_t x) {
   uint8_t numDigits= 0;
   while(x != 0) {
      x= x >> 1;
      numDigits++;
   }
   return numDigits;
}
