/*
 * transformation_matrix.c
 *
 *  Created on: Dec 16, 2010
 *      Author: hsrmqp
 */

#include "utils/matrix.h"
#include "math.h"
#include "utils/transformation_matrix.h"

/**
 * Set the values in a transformation matrix.
 * @param tfMatrix Instance of a 3x3 matrix
 * @param x_off The translation in the x direction.
 * @param y_off The translation in the y direction.
 * @param theta The angular rotation in radians.
 */
void tf_matrix_set(matrix_t* tfMatrix, float x_off, float y_off, float theta) {
   //rotation
   matrix_set(tfMatrix, 0, 0, cos(theta));
   matrix_set(tfMatrix, 0, 1, -sin(theta));
   matrix_set(tfMatrix, 1, 0, sin(theta));
   matrix_set(tfMatrix, 1, 1, cos(theta));

   //translation
   matrix_set(tfMatrix, 0, 2, x_off);
   matrix_set(tfMatrix, 1, 2, y_off);

   //last row, always [0 0 1]
   matrix_set(tfMatrix, 2, 0, 0);
   matrix_set(tfMatrix, 2, 1, 0);
   matrix_set(tfMatrix, 2, 2, 1);
}

/**
 * Set up a vector matrix.
 * @param tfVector Instance of a 3x1 matrix.
 * @param x x component of the vector.
 * @param y y component of the vector.
 */
void tf_vector_set(matrix_t* tfVector, float x, float y) {
   matrix_set(tfVector, 0, 0, x);
   matrix_set(tfVector, 1, 0, y);
   matrix_set(tfVector, 2, 0, 1);
}

/**
 * Get the x component of a transformation vector.
 * @param tfVector Instance of a transformation vector.
 */
float tf_vector_getx(matrix_t* tfVector) {
   return matrix_get(tfVector, 0, 0);
}

/**
 * Get the y component of a transformation vector.
 * @param tfVector Instance of a transformation vector.
 */
float tf_vector_gety(matrix_t* tfVector) {
   return matrix_get(tfVector, 1, 0);
}


/**
 * Perform a single transformation on a vector.
 * @param tfMatrix Instance of a transformation matrix.
 * @param inputVector The vector to transform.
 * @param outputVector Location to store output.
 */
void tf_transform(matrix_t* tfMatrix, matrix_t* inputVector, matrix_t* outputVector) {
   matrix_multiply(tfMatrix, inputVector, outputVector);
}

/**
 * Transform a chain of transformation matrices.
 * @param tfMatrixChain An array of pointers to transformation matrices.
 *                      This should be null terminated.
 * @param inputVector The vector to transform.
 * @param outputVector Location to store the output.
 */
void tf_transform_chain(matrix_t** tfMatrixChain, matrix_t* inputVector, matrix_t* outputVector) {
   //create temporary matrix storage for multiplications
   matrix_t tmpMatrix;
   matrix_create(&tmpMatrix, 3, 3);

   while((tfMatrixChain+1)!= 0) {
      matrix_multiply(*tfMatrixChain, *(tfMatrixChain+1), &tmpMatrix);
      tfMatrixChain++;
   }

   matrix_multiply(&tmpMatrix, inputVector, outputVector);

   matrix_delete(&tmpMatrix);
}
