/*
 * xbee.c
 *
 *  Created on: Dec 6, 2010
 *      Author: hsrmqp
 */

#include "xbee.h"
#include "includes.h"
#include "utils/serialize.h"
#include "utils/emb_malloc.h"

//Global variables
static const xbee_setup_t* sSetup= 0;

static xbee_address_lookup_t sAddressLookupTable[XBEE_MAX_ADDRESSES];
static uint8_t sNumAddressEntries= 0; //number of entries in table

static uint8_t sframeID= 0; //incremental counter for frame id
static uint8_t sLastSentAD= 0; //track which AD was last sent something
static uint8_t sTXserialFrameBuffer[XBEE_MAX_PAYLOAD+18];  //max payload + packet overhead
static uint8_t sRXserialFrameBuffer[XBEE_MAX_PAYLOAD+16];  //max payload + packet overhead
//static space for rx storage
static xbee_api_tx_req_frame sTXFrame;
static xbee_api_rx_pac_frame_t sRXFrame;
static xbee_api_tx_status_frame_t sTXStatusFrame;

//Static Prototypes
static void handle_xbee_rx_packet(xbee_api_rx_pac_frame_t* rxFrame);
static void handle_xbee_tx_status_packet(xbee_api_tx_status_frame_t* txStatusFrame);
static void handle_xbee_packet(uint8_t* rxSerialPacket);
static uint8_t lookupAddressDescriptor(uint64_t addr64bit, uint16_t addr16bit);
static uint8_t addAddressTableEntry(uint64_t addr64bit, uint16_t addr16bit, uint8_t index);
//static void removeAddressTableEntry(uint8_t ad); currently unused
static void initAddressTable(void);
static void createTXAPIFrame(xbee_api_tx_req_frame* txFrame,
                           uint64_t destAddr64bit, uint16_t destAddr16bit,
                           uint8_t broadcastRadius,
                           uint8_t optionsBitmask,
                           uint8_t frameID,
                           const uint8_t* payload,
                           uint8_t payloadSize);
static uint8_t calculateAPIFrameChecksum(uint8_t serialData[]);
static uint16_t serializeTXAPIFrame(xbee_api_tx_req_frame* txFrame, uint8_t* serialFrame);
static bool deSerializeRXAPIFrame(xbee_api_rx_pac_frame_t* rxFrame,
                                  uint8_t* serialFrame);
static bool deSerializeTXStatusAPIFrame(xbee_api_tx_status_frame_t* txStatusFrame,
                                        uint8_t* serialFrame);
static void writeRawToUSART(const uint8_t* bytes, uint8_t numBytes);

void init_xbee(xbee_setup_t* setup) {
   sSetup= setup;
   //set all entries in table to be unknown
   initAddressTable();
   //add parent of scout
   //addAddressTableEntry(0x0013A200403B0E27ull, 0xFFFE, 0); //Q addr
   //addAddressTableEntry(0x0013A200405D7EF0ull, 0xFFFE, 0); //W1 addr
   addAddressTableEntry(0x0013A200405D7EEFull, 0xFFFE, 0); //W2 addr
}

void xbee_receive_task(void* pd) {
   uint16_t i= 0; //index for rx buffer, also number of bytes read in
   uint16_t totalSize;
   for(;;) {
      i= 0;
      totalSize= 0xFFFF; //can't actually be this big but it will be set to a
                         //real value once we get bytes 2-3 of the packet
      bool packetComplete= false;
      //loop until a full packet is received
      while(!packetComplete) {
         //grab a byte off the usart and block if there is nothing to receive
         sRXserialFrameBuffer[i]= USART_GetByte_Block(sSetup->usart_info,
                                                      xbee_rx_data_sem);
         //throw out if the 1st byte isn't the start delimiter
         if(i == 0 && sRXserialFrameBuffer[i] != XBEE_START_DELIM) {
            i= 0;
         }
         else {
            i++;
         }

         //read the packet length if enough bytes have been read in
         if(i == 3) {
            totalSize= 0;
            totalSize|= sRXserialFrameBuffer[1] << 8;
            totalSize|= sRXserialFrameBuffer[2];
            totalSize+= 4; //4 bytes for start delim, length, and checksum
         }

         //check if the entire packet has been received
         if(i >= totalSize) {
            packetComplete= true;
         }
      }
      if(sRXserialFrameBuffer[totalSize - 1] != calculateAPIFrameChecksum(sRXserialFrameBuffer)) {
         //do nothing
      }
      else {
         //handle the packet
         handle_xbee_packet(sRXserialFrameBuffer);
      }
   }
}

/**
 * Send a generic array of bytes to the given address descriptor.
 * @param ad The address descriptor to send to.
 * @param data Array of bytes to send.
 * @param size The number of bytes in the data array.
 * @return XBEE_SEND_SUCCESS - Everything went ok
 *         XBEE_SEND_FAILURE - The destination did not receive the packet
 *         XBEE_SEND_NO_AD   - The given ad does not exist
 */
int8_t xbee_send(uint8_t ad, const uint8_t* data, uint8_t size){
   //if size is zero then just do nothing
   if(size == 0) {
      //if you want to send zero bytes then you're already done
      return XBEE_SEND_SUCCESS;
   }
   //make sure ad is in bounds to avoid array out of bounds errors
   if(ad > sNumAddressEntries) {
      return XBEE_SEND_NO_AD;
   }
   //get the 64 and 16 bit address for the descriptor ad
   uint16_t addr16bit= sAddressLookupTable[ad].networkAddr;
   uint64_t addr64bit= sAddressLookupTable[ad].addr64bit;

   if(addr16bit == XBEE_16BIT_UNKNOWN && addr64bit == XBEE_64BIT_UNKNOWN) {
      return XBEE_SEND_NO_AD;
   }

   //increment frame ID # and make sure it is non-zero
   sframeID++;
   if(sframeID == 0) {
      sframeID++;
   }

   //remember last AD we sent to
   sLastSentAD= ad;

   uint8_t* segmentedData= emb_malloc(size + 4);
   segmentedData[0]= 0;
   segmentedData[1]= 1;
   segmentedData[2]= 0;
   segmentedData[3]= 0;

   for(uint8_t i= 0; i < size; i++) {
      segmentedData[i+4]= data[i];
   }
   createTXAPIFrame(&sTXFrame, addr64bit, addr16bit, 0, 0, sframeID, segmentedData, size+4);


   //convert api frame to serial array of bytes
   uint16_t totalFrameSize;
   totalFrameSize= serializeTXAPIFrame(&sTXFrame, sTXserialFrameBuffer);

   emb_free(segmentedData);

   //write the bytes out the USART
   writeRawToUSART(sTXserialFrameBuffer, totalFrameSize);

   //sleep until response comes in and return that
   uint8_t err;
   OS_FLAGS statusFlag;
   statusFlag= OSFlagPend(xbee_tx_status_flag, XBEE_TX_SUCCESS_FLAG + XBEE_TX_FAIL_FLAG,
                          OS_FLAG_WAIT_SET_ANY + OS_FLAG_CONSUME, XBEE_TX_TIMEOUT, &err);
   if(err == OS_ERR_TIMEOUT) {
      return XBEE_SEND_FAILURE;
   }
   else if(statusFlag == XBEE_TX_SUCCESS_FLAG) {
      return XBEE_SEND_SUCCESS;

   }
   else {
      return XBEE_SEND_FAILURE;
   }

}

/**
 * Handler for a received rx packet.  This function formats the data in a way
 * to be better presented to the communications module.  The rx data is assumed
 * to come in the form of [number of bytes in sequence,
 * sequence number (starting at 0), rest of the data].  For the scout we can't
 * buffer any more than 1 xbee packet so it isn't worth it to even look at
 * these because we don't have the RAM to handle it even if we wanted to.
 */
static void handle_xbee_rx_packet(xbee_api_rx_pac_frame_t* rxFrame) {
   int8_t descriptor; //entry in address lookup table

   descriptor= lookupAddressDescriptor(rxFrame->addr64bit, rxFrame->addr16bit);

   if(descriptor < 0) {
      //packet not from valid sender: discard
      return;
   }

   //hand off to listener if one exists
   //the first 4 bytes are chopped off because they contain the sequencing
   //information.
   uint8_t* rxData= rxFrame->data + 4;
   uint8_t rxPayloadLength= rxFrame->payloadLength - 4;
   if(sSetup->callback!= 0) {
      sSetup->callback(descriptor, rxData, rxPayloadLength);
   }

}

/**
 * Handler for a received tx status packet.
 */
static void handle_xbee_tx_status_packet(xbee_api_tx_status_frame_t* txStatusFrame) {
   if(txStatusFrame->frameID == sframeID) {
      uint8_t err;
      //for now, just check if tx is good/bad.  probably check more stuff later
      if(txStatusFrame->deliveryStatus == 0x00) {
         //add the discovered 16 bit network address
         sAddressLookupTable[sLastSentAD].networkAddr= txStatusFrame->addr16bit;
         OSFlagPost(xbee_tx_status_flag, XBEE_TX_SUCCESS_FLAG, OS_FLAG_SET, &err);
      }
      else {
         OSFlagPost(xbee_tx_status_flag, XBEE_TX_FAIL_FLAG, OS_FLAG_SET, &err);
      }
   }
}

/**
 * Look up an address descriptor based on the given 64/16 bit address.
 * @param addr64bit The 64 bit address.  Use 0xFFFFFFFFFFFFFFFF if the address
 *                  is unknown.
 * @param addr16bit The 16 bit network address.  Use 0xFFFE if the address is
 *                  unknown.
 * @return The address descriptor matching the given addresses or -1 if none
 *         found.
 */
static uint8_t lookupAddressDescriptor(uint64_t addr64bit, uint16_t addr16bit) {
   //both 64 and 16 bit can't be unknown
   if( (addr64bit == 0xFFFFFFFFFFFFFFFFULL) && (addr16bit == 0xFFFE) ) {
      return -1;
   }
   //check every table entry looking for a match of either the 64 bit or 16 bit
   //address.  return index if found
   //if only 64 bit is known
   if(addr16bit == XBEE_16BIT_UNKNOWN) {
      for(uint8_t i= 0; i < sNumAddressEntries; i++) {
         if(sAddressLookupTable[i].addr64bit == addr64bit) {
            return i;
         }
      }
   }
   //if only 16 bit is known
   else if(addr64bit == XBEE_64BIT_UNKNOWN) {
      for(uint8_t i= 0; i < sNumAddressEntries; i++) {
         if(sAddressLookupTable[i].networkAddr == addr16bit) {
            return i;
         }
      }
   }

   //if both are known
   else {
      for(uint8_t i= 0; i < sNumAddressEntries; i++) {
         //lookup using 64 bit because that one never changes
         //from run to run.  The 16 bit address replaces the old
         //one in the table.  That way the 16 bit network address
         //can change at runtime and be used for future reference.
         if(sAddressLookupTable[i].addr64bit == addr64bit) {
            sAddressLookupTable[i].networkAddr= addr16bit;
            return i;
         }
      }
   }

   return -1;
}

/**
 * Add an entry to the address table.
 * @param addr64bit The 64 bit network address to add.  Use 0xFFFFFFFFFFFFFFFF if the address
 *                  is unknown.
 * @param addr16bit The 16 bit network address to add.  Use 0xFFFE if the address is
 *                  unknown.
 * @return The address descriptor that the entry is assigned.  Returns 0xFF if an
 *         error occurs such as making 64 bit and 16 bit address both unknown or the
 *         table is full.
 */
static uint8_t addAddressTableEntry(uint64_t addr64bit, uint16_t addr16bit, uint8_t index) {
   //make sure both 64 bit and 16 bit aren't unknown
   if((addr64bit == 0xFFFFFFFFFFFFFFFF) && (addr16bit == 0xFFFE)) {
      return 0xFF;
   }

   //check if the number of entries  is equal to the max allowed
   if(sNumAddressEntries == XBEE_MAX_ADDRESSES) {
      return 0xFF;
   }

   //otherwise ok to add to table
   sAddressLookupTable[index].addr64bit= addr64bit;
   sAddressLookupTable[index].networkAddr= addr16bit;
   sNumAddressEntries++;

   return index;
}

/**
 * CURRENTLY NOT USED
 * Remove an address table entry.  This will remove the entry from the table
 * and shift all other entries down if needed.
 * @param ad The entry to remove.
 *
static void removeAddressTableEntry(uint8_t ad) {
   //mark the entry ad as unused by setting 16 and 64 bit address
   //to be unknown
   sAddressLookupTable[ad].addr64bit= XBEE_64BIT_UNKNOWN;
   sAddressLookupTable[ad].networkAddr= XBEE_16BIT_UNKNOWN;

   //decrement the number of entries in table
   sNumAddressEntries--;
}
*/

/**
 * Set up the address table with all entries unknown.
 */
static void initAddressTable(void) {
   for(uint8_t i= 0; i < XBEE_MAX_ADDRESSES; i++) {
      sAddressLookupTable[i].addr64bit= XBEE_64BIT_UNKNOWN;
      sAddressLookupTable[i].networkAddr= XBEE_16BIT_UNKNOWN;
   }
}

/**
 * Handle an incoming xbee rx serial packet.  The serial packet is an unfiltered
 * raw stream of bytes sent from the xbee including any delimiters.
 * @param rxSerialPacket A raw stream of bytes
 */
static void handle_xbee_packet(uint8_t* rxSerialPacket) {
   //switch based on packet type
   switch(rxSerialPacket[3]) {
   case XBEE_API_RX_PACKET:
      if(deSerializeRXAPIFrame(&sRXFrame, rxSerialPacket)) {
         //RX callback
         handle_xbee_rx_packet(&sRXFrame);
      }
      break;
   case XBEE_API_TX_STATUS:
      if(deSerializeTXStatusAPIFrame(&sTXStatusFrame, rxSerialPacket)) {
         //TX status callback
         handle_xbee_tx_status_packet(&sTXStatusFrame);
      }
      break;
   case XBEE_API_MODEM_STATUS:
      if(rxSerialPacket[4] == 0x02) {
      }
      break;
   default:
      //do nothing
      break;
   }
}

/**
 * Setup a transmit API frame.
 * @param txFrame An uninitialized instance of a transmit request frame
 * @param destAddr64bit The 64 bit network address for the destination
 * @param destAddr16Bit The 16 bit network address for the destination
 * @param broadcastRadius The number of hops to allow in a broadcast (0 for maximum)
 * @param optionsBitmask Transmit options bitmask
 *                       BIT0: Disable ACK
 *                       BIT1: Enable AES encryption
 *                       BIT3: Use Extended timeout
 * @param frameID Must be non-zero and is used to identify the frame on the
 *                return rx status.
 */
static void createTXAPIFrame(xbee_api_tx_req_frame* txFrame,
                           uint64_t destAddr64bit, uint16_t destAddr16bit,
                           uint8_t broadcastRadius,
                           uint8_t optionsBitmask,
                           uint8_t frameID,
                           const uint8_t* payload,
                           uint8_t payloadSize) {
   txFrame->addr64bit= destAddr64bit;

   txFrame->addr16bit= destAddr16bit;

   txFrame->broadcastRadius= broadcastRadius;

   txFrame->options= optionsBitmask;

   txFrame->frame_id= frameID;

   txFrame->data= payload;

   txFrame->payloadLength= payloadSize;

}

/**
 * Calculate the checksum byte of a TX api frame.
 * @param txFrame An instnace of a transmit api frame struct
 */
static uint8_t calculateAPIFrameChecksum(uint8_t serialData[]) {
   uint8_t checksum= 0;

   uint8_t length= 0;
   length|= serialData[1] << 8;
   length|= serialData[2];
   //checksum is computed by adding up all bytes that are part of the api frame
   //specific frame
   //i goes from 3 to length +3 because there are 3 bytes at the beginning of
   //the serial packet which don't count towards the checksum
   for(uint16_t i= 3; i < length + 3; i++) {
      checksum+= serialData[i];
   }

   checksum= 0xFF - checksum;

   return checksum;
}

/**
 * This will convert a completed api frame into a serial format ready to be
 * written out to the usart.  This function accounts for any endian issues.
 * @param txFrame A completed tx api frame.
 * @param serialFrame An allocated space to store the serialized frame
 * @return The total size of the frame in bytes
 */
static uint16_t serializeTXAPIFrame(xbee_api_tx_req_frame* txFrame, uint8_t* serialFrame) {
   //the length of the frame NOT INCLUDING the start delimiter byte, the length byte
   //or the checksum byte.  These are excluded due to the api format required by xbee
   uint16_t frameLength= XBEE_TX_FRAME_LENGTH(txFrame);
   uint16_t offset= 0;
   //add start delimiter
   serialFrame[0]= XBEE_START_DELIM;
   offset++;
   //add length MSB first
   offset+= serializeInt16(frameLength, &serialFrame[offset]);
   //add frame type
   serialFrame[3]= XBEE_API_TX_REQ;
   offset++;
   //add frame ID
   serialFrame[4]= txFrame->frame_id;
   offset++;
   //add 64 bit destination address MSB first
   offset+= serializeInt64(txFrame->addr64bit, &serialFrame[offset]);
   //add 16 bit destination address MSB first
   offset+= serializeInt16(txFrame->addr16bit, &serialFrame[offset]);
   //add broadcast radius
   serialFrame[offset]= txFrame->broadcastRadius;
   offset++;
   //add options bitmask
   serialFrame[offset]= txFrame->options;
   offset++;
   //add the payload
   for(uint8_t i= 0; i < txFrame->payloadLength; i++) {
      serialFrame[offset+i]= txFrame->data[i];
   }
   //add the checksum byte
   serialFrame[offset+txFrame->payloadLength]= calculateAPIFrameChecksum(serialFrame);

   return frameLength + 4; //the total size of the frame is the frame length
                           //specified in the api packet + the 2 length bytes
                           //start delimiter, and checksum excluded in the
                           //original calculation

}

/**
 * Convert a serial frame into an rx packet struct.
 * @param rxFrame Instance of an rx packet
 * @param serialFrame Array of bytes containing the received serial frame
 * @return true if the frame could be correctly deserialized, false otherwise
 */
static bool deSerializeRXAPIFrame(xbee_api_rx_pac_frame_t* rxFrame, uint8_t* serialFrame) {
   //check that the first byte is the start delimiter
   if(serialFrame[0] != XBEE_START_DELIM) {
      return false;
   }
   //make sure that frame type is rx packet
   if(serialFrame[3] != XBEE_API_RX_PACKET) {
      return false;
   }
   uint16_t offset= 1;
   uint16_t frameLength; //the 2 byte length as specified in the serial frame
   //pull out the frame length from the serial frame, this is used to find out
   //the size of the payload
   offset+= deSerializeInt16(&frameLength, &serialFrame[offset]);
   offset++; //skip the packet opcode since we've already verified that

   //subtract 12 bytes of overhead (start delim, checksum, etc.)
   uint8_t payloadLength= frameLength - 12;

   //pull out 64 bit address
   offset+=deSerializeInt64(&(rxFrame->addr64bit), &serialFrame[offset]);

   //pull out 16 bit address
   offset+=deSerializeInt16(&(rxFrame->addr16bit), &serialFrame[offset]);

   //pull out rx options
   rxFrame->options= serialFrame[offset];
   offset++;

   //pull out the payload
   rxFrame->data= &serialFrame[offset];
   rxFrame->payloadLength= payloadLength;

   //return success
   return true;

}

/**
 * Convert a serial frame into a TX status struct.
 * @param txStatusFrame An instance of a transmit status frame struct
 * @param serialFrame An array of bytes containing the recieved serial frame
 */
static bool deSerializeTXStatusAPIFrame(xbee_api_tx_status_frame_t* txStatusFrame,
                                        uint8_t* serialFrame) {
   //check that the first byte is the start delimiter
   if(serialFrame[0] != XBEE_START_DELIM) {
      return false;
   }
   //make sure that frame type is tx status
   if(serialFrame[3] != XBEE_API_TX_STATUS) {
      return false;
   }
   uint8_t offset= 1;
   uint16_t frameLength; //the 2 byte length as specified in the serial frame
   //pull out the frame length from the serial frame, this is used to find out
   //the size of the payload
   offset+= deSerializeInt16(&frameLength, &serialFrame[offset]);
   offset++; //skip over 1 byte because we already confirmed that the frame
             //opcode is correct
   //pull out the frame ID
   offset+= deSerializeInt8(&(txStatusFrame->frameID), &serialFrame[offset]);
   //pull out 16 bit address
   offset+= deSerializeInt16(&(txStatusFrame->addr16bit), &serialFrame[offset]);

   //pull out retry count
   txStatusFrame->retryCount= serialFrame[offset];
   offset++;

   //pull out delivery status
   txStatusFrame->deliveryStatus= serialFrame[offset];
   offset++;

   //pull out discovery status
   txStatusFrame->discoverStatus= serialFrame[offset];
   offset++;

   return true;
}

/**
 * Write bytes out to the xbee's usart port.  This function WILL block if the
 * buffer is full until every byte is written.
 * @param bytes The array of bytes to write.
 * @param numBytes The number of bytes to write (up to 255)
 */
static void writeRawToUSART(const uint8_t* bytes, uint8_t numBytes) {
   //write put each byte in the array
   for(uint8_t i= 0; i < numBytes; i++) {
      //putbyte function is placed in a while loop so it will continue to try
      //to place byte in the buffer until it succeeds
      while(!USART_TXBuffer_PutByte(sSetup->usart_info, bytes[i]));
   }
}
