/*
 * xmega128a1_board.c
 *
 *  Created on: Nov 28, 2010
 *      Author: hsrmqp
 */

#include "avr/io.h"

#include "includes.h"
#include "utils/system_time.h"

#include "xmega128a1_board.h"
#include "xmega_drivers.h"

#include "encoder_counter.h"
#include "mouse_sensor.h"
#include "hbridge.h"
#include "servo.h"
#include "xbee.h"
#include "rangefinder.h"
#include "beacondetector.h"
#include "utils/debug_module.h"
#include "communications.h"

//Global Setup Structs
//SPI
static SPI_Master_t spiDSetup;
//USART
#ifdef USART_XBEE_ENABLE
static USART_data_t xbee_USART_data;
#endif

#ifdef USART_DEBUG_ENABLE
static USART_data_t debug_USART_data;
#endif


//Static function prototypes (for XMega Level Drivers)
static void setSysClock32Mhz(bool lock);
static void initADC(void);
static void initMotorPWM(void);
static void initServoPWM(void);
static void initTimers(void);
static void initUSART(void);
static void initGPIO(void);
static void initInterrupts(void);


//Static function prototypes (for Chip Level Drivers)
static void initUSART(void);
static void superInitEncoderCounter(void);
static void superInitHBridge(void);
static void superInitServo(void);
static void superInitMouseSensor(void);
static void initSPI_EncCntr(SPI_Master_t* spiMasterSetup); //helpers for spi devices
static void initSPI_MouSense(SPI_Master_t* mouSPIsetup);
static void superInitXbee(void);
static void superInitRangeFinder(void);
static void superInitBeaconDetector(void);
#ifdef USART_DEBUG_ENABLE
static void superInitDebugModule(void);
#endif

/**
 * Perform basic initialization to set up the xmega peripherals/system to work for our project.
 */
void low_level_init() {
   //Clock
   setSysClock32Mhz(true);
   //Timers
   initTimers();
   //ADC
   initADC();
   //DMA
   //initDMA();
   //PWM
   initMotorPWM();
   initServoPWM();
   //USART
   initUSART();
   //GPIO
   initGPIO();
   //Inerrupts
   initInterrupts();
}

/**
 * Init chip level driver stuff
 */
void chip_level_init() {

   //setup Debugging support
#ifdef USART_DEBUG_ENABLE
   superInitDebugModule();
#endif

   //setup encoder
   superInitEncoderCounter();

   //Setup H Bridge
   superInitHBridge();

   //Setup Mouse Sensor
   superInitMouseSensor();

   //Setup Servo
   superInitServo();

   //Setup Xbee
#ifdef USART_XBEE_ENABLE
   superInitXbee();
#endif

   //Setup Range Finders
   superInitRangeFinder();

   //Setup beacon detectors
   superInitBeaconDetector();

}

/**
 * Sets the System Clock to run at 32Mhz and disables all other clocks
 * Will not return until clock source is steady
 *
 * @param lock whether or not to lock the system clock config until the next reset
 */
static void setSysClock32Mhz(bool lock){
	// Enable the 32MHz internal oscillator which is used as the system clock
	CLKSYS_Enable( OSC_RC32MEN_bm );
	do {} while ( CLKSYS_IsReady( OSC_RC32MRDY_bm ) == 0 );
	// Enable the 32KHz internal oscillator which is used as reference for the DFLL
	CLKSYS_Enable( OSC_RC32KEN_bm );
	do {} while ( CLKSYS_IsReady( OSC_RC32KRDY_bm ) == 0 );
	//Start Autocalibration without the external crystal
	CLKSYS_AutoCalibration_Enable( OSC_RC32MCREF_bm, false );
	//Set the 32MHz clock as the main clock source
	CLKSYS_Main_ClockSource_Select( CLK_SCLKSEL_RC32M_gc );
	//Disable the unused 2MHz clock
	CLKSYS_Disable( OSC_RC2MEN_bm );

	//Lock the calibratioon (if desired) until a hard chip reset
	if (lock){
		CLKSYS_Configuration_Lock();
	}

}

/**
 * Set up the ADC.
 */
/**
 * This function, sets up ADC A to take a single ended unsigned reading on pin 1.
 * The funciton will automatically set the "negative" pin to be ground so that we
 * don't have to worry about setting that.  We will set the function to run at 500KSPS
 * since this is well within the range of the ADC.  Since we are running at 32MHz this
 * will require a prescaler of 64.  The ADC has 12 bit resolution by default.
 *
 * We need to set this to use the 1V reference and I dunno how to do that
 */
static void initADC(void){

   /* Move stored calibration values to ADC A */
   ADC_CalibrationValues_Load(&ADCA);

    /* Set up ADC A to have unsigned conversion mode and 12 bit resolution. */
   ADC_ConvMode_and_Resolution_Config(&ADCA, false, ADC_RESOLUTION_12BIT_gc);

    // The ADC has different voltage reference options, controlled by the REFSEL bits in the
    // REFCTRL register. Here the internal accurate 1.00V reference is selected
   ADC_Reference_Config(&ADCA, ADC_REFSEL_INT1V_gc);

   // The clock into the ADC decide the maximum sample rate and the conversion time, and
    // this is controlled by the PRESCALER bits in the PRESCALER register. Here, the
    // Peripheral Clock is divided by 64 ( gives 500 KSPS with 32Mhz clock )
   ADC_Prescaler_Config(&ADCA, ADC_PRESCALER_DIV64_gc);

    // The used Virtual Channel (CH0) must be set in the correct mode
    // In this task we will use single ended input, so this mode is selected

    /* Setup channel 0 to have single ended input. */
   ADC_Ch_InputMode_and_Gain_Config(&ADCA.CH0, ADC_CH_INPUTMODE_SINGLEENDED_gc, ADC_CH_GAIN_1X_gc);

   /* Setup channel 1 to have single ended input. */
  ADC_Ch_InputMode_and_Gain_Config(&ADCA.CH1, ADC_CH_INPUTMODE_SINGLEENDED_gc, ADC_CH_GAIN_1X_gc);


    // Setting up the which pins to convert for the first IR Sensor on PortA Pin0.
    // Note that the negative pin is internally connected to ground
   ADC_Ch_InputMux_Config(&ADCA.CH0, ADC_CH_MUXPOS_PIN0_gc, ADC_CH_MUXNEG_PIN0_gc);

   // Setting up the which pins to convert for the second IR Sensor on PortA Pin1.
   // Note that the negative pin is internally connected to ground
  ADC_Ch_InputMux_Config(&ADCA.CH1, ADC_CH_MUXPOS_PIN1_gc, ADC_CH_MUXNEG_PIN0_gc);


    // Before the ADC can be used it must be enabled
   ADC_Enable(&ADCA);


    // Wait until the ADC is ready
    ADC_Wait_32MHz(&ADCA);

    // In the while(1) loop, a conversion is started on CH0 and the 8 MSB of the result is
    // ouput on the LEDPORT when the conversion is done

   /* Get offset value for ADC B.  */
//   return(ADC_Offset_Get_Unsigned(&ADCA, &(ADCA.CH0), true));
}

/**
 * This function will initialize the motor PWM
 *
 * Motors are running off of TC0 on Port C.  They are running off of CCA for Motor 0 and CCB for Motor 1
 * This function will set the frequency to 8KHz, and set the current period to be 0, so that the motors
 * will not move when they are initialized
 */

static void initMotorPWM(void){
   //Set the clock source for TCC0 to be the system clock with a prescale of 4
   TC0_ConfigClockSource(&TCC0, TC_CLKSEL_DIV4_gc);

   //Set the motor PWM to be a max period of 1000 count,MOTOR_PWM_PERIOD
   //with our selected clock this will give us a 8KHz frequency
   TC_SetPeriod( &TCC0, MOTOR_PWM_PERIOD);

   //Select the Waveform Generation Mode (We want Single Slope PWM)
   TC0_ConfigWGM( &TCC0, TC_WGMODE_SS_gc );

   //We need to set CCA and CCB to be the Compare / Capture Channels
   TC0_EnableCCChannels( &TCC0, TC0_CCAEN_bm); // CCA will be Motor 0
   TC0_EnableCCChannels( &TCC0, TC0_CCBEN_bm); // CCB will be Motor 1

   //We need to set the Period of CCA & CCB to be zero, this is so the motors will not move after initialization
   TC_SetCompareA( &TCC0, 0); //Set CCA period to be Zero
   TC_SetCompareB( &TCC0, 0); //Set CCB period to be Zero

   /* The corresponding port pins MUST be output for the Waveform to be visible */
   PORTC.DIRSET = PIN0_bm | PIN1_bm; //We only want to set pins 0 and 1 to be output
}

/**SERVO_MAX_CC_LEVEL
 * This function will initalize the servo PWM
 *
 * The Servo is running off of TC1 on Port C, They are running off of CCA.  This function will set the
 * frequency to be 50Hz, and set the current period to be 0, so that the servo will be in a "home" position
 * when this function starts
 */
static void initServoPWM(void){
   //Set the clock source for TCC1 to be the system clock with a prescale of 64
   TC1_ConfigClockSource(&TCC1, TC_CLKSEL_DIV64_gc);

   //Set the servo PWM to be a max period of 40000, with our selected clock this will give us a 50Hz frequency
   TC_SetPeriod( &TCC1, SERVO_PWM_MAX_PERIOD);

   //Select the Waveform Generation Mode (We want Single Slope PWM)
   TC1_ConfigWGM( &TCC1, TC_WGMODE_SS_gc );

   //We need to set CCA to be the Compare / Capture Channel
   TC1_EnableCCChannels( &TCC1, TC1_CCAEN_bm); // CCA will be Servo 0

   //We need to set the Period of CCA to be zero, this will put the servo in "home" position
   TC_SetCompareA( &TCC1, SERVO_MIN_CC_LEVEL); //Set CCA period to be Zero

   /* The corresponding port pins MUST be output for the Waveform to be visible */
   PORTC.DIRSET = PIN4_bm; //We only want to set pin 4 to be output
}

/**
 * Set up the timer for the heartbeat of uCOS and system time
 */
static void initTimers(void) {
   //  UCOS
   //Set the clock source for TCF1 to be sysClock/64
   TC1_ConfigClockSource(&TCF1, TC_CLKSEL_DIV64_gc);

   //use normal (i.e. none) waveform generation mode
   TC1_ConfigWGM(&TCF1, TC_WGMODE_NORMAL_gc);

   //set the period to 5000 (100Hz)
   TC_SetPeriod(&TCF1, 5000);

   //set Interrupt level.  use low level because the resolution on the os timer
   //is supposed to be fairly poor anyways
   TC1_SetOverflowIntLevel( &TCF1, TC_OVFINTLVL_LO_gc );


   //  System Timer
   //Set the clock source for TCF0 to be sysClock/64
   TC0_ConfigClockSource(&TCF0, TC_CLKSEL_DIV64_gc);

   //use normal (i.e. none) waveform generation mode
   TC0_ConfigWGM(&TCF0, TC_WGMODE_NORMAL_gc);

   //set the period to 50 (10kHz)
   TC_SetPeriod(&TCF0, 50);

   //set Interrupt level.  use high level so it will take priority
   //over all other interrupts
   TC0_SetOverflowIntLevel( &TCF0, TC_OVFINTLVL_HI_gc );
}

/*
 * This function sets the GPIO
 */
static void initGPIO(){

   //PORTA
   PORTA.DIRCLR= PIN0_bm | PIN1_bm | PIN2_bm | PIN3_bm; //use pins 2 and 3 for candle sense input
														//use pins 0 and 1 for the ir detectors

   //PORTB
   PORTB.DIRCLR= PIN0_bm; //use pin 0 for bump sense input

   //PORTC
   PORTC.DIRSET= PIN2_bm | PIN3_bm | PIN5_bm | PIN6_bm | PIN7_bm;  //set pins 2,3,5,6 for motor (pin7 for servo enable)
                                             //direction control output
   PORTC.OUTCLR= PIN2_bm | PIN3_bm | PIN5_bm | PIN6_bm;  //default direction bits to 0
                                             //to put brakes on motors
   PORTC.OUTSET= PIN7_bm; //disable servo on startup



   //PORTD
   PORTD.DIRSET= PIN0_bm | PIN1_bm; //use pin 0 and pin1 for mouse sensor cs
   PORTD.DIRSET= PIN4_bm;           //Set to output to keep spi in master mode
   PORTD.OUTSET= PIN0_bm | PIN1_bm; //default pins to high (disabled)

   //PORTE
   PORTE.DIRSET= PIN0_bm | PIN1_bm | PIN4_bm | PIN5_bm;  //use pins 0,1,4,5 for chip select
                                             //and enable for IR sensors and
                                             //encoders
   PORTE.OUTSET= PIN0_bm | PIN1_bm | PIN4_bm | PIN5_bm; //default pins to high (disabled)

   //PORTF-PORTP not used
   PORTK.DIRSET= PIN7_bm | PIN6_bm | PIN5_bm | PIN4_bm; //LED's
   PORTK.OUT= PIN7_bm | PIN6_bm | PIN5_bm | PIN4_bm;

}

/**
 * Set up the PMIC for round robin scheduling and allow high, medium, and low
 * level interrupts to occur.  This function does NOT enable the global
 * interrupt enable bit.
 */
static void initInterrupts(void) {
   PMIC.CTRL|= PMIC_RREN_bm | PMIC_HILVLEN_bm | PMIC_MEDLVLEN_bm | PMIC_LOLVLEN_bm;
}
//init function for encoder counter chip

static void initUSART(void) {
   //XBEE
#ifdef USART_XBEE_ENABLE
	   // Use USARTD0 and initialize buffers
	   USART_InterruptDriver_Initialize(&xbee_USART_data, &USARTD0, USART_DREINTLVL_LO_gc);

	   // USARTD0, 8 Data bits, No Parity, 1 Stop bit
	   USART_Format_Set(xbee_USART_data.usart, USART_CHSIZE_8BIT_gc,
						USART_PMODE_DISABLED_gc, false);

	   // Enable RXC interrupt
	  USART_RxdInterruptLevel_Set(xbee_USART_data.usart, USART_RXCINTLVL_LO_gc);

     // Set Baudrate to 115200 bps:
		int8_t xbeeScale = -6; //Done to supress a warning
		USART_Baudrate_Set(&USARTD0, 1047 , (uint8_t)xbeeScale);

		/* Enable both RX and TX. */
		USART_Rx_Enable(xbee_USART_data.usart);
		USART_Tx_Enable(xbee_USART_data.usart);

		// PD3 (TX) as output
		PORTD.DIRSET   = PIN3_bm;
		// PD2 (RX) as input
		PORTD.DIRCLR   = PIN2_bm;
#endif


    //DEBUG USART
#ifdef USART_DEBUG_ENABLE
		// Use USARTE0 and initialize buffers
		USART_InterruptDriver_Initialize(&debug_USART_data, &USARTE0, USART_DREINTLVL_LO_gc);

		// USARTE0, 8 Data bits, No Parity, 1 Stop bit
		USART_Format_Set(debug_USART_data.usart, USART_CHSIZE_8BIT_gc,
						 USART_PMODE_DISABLED_gc, false);

		// Enable RXC interrupt
	   USART_RxdInterruptLevel_Set(debug_USART_data.usart, USART_RXCINTLVL_LO_gc);


		 int8_t debugScale = -6; //Done to surpress a warning
		 USART_Baudrate_Set(&USARTE0, 1047 , (uint8_t)debugScale);

		 /* Enable both RX and TX. */
		 USART_Rx_Enable(debug_USART_data.usart);
		 USART_Tx_Enable(debug_USART_data.usart);

		 // PD3 (TX) as output
		 PORTE.DIRSET   = PIN3_bm;
		 // PD2 (RX) as input
		 PORTE.DIRCLR   = PIN2_bm;
#endif


}

static void superInitRangeFinder(void){
   static range_finder_setup_t rangeFinderSetup;

   static ADC_t* rangeFinderADCs[] ={&ADCA, &ADCA};
   static ADC_CH_t* rangeFinderChannels[]={&(ADCA.CH0), &(ADCA.CH1)};
   static PORT_t* rangeFinderEnablePorts[]={&PORTE, &PORTE};
   static uint8_t rangeFinderEnablePins[]={PIN4_bm, PIN5_bm};
   static int8_t rangeFinderOffsets[2];
   static int32_t rangeFinderKprime[]= {34425, 34425};
   static int32_t rangeFinderBprime[]= {-143, -143};
   static int32_t rangeFinderK[]= {1, 1};

   rangeFinderOffsets[0] = ADC_Offset_Get_Unsigned(&ADCA, &(ADCA.CH0), false);
   rangeFinderOffsets[1] = ADC_Offset_Get_Unsigned(&ADCA, &(ADCA.CH1), false);

   rangeFinderSetup.adc = rangeFinderADCs;
   rangeFinderSetup.channel = rangeFinderChannels;
   rangeFinderSetup.enablePort = rangeFinderEnablePorts;
   rangeFinderSetup.enablePin = rangeFinderEnablePins;
   rangeFinderSetup.offset = rangeFinderOffsets;

   //range finder calibration values
   rangeFinderSetup.kprime= rangeFinderKprime;
   rangeFinderSetup.bprime= rangeFinderBprime;
   rangeFinderSetup.k= rangeFinderK;


   init_rangefinder(&rangeFinderSetup);
}

//Initialize the beacon detectors
static void superInitBeaconDetector(void){
	static beacon_detector_setup_t beaconDetectorSetup;

	// Declare input ports and pins respectively
	static PORT_t* beaconDetectorPorts[]={&PORTA, &PORTA};
	static uint8_t beaconDetectorPins[]={PIN2_bm, PIN3_bm};

	beaconDetectorSetup.port = beaconDetectorPorts;
	beaconDetectorSetup.pin = beaconDetectorPins;

	init_beaconDetector(&beaconDetectorSetup);
}

//Init Encoder Counters
static void superInitEncoderCounter(void) {
   static encoder_counter_setup_t encSetup;
   static PORT_t* encCntrSSPorts[]= {&PORTE, &PORTE};
   static uint8_t encCntrSSPinMasks[]= {PIN0_bm, PIN1_bm};

   encSetup.spiDevice= &spiDSetup;
   encSetup.SPIinitFunc= &initSPI_EncCntr;
   encSetup.ssPort= encCntrSSPorts;
   encSetup.ssPinMask= encCntrSSPinMasks;

   init_encoder_counter(&encSetup);
}

//init function for Hbridge
static void superInitHBridge(void){
	static hbridge_setup_t hbridgeSetup;
	static direcPins_t m0Pins;
	static direcPins_t m1Pins;
	m0Pins.dir0 = PIN2_bm;
	m0Pins.dir1 = PIN3_bm;
	m1Pins.dir0 = PIN5_bm;
	m1Pins.dir1 = PIN6_bm;
	static direcPins_t motorDirPins[2];
	motorDirPins[0]= m0Pins;
	motorDirPins[1]= m1Pins;
	static TC0_t* pwmTimers[]= {&TCC0, &TCC0};
	static TCCOMPARE_t compare[]= {TC_CCA_REG, TC_CCB_REG};
	static PORT_t* outputPorts[]= {&PORTC, &PORTC};

	hbridgeSetup.motorPWMTimer = pwmTimers;
	hbridgeSetup.motorCompare = compare;
	hbridgeSetup.motorOutputPort = outputPorts;
	hbridgeSetup.motorDirPins = motorDirPins;

	init_hbridge(&hbridgeSetup);
}

//init function for the servo
static void superInitServo(void){
	static servo_setup_t servoSetup;

	static TC1_t* servoPWMTimers[]={&TCC1};
	static TCCOMPARE_t servoPWMCCRegisters[]={TC_CCA_REG};
	static PORT_t* servoEnablePorts[]={&PORTC};
	static uint8_t servoEnablePins[]={PIN7_bm};

	servoSetup.servoPWMTimer = servoPWMTimers;
	servoSetup.servoCompare = servoPWMCCRegisters;
	servoSetup.servoEnablePort = servoEnablePorts;
	servoSetup.servoEnablePin = servoEnablePins;
	servoSetup.offset= 7;

	init_servo(&servoSetup);
}

//init function for the mouse sensors
static void superInitMouseSensor(void){
   static mouse_sensor_setup_t mouseSetup;

   static PORT_t* mouseSSPorts[]= {&PORTD, &PORTD};
   static uint8_t mouseSSPinMasks[]= {PIN0_bm, PIN1_bm};

   mouseSetup.spiDevice = &spiDSetup;
   mouseSetup.SPIinitFunc= &initSPI_MouSense;
   mouseSetup.ssPort = mouseSSPorts;
   mouseSetup.ssPinMask = mouseSSPinMasks;

   init_mouse_sensor(&mouseSetup);
}

#ifdef USART_XBEE_ENABLE
static void superInitXbee(void) {
   static xbee_setup_t xbeeSetup;

   xbeeSetup.usart_info= &xbee_USART_data;
   xbeeSetup.CTS_Port= &PORTE;
   xbeeSetup.CTS_PinMask= PIN7_bm;
   xbeeSetup.RTS_Port= &PORTE;
   xbeeSetup.RTS_PinMask= PIN6_bm;
   xbeeSetup.callback= remote_RXDriverCallback;

   init_xbee(&xbeeSetup);
}
#endif
#ifdef USART_DEBUG_ENABLE
static void superInitDebugModule(void) {

   static debug_module_setup_t debugSetup;

   debugSetup.usart= &debug_USART_data;

   init_debug_module(&debugSetup);

}
#endif

static void initSPI_EncCntr(SPI_Master_t* encSPIsetup) {
   //set up spi port D as a master with normal polarity and rising edge sample
   //interrupts are disabled and the clock speed is 1MHz
   SPI_MasterInit(&spiDSetup, &SPID, &PORTD, false,
                  SPI_MODE_0_gc, SPI_INTLVL_OFF_gc, true, SPI_PRESCALER_DIV64_gc);
}

static void initSPI_MouSense(SPI_Master_t* mouSPIsetup) {
   //set up spi port D as a master with inverse polarity and rising edge sample
   //interrupts are disabled and the clock speed is 1MHz
   SPI_MasterInit(&spiDSetup, &SPID, &PORTD, false,
                  SPI_MODE_3_gc, SPI_INTLVL_OFF_gc, true, SPI_PRESCALER_DIV64_gc);
}

/******************************************************************************
 *                                     ISRs
 *****************************************************************************/
/********************       OS System Timer ISR            *******************/
ISR( TCF1_OVF_vect, ISR_NAKED ) {
   UCOS_ENTER_ISR("TCF1_OVF");
   OSTimeTick();
   UCOS_EXIT_ISR();
}

/*****************       System Clock NOT KNOWN TO UCOS     *******************/
ISR( TCF0_OVF_vect, ISR_NAKED ) {
   ENTER_ISR();
   system_time_ISR_handler();
   EXIT_ISR();
}

//XBEE USART
//  Receive complete interrupt service routine.
//  Calls the common receive complete handler with pointer to the correct USART
//  as argument.
#ifdef USART_XBEE_ENABLE
ISR( USARTD0_RXC_vect, ISR_NAKED )
{
   UCOS_ENTER_ISR("USARTD0_RXC");
   USART_RXComplete_Sem( &xbee_USART_data, xbee_rx_data_sem );
   UCOS_EXIT_ISR();
}


//  Data register empty  interrupt service routine.
//  Calls the common data register empty complete handler with pointer to the
//  correct USART as argument.

ISR( USARTD0_DRE_vect, ISR_NAKED )
{
   UCOS_ENTER_ISR("USARTD0_DRE");
   USART_DataRegEmpty( &xbee_USART_data );
   UCOS_EXIT_ISR();
}
#endif

//DEBUG USART
//  Receive complete interrupt service routine.
//  Calls the common receive complete handler with pointer to the correct USART
//  as argument.
#ifdef USART_DEBUG_ENABLE
#ifndef DEBUG_BLOCKING_PRINTS
ISR( USARTE0_RXC_vect, ISR_NAKED )
{
   UCOS_ENTER_ISR("USARTE0_RXC");
   USART_RXComplete( &debug_USART_data );
   UCOS_EXIT_ISR();
}

//  Data register empty  interrupt service routine.
//  Calls the common data register empty complete handler with pointer to the
//  correct USART as argument.
ISR( USARTE0_DRE_vect, ISR_NAKED )
{
   UCOS_ENTER_ISR("USARTE0_DRE");
   USART_DataRegEmpty( &debug_USART_data );
   UCOS_EXIT_ISR();
}
#endif
#endif

